#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"
#include "mult.h"

extern FILE *mltemp;
extern FILE *bdsyntemp;
extern FILE *outfile;
extern int watch_espresso;
extern int assigns_to_ML;
extern int collapse_level;
extern int short_1_bit;

static int assign_num;             /* number of current assignment */
static int pre_lif();
static int pre_lif_process();

/*
 * This lififies (runs through ml) all assignments that are necessary
 * for bdsyn's final output.
 */
lifassign( root )
node *root;
{
    node *assign;
    node *lhs;			/* LHS of assign */

    assign_num = 0;
    symtab_foreach(global_syms, pre_lif, T_IS_OUTPUT_PORT);

    /* collapse each expression with needlif. */
    for (assign = root; assign != NULL; assign = assign->next) {
	lhs = assign->un.assign.lexpr;
	if (assign->un.assign.needlif) {
	    lif_process(lhs->un.bdvar.name, lhs->un.bdvar.lbit,
		    lhs->un.bdvar.rev);
/*
	    if (!lhs->un.bdvar.name->u.state.is_temp) {
*/
	    if (collapse_level > 0) {
		(void) fprintf(outfile, "%s%s",
			assign->un.assign.rexpr->un.lif.names,
			assign->un.assign.rexpr->un.lif.table);
	    }
	}
    }
    assigns_to_ML = assign_num;
    if ( watch_espresso >= 1 ) {
	(void) printf("\n");
    }
}


/*
 * lif_process will create a lif table for the given rev of the
 * given var.
 */
lif_process( name, nbit, rev )
symbol *name;			/* symbol pointer to var */
int nbit;			/* bit number of var */
usage *rev;			/* rev of var */
{
    node *assign;		/* Assign statement for var */
    node *lhs;			/* LHS of assign */
    symbol *symbolent;		/* main symtab entry for LHS */
    char myname[512];		/* name of var */
    char buf[1030];
    char *opname;		/* name of RHS of assign */
    int buflen;			/* number of chars from fgets */
    int tablelen;		/* chars in lif table >= buflen */
    int namelinelen;		/* chars on one line >= buflen for .names */
    node *lifnode;		/* result of ml */

    if ( rev == NIL(usage) )
	ierror("No rev in lif_process", NIL(node));
    assign = rev->assign;
    if ( assign->un.assign.rexpr->type != T_LIF ) {   /* If not already done */
	lhs = assign->un.assign.lexpr;
	symbolent = lhs->un.bdvar.name;

	/* give LHS's name and bit to node_lif so it can forward subst */
	opname = node_lif(assign->un.assign.rexpr, symbolent->name,
		lhs->un.bdvar.lbit, ++assign_num);

	/* last rev to Output Port has no rev string */
	if ( (symbolent->type == T_IS_OUTPUT_PORT) && islastupdate(lhs) ) {
		if (short_1_bit && symbolent->u.state.nobit) {
	    	(void) sprintf( myname, "%s", name->name );
		}
		else {
	   		(void) sprintf( myname, "%s<%d>", name->name, nbit );
		}
	} else {
	    (void) sprintf( myname,"%s<%d>%d.%d", name->name, nbit,
		    rev->block->id, rev->rev );
	}
	/* Hook up LHS and RHS */
	(void) fputs(".names ", bdsyntemp);
	(void) fputs(opname, bdsyntemp);
	putc(' ', bdsyntemp);
	(void) fputs(myname, bdsyntemp);
	/* table is "1 1" */
	if (collapse_level > 0) {
	    (void) fputs("\n1 1\n.end\n", bdsyntemp);
	} else {
	    (void) fputs("\n1 1\n", bdsyntemp);
	}
	if ( watch_espresso >= 2 ) {
	    (void) printf( ".names %s %s\n1 1\n.end\n", opname, myname );
	}
	FREE( opname );
	/* send out for flatten */
	if ( watch_espresso >= 1 ) {
	    putchar('#');
	    if( assign_num % 75 == 0 ) {
		(void) printf("%d\n", assign_num);
	    }
	    (void) fflush( stdout );
	}
	(void) fflush(bdsyntemp);

	/* read table back from mis */
	if (collapse_level > 0) {
	    node_dispose(assign->un.assign.rexpr);
	    lifnode = assign->un.assign.rexpr = node_create_fi( T_LIF, assign );
	    lifnode->un.lif.table = ALLOC( char, 32 );
	    *lifnode->un.lif.table = '\0';
	    tablelen = 0;
	    if ( fgets( buf, 1024, mltemp ) == NULL )
		ierror("Unexpected EOF from mis", (node *) NULL);
	    while ( strcmp( buf, ".end\n" ) != 0 ) {
		buflen = strlen( buf );
		if ( *buf != '.' ) {
		    lifnode->un.lif.table = REALLOC( char,
			    lifnode->un.lif.table, tablelen + buflen + 1 );
		    (void) strcpy( lifnode->un.lif.table + tablelen, buf );
		    tablelen += buflen;
		    while( buf[ buflen - 1 ] != '\n' ) {
			(void) fgets( buf, 1024, mltemp );
			buflen = strlen( buf );
			lifnode->un.lif.table = REALLOC( char,
				lifnode->un.lif.table, tablelen + buflen + 1 );
			(void) strcpy( lifnode->un.lif.table + tablelen, buf );
			tablelen += buflen;
		    }
		} else if ( strncmp( buf, ".names", 6 ) == 0 ) {
		    namelinelen =buflen;
		    lifnode->un.lif.names = ALLOC( char, buflen + 1 );
		    (void) strcpy( lifnode->un.lif.names, buf );
		    while( buf[ buflen - 1 ] != '\n' ) {
			(void) fgets( buf, 1024, mltemp );
			buflen = strlen( buf );
			lifnode->un.lif.names = REALLOC( char,
				lifnode->un.lif.names, namelinelen + buflen + 1 );
			(void) strcpy( lifnode->un.lif.names + namelinelen, buf );
			namelinelen += buflen;
		    }
		} else {	/* ignore it */
		    while( buf[ buflen - 1 ] != '\n' ) {
			(void) fgets( buf, 1024, mltemp );
			buflen = strlen( buf );
		    }
		}
		(void) fgets( buf, 1024, mltemp );
	    }
	}
    }
}


/*
 * This walks over the transistive fanin and marks nodes for which lif
 * tables will be required.  Called on each output port by lifassign.
 */
static int pre_lif( p )
symbol *p;
{
    int i;
    char warn[512];
    node *assign;		/* last assign to Output port */

    for ( i = p->u.state.hibit; i>= p->u.state.lobit ; i-- ) {
	if ( (p->u.state.versions != NULL) &&
	(p->u.state.versions[i - p->u.state.lobit].assigned.current == 1) ) {
	    if ( p->u.state.versions[i - p->u.state.lobit].assigned.previous == 1 ) {
		(void) sprintf(warn,
			"Output port %s<%d> assigned by two disjoint routines",
			p->name, i);
		warning(warn, NIL(node));
	    }
	    p->u.state.versions[i - p->u.state.lobit].assigned.previous = 1;
	    p->u.state.versions[i - p->u.state.lobit].assigned.current = 0;
	    assign = p->u.state.versions[i-p->u.state.lobit].recent->assign;
	    if (!assign->un.assign.prelifed) {   /* If not already done */
		(void) pre_lif_process(assign->un.assign.rexpr, p->name, i);
	    }
	    assign->un.assign.needlif = 1;
	    assign->un.assign.prelifed = 1;
	}
    }
    return(1);                  /* tell symtab_foreach to continue */
}


/*
 * This will return true if this revision is the last to the
 * variable.
 */
islastupdate( varnode )
node *varnode;			/* revision in question */
{
    int lodecl;			/* low declared bit of var */
    symbol *symbolent;		/* entry in symbol table */
    revision *vers;		/* revision info for var */

    symbolent = varnode->un.bdvar.name;
    lodecl = symbolent->u.state.lobit;
    vers = symbolent->u.state.versions;
    return (vers[varnode->un.bdvar.lbit-lodecl].recent ==
	    varnode->un.bdvar.rev);
}


/*
 * Will scan a fanout tree recursively backward from the fanouts.
 * It will mark the nodes that require a lif table in the final
 * output.
 * Will return true if the node below is a DONT_CARE.  This way we
 * can easily force DC's to get their own output table.  This also gives
 * a quick check that DC's are in simple assignments.
 */
static int pre_lif_process( root, varname, bit )
node *root;
char *varname;			/* name of lhs */
int bit;			/* bit of lhs */
{
    symbol *symbolent;		/* symbol table entry */
    node *lastx;		/* pointer to hashed rhs for forward subst */
    node *op;			/* operand of unary AND or OR */

    switch ( root->type ) {
	
	case T_IS_STATE:
	    symbolent = root->un.bdvar.name;
	    if (root->un.bdvar.rev == NIL(usage)) {
		lastx = NIL(node);
	    } else {
		lastx = root->un.bdvar.rev->assign;
	    }

	    /* process DONT_CARE */
	    if ( upper_strcmp( symbolent->name, "DONT_CARE" ) == 0 ) {
		if (lastx != NIL(node)) {
		    error("Assignment TO DONT_CARE detected", root);
		} else {
		    return 1;
		}
	    }

	    /* Scan down tree for variable (unless there is no tree) */
	    if (lastx != NIL(node)) {
		/* check for forward subst condition */
		if ((collapse_level == 2) &&
			(upper_strcmp(varname, symbolent->name ) == 0) &&
			(root->un.bdvar.lbit == bit) ) {
		    if ( !lastx->un.assign.prelifed ) {
			/* Unless DC is detected, create no lif table
			 * since we want to collapse assignments to
			 * a single variable */
			if (pre_lif_process(lastx->un.assign.rexpr,
				varname, bit)) {
			    lastx->un.assign.needlif = 1;
			}
		    }
		} else {
		    if (!lastx->un.assign.prelifed) {
			(void) pre_lif_process(lastx->un.assign.rexpr,
				root->un.bdvar.name->name,
				root->un.bdvar.lbit);
		    }
		    lastx->un.assign.needlif = 1;
		}
		lastx->un.assign.prelifed = 1;
	    }
	    break;


        case T_AND: case T_OR: case T_NAND:
	case T_NOR: case T_XOR: case T_EQV:
	    if (pre_lif_process(root->un.binary.lexpr, varname, bit) ||
		    pre_lif_process(root->un.binary.rexpr, varname, bit)) {
		error("DONT_CARE must be in a simple assignment", root);
	    }
	    break;

        case T_NOT:
	    if (pre_lif_process(root->un.unary.expr, varname, bit)) {
		error("DONT_CARE must be in a simple assignment", root);
	    }
	    break;

        case T_UNARY_OR: case T_UNARY_AND:
	    for(op = root->un.unary.expr; op != NULL; op = op->conc) {
		if (pre_lif_process(op, varname, bit)) {
		    error("DONT_CARE must be in a simple assignment", root);
		}
	    }
	    break;

	case T_IS_CONSTANT:
	    break;
	
	default:
	    ierror("pre_lif_process on ?", root);

    }
    return 0;
}
