#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"
#include "eval.h"


/*
 * meta is the routine that performs replacement of meta variables
 * with constants.  It also unrolls FOR loops.
 * Meta is also forced to process SELECTONE and SELECTALL ( turn them
 * into IFs ), and evaluate all IF and SELECT conditions.  This is
 * necessary to correctly propagate all sorts of meta structures.
 */
meta(root)
node *root;
{
    node *current;		/* current node being processed */
    int temp;
    node *lhs;			/* LHS of assign */
    int loopindex;
    node *from, *to, *by;
    node *loopmeta;		/* meta var = un.bdif.var */
    node *forblock;		/* block replicated by FOR */
    node *forreplace;		/* whole block to replace to FOR */
    node *forlast;		/* current last node of forreplace */
    node *nindex;		/* node index */
    node *nextsave;		/* save what node follows on ->next */
    int constant;			/* is the expression constant */
    int width;			/* width of expression */
    long mask;			/* mask off high bit of short meta */

    for ( current = root; current != NULL; current = current->next ) {
	if ( current->conc != NULL ) {
	    meta( current->conc );
	}
	switch ( current->type ) {

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
		meta( current->un.unary.expr );
		break;
    
	    case T_AND: case T_OR: case T_NAND:
	    case T_NOR: case T_XOR: case T_EQV:
	    case T_EQL: case T_NEQ: case T_LEQ:
	    case T_GEQ: case T_LSS: case T_GTR:
	    case T_SL0: case T_SR0: case T_SL1:
	    case T_SR1: case T_SLR: case T_SRR:
	    case T_SXT: case T_ZXT: case T_OXT:
	    case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
		meta( current->un.binary.rexpr );
		meta( current->un.binary.lexpr );
		break;
    
	    case T_BIT:
		meta( current->un.ternary.rexpr );
		meta( current->un.ternary.mexpr );
		meta( current->un.ternary.lexpr );
		break;
    
	    case T_EQ:
		meta( current->un.assign.rexpr );
		lhs = current->un.assign.lexpr;
		if ( (lhs->type == T_IS_STATE) &&
		     (lhs->un.bdvar.name->type == T_IS_META) ) {
		    current->un.assign.rexpr =
			    node_eval( current->un.assign.rexpr );
		    if ( (current->un.assign.rexpr->type != T_IS_CONSTANT) ||
			    current->un.assign.rexpr->conc != NULL )
			error("Assignment of non-constant to meta", current);
		    lhs->un.bdvar.name->VALUE =
			    current->un.assign.rexpr->un.constant.num;
		    /* turn assign into no-op */
		    NODE_DISPOSE( current->un.assign.lexpr );
		    NODE_DISPOSE( current->un.assign.rexpr );
		    current->type = T_IS_CONSTANT;
		    current->un.constant.num = current->un.constant.dont_care = 0;
		    current->un.constant.hbit = 0;
		} else {
		    meta( lhs );
		}
		break;

	    case T_IS_STATE:
		if ( current->un.bdvar.name->type == T_IS_META ) {
		    width = current->un.bdvar.hbit -
			    current->un.bdvar.lbit + 1;
/*		    mask = ~( -1 << (current->un.bdvar.hbit + 1) );*/
		    mask = maskbits(current->un.bdvar.hbit);
		    temp = current->un.bdvar.name->VALUE;
		    temp &= mask;
		    temp >>= (unsigned) current->un.bdvar.lbit;
		    current->type = T_IS_CONSTANT;
		    current->un.constant.num = temp;
		    current->un.constant.dont_care = 0;
		    current->un.constant.hbit = width - 1;
		}
		break;
    
	    case T_IS_CONSTANT:
	    case T_OTHERWISE:
	    case T_LEAVE:
		break;

	    case T_LABEL:
		meta( current->un.label.stmt );
		break;

	    case T_IF:
		meta( current->un.bdif.expr );
		current->un.bdif.expr = complex( current->un.bdif.expr,
			&constant, &width );
		if ( constant ) {
		    current->un.bdif.expr = node_eval( current->un.bdif.expr );
		    if ( current->un.bdif.expr->un.constant.num & 1 ) {
			NODE_DISPOSE( current->un.bdif.elseb );
			current->un.bdif.elseb = NULL;
		    } else {
			NODE_DISPOSE( current->un.bdif.thenb );
			current->un.bdif.thenb = NULL;
		    }
		}
		meta( current->un.bdif.thenb );
		meta( current->un.bdif.elseb );
		break;

	    case T_FOR:
	    case T_FOR_DOWN:
		loopmeta = current->un.bdfor.var;
		if ( loopmeta == NULL ) {	/* For already processed */
		    meta( current->un.bdfor.stmt );
		} else {
		    if ( (loopmeta->type != T_IS_STATE) ||
			    (loopmeta->un.bdvar.name->type != T_IS_META) )
			error( "Illegal loop index", loopmeta );
		    meta( current->un.bdfor.from );
		    from = current->un.bdfor.from =
			    node_eval( current->un.bdfor.from );
		    meta( current->un.bdfor.to );
		    to = current->un.bdfor.to = node_eval( current->un.bdfor.to );
		    if ( current->un.bdfor.by == NULL ) {
			by = current->un.bdfor.by =
				node_create_fi( T_IS_CONSTANT, current );
			by->un.constant.num = 1;
			by->un.constant.dont_care = 0;
			by->un.constant.hbit = 31;
		    } else {
			meta( current->un.bdfor.by );
			by = current->un.bdfor.by =
				node_eval( current->un.bdfor.by );
		    }
		    if ( (from->type != T_IS_CONSTANT) || (from->conc != NULL) ||
			 (to->type != T_IS_CONSTANT) || (to->conc != NULL) ||
			 (by->type != T_IS_CONSTANT) || (by->conc != NULL) ) {
			error( "Illegal argument to FOR", current );
		    }
		    if ( by->un.constant.num == 0 )
			error( "BY is zero", by );

		    if ( current->type == T_FOR_DOWN ) {
			by->un.constant.num *= -1;
		    }
		    forreplace = forlast = NULL;
		    for ( loopindex = from->un.constant.num;
			    (( by->un.constant.num > 0 ) 
				    ? loopindex <= to->un.constant.num
				    : loopindex >= to->un.constant.num);
			    loopindex += by->un.constant.num ) {
			forblock = node_dup( current->un.bdfor.stmt );
			loopmeta->un.bdvar.name->VALUE = loopindex;
			meta( forblock );
			if ( forlast == NULL ) {
			    forreplace = forlast = forblock;
			} else {
			    forlast->next = forblock;
			}
			while ( forlast->next != NULL ) forlast = forlast->next;
		    }
		    node_dispose( current->un.bdfor.stmt );
		    current->un.bdfor.stmt = forreplace;
		    /* dispose index to flag that FOR expansion has been done */
		    NODE_DISPOSE( current->un.bdfor.var );
		    current->un.bdfor.var = NULL;
		}
		break;
    
	    case T_CASE:
		meta( current->un.bdcase.expr );
		current->un.bdcase.expr =
			complex( current->un.bdcase.expr, &constant, &width );
		if ( constant ) {
		    current->un.bdcase.expr =
			    node_eval( current->un.bdcase.expr );
		    if ( !(current->un.bdif.expr->un.constant.num & 1) ) {
			NODE_DISPOSE( current->un.bdcase.stmt );
			current->un.bdcase.stmt = NULL;
		    }
		}
		meta( current->un.bdcase.stmt );
		break;
    
	    case T_SELECTONE: case T_SELECTALL:
		nextsave = current->next;
		doselect( current );
		for ( nindex = current; nindex->next != NULL;
			nindex = nindex->next ) {
		}
		nindex->next = nextsave;
		break;

	    case T_SELECT:
		otherw( current, 1 );
		if ( current->un.select.expr != NULL ) 
		    ierror( "Bad otherw in meta", root );
		meta( current->un.select.bdcase );
		break;
    
	    default:
		ierror("Unknown type in meta", root );
	}
    }
}
