/*
 * Structure to keep track of nesting levels, and clauses of
 * IF and SELECT
 */
typedef struct block_struct block;
struct block_struct {
    short level;		/* level of nesting */
    short id;			/* unique number for this block */
    block *parent;		/* block that contains this one */
    block *child;		/* first clause of first nested IF or SEL */
    block *nextclause;		/* Next Clause of same IF or SEL */
    block *nextgroup;		/* first clause of next IF or SEL contained
				 * by same block */
};


/*
 * Structure that is unique to every revision of every bit of every
 * variable
 */
typedef struct usage_struct usage;
struct usage_struct {
    block *block;		/* block of revision */
    int rev;			/* which revision in block */
    usage *prev;		/* usage of var before this one */
    node *assign;		/* the assignment to this usage */
};


/*
 * Structure to hold info about the latest revision of a variable
 */
typedef struct revision_struct revision;
struct revision_struct {
    usage *recent;		/* most recent revision to var */
    struct assigned {
	unsigned previous : 1;	/* assigned in previous routine */
	unsigned current : 1;	/* assigned in this routine */
	unsigned unspec : 1;	/* unspec of this bit used */
    } assigned;
    node *last;			/* last use of variable in seq order */
};


/*
 * Structure to keep track of past revisions to variables
 */
typedef struct revision_list revinfo;
struct revision_list {
    revinfo *next;		/* next variable that was changed */
    symbol *name;		/* which variable am I */
    int bit;			/* and what bit? */
    usage *info;		/* Last change to me in this block */
};

node *subst_version();
revision *rev_array();
extern block *rootblock;
