#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"
#include "mult.h"

block *rootblock;

/*
 * Allocate and initialize new block
 */
block *alloc_block(parent)
block *parent;
{
    block *newblock;
    static short id = 0;

    newblock = ALLOC(block, 1);
    newblock->id = id++;
    if (parent == NIL(block)) {
	newblock->level = 0;
    } else {
	newblock->level = parent->level + 1;
    }
    newblock->nextclause = NIL(block);
    newblock->nextgroup = NIL(block);
    newblock->child = NIL(block);
    newblock->parent = parent;
    return newblock;
}


/*
 * This will scan the tree and record the last use of a variable.
 * Also this will create a hierarchy of blocks that represent the
 * bodies of IF and SELECT clauses.
 */
premult( root, currblock )
node *root;
block *currblock;		/* block for this clause- NIL if top */
{
    symbol *varinfo;		/* symtab entry for var */
    revision *vers;		/* array of revisions for a var */
    int bitnum;			/* bit of var we are working on */
    int lodecl;			/* lowest declared bit of var */
    node *exprroot;		/* root of expr on list */
    node *wordroot;		/* root of a conc'ed word */
    node *nindex;		/* node index */
    block *bindex;		/* block index */
    block *lastgroup;		/* last IF or SEL of this block encountered */

    /* Create block root if first call */
    if (currblock == NIL(block)) {
	currblock = rootblock = alloc_block(NIL(block));
    }
    lastgroup = NIL(block);

    for ( exprroot = root; exprroot != NULL; exprroot = exprroot->next ) {
	for ( wordroot=exprroot; wordroot != NULL; wordroot = wordroot->conc ) {
	    switch( wordroot->type ) {

		case T_IS_CONSTANT:
		case T_OTHERWISE:
		    break;

		case T_BUF: case T_NOT: case T_WIDTH:
		case T_UNARY_PLUS: case T_UNARY_MINUS:
		case T_UNARY_AND: case T_UNARY_OR:
		    premult( wordroot->un.unary.expr, currblock);
		    break;

		case T_AND: case T_OR: case T_NAND:
		case T_NOR: case T_XOR: case T_EQV:
		case T_EQL: case T_NEQ: case T_LEQ:
		case T_GEQ: case T_LSS: case T_GTR:
		case T_SXT: case T_ZXT: case T_OXT:
		case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
		case T_SL0: case T_SR0: case T_SL1: case T_SR1:
		case T_SLR: case T_SRR:
		    premult( wordroot->un.binary.lexpr, currblock );
		    premult( wordroot->un.binary.rexpr, currblock );
		    break;

		case T_BIT:
		    premult( wordroot->un.ternary.lexpr, currblock );
		    premult( wordroot->un.ternary.mexpr, currblock );
		    premult( wordroot->un.ternary.rexpr, currblock );
		    break;

		case T_EQ:
		    premult( wordroot->un.assign.rexpr, currblock );
		    break;

		case T_SELECT:
		    for ( nindex = wordroot->un.select.bdcase;
			    nindex != NULL; nindex = nindex->next ) {
			premult( nindex->un.bdcase.expr, currblock );
		    }
		    if (lastgroup == NIL(block)) {
			bindex = currblock->child = alloc_block(currblock);
		    } else {
			bindex = lastgroup->nextgroup = alloc_block(currblock);
		    }
		    lastgroup = bindex;
		    for (nindex = wordroot->un.select.bdcase;
			    nindex != NULL; nindex = nindex->next) {
			premult(nindex->un.bdcase.stmt, bindex);
			if (nindex->next != NIL(node)) {
			    bindex->nextclause = alloc_block(currblock);
			    bindex = bindex->nextclause;
			}
		    }
		    break;

		case T_IF:
		    premult( wordroot->un.bdif.expr, currblock );
		    if (lastgroup == NIL(block)) {
			bindex = currblock->child = alloc_block(currblock);
		    } else {
			bindex = lastgroup->nextgroup = alloc_block(currblock);
		    }
		    lastgroup = bindex;
		    premult( wordroot->un.bdif.thenb, bindex );
		    bindex->nextclause = alloc_block(currblock);
		    bindex = bindex->nextclause;
		    premult( wordroot->un.bdif.elseb, bindex );
		    break;

		case T_IS_STATE:
		    /* get version info (or make it) */
		    varinfo = wordroot->un.bdvar.name;
		    lodecl = varinfo->u.state.lobit;
		    if ( varinfo->u.state.versions == NULL ) {
			vers = varinfo->u.state.versions = rev_array(
				varinfo->u.state.hibit - lodecl + 1 );
		    } else {
			vers = varinfo->u.state.versions;
		    }
		    bitnum = wordroot->un.bdvar.lbit;
		    /* point to this as last use (may be updated later) */
		    vers[bitnum-lodecl].last = wordroot;
		    break;

		default:
		    (void) fprintf(stderr,"%d",wordroot->type);
		    ierror("bad type in premult",wordroot); 
	    }
	}
    }
}
