#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"

node *comparexpr();


/*
 * This routine reduces a SELECTONEs and SELECTALLs to a
 * collection of IFs.
 */
void
doselect( selectnode )
node *selectnode;
{
    node *ifnode;
    node *ifn_place;
    node *current_case;
    void otherw();

    /* reduce the case conditions to boolean expr's and
     * turn otherwise into one big fat IF
     */
    otherw( selectnode, 0 );

    if( selectnode->un.select.bdcase == NULL ) {
	ifnode = NULL;
    } else {

	/* SELECTONE becomes nested IFs */
        if( selectnode->type == T_SELECTONE ) {
	    current_case = selectnode->un.select.bdcase;
	    ifnode = node_create_fi( T_IF,current_case );
	    ifn_place = ifnode;
	    while( current_case != NULL ) {
	        ifn_place->un.bdif.expr = current_case->un.bdcase.expr;
	        current_case->un.bdcase.expr = NULL;
	        ifn_place->un.bdif.thenb = current_case->un.bdcase.stmt;
	        current_case->un.bdcase.stmt = NULL;
	        if ( current_case->next == NULL) {
		    ifn_place->un.bdif.elseb = NULL;
	        } else {
		    /* nest another IF statement */
	            ifn_place->un.bdif.elseb =
			    node_create_fi(T_IF,current_case->next);
	            ifn_place = ifn_place->un.bdif.elseb;
	        }
	        /* move to next case */
	        current_case = current_case->next;
	    }

	} else {
	    /* SELECTALL becomes many single IFs */
	    current_case = selectnode->un.select.bdcase;
	    ifnode = node_create_fi( T_IF,current_case );
	    ifn_place = ifnode;
	    while( current_case != NULL ) {
	        ifn_place->un.bdif.expr = current_case->un.bdcase.expr;
	        current_case->un.bdcase.expr = NULL;
	        ifn_place->un.bdif.thenb = current_case->un.bdcase.stmt;
	        current_case->un.bdcase.stmt = NULL;
		ifn_place->un.bdif.elseb = NULL;
	        if ( current_case->next != NULL) {
		    /* make another IF statement */
	            ifn_place->next =
			    node_create_fi(T_IF,current_case->next);
	            ifn_place = ifn_place->next;
	        }
	        /* move to next case */
	        current_case = current_case->next;
	    }
	}

	/* now throw out selectnode's children link on IF block */
	/* turn selectnode into a no-op. */
	node_dispose( selectnode->un.select.bdcase );
	selectnode->next = ifnode;
	selectnode->type = T_IS_CONSTANT;
	selectnode->un.constant.num = selectnode->un.constant.hbit = 0;
	selectnode->un.constant.dont_care = 0;
    }
}


/*
 * This routine has two main purposes:
 * The first is to take each case list ("   [1,5,2]:   ")
 * and turn it into a single booloean expression.
 * The second is to generate the the the condition for the
 * otherwise clause of the case.
 * Two otherwises generate an error.  The otherwise is taken
 * to be at the end of the case regardless of its original
 * position.  (A warning is gernerated.)
 */
void
otherw( selectnode, mkotherw )
node *selectnode;
int mkotherw;		/* if true add otherwise clause if missing */
{
    node *cclist;		/* concat list of all compares of a case */
    node *ccl_place;		/* most recent addition to cclist */
    node *allist;		/* concat list of all compares of select */
    node *all_place;		/* most recently added to allist */
    node *current_case;		/* current case of scan */
    node *last_case;		/* predecessor of current case */
    node *current_expr;		/* expression of case */
    node *next_expr;		/* next expr to process */
    node *notnode;		/* temporaries for adding final otherwise expr*/
    node *temp;
    node *otherw_case;		/* case node for the otherwise */
    int otherw_count;		/* How many otherwises */

    /* return if processing has been done */
    if ( selectnode->un.select.expr == NULL ) {
	return;
    }
    otherw_count = 0;
    current_case = selectnode->un.select.bdcase;
    last_case = NULL;
    allist = all_place = NULL;

    /*go through each case block in the select */
    while( current_case != NULL ) {

	if (otherw_count != 0 )
	    warning("Otherwise pushed to the end of select",selectnode);

	/* cut off old expression list so it may be redone */
	current_expr = current_case->un.bdcase.expr;
	current_case->un.bdcase.expr = NULL;

	/* initialize a part of new expr list */
        cclist = ccl_place = NULL;

	/* go through expr's on case list */
	while( current_expr != NULL ) {

	    /* disconnect this expr form others */
	    next_expr = current_expr->next;
	    current_expr->next = NULL;

	    if( current_expr->type == T_OTHERWISE ) {
		otherw_count ++;
		if ( otherw_count > 1 )
		    error("multiple otherwise",current_expr);

		/* save pointer to otherw case */
		otherw_case = current_case;
		/* break out of while loop */
		NODE_DISPOSE( next_expr );
		next_expr = NULL;
		NODE_DISPOSE( current_expr );
		/* throw out the cclist that we were making */
		node_dispose ( cclist );
		cclist = ccl_place = NULL;
	    } else {

		/* Build a concat list of compares */
		if (cclist == NULL) {
		    cclist = ccl_place =
			comparexpr( node_dup( selectnode->un.select.expr),
				current_expr );
		} else {
		    ccl_place->conc =
			comparexpr( node_dup( selectnode->un.select.expr),
				current_expr );
		    ccl_place = ccl_place->conc;
		}
	    }

	    /* go to next expr */
	    current_expr = next_expr;
	}

	if ( cclist == NULL ) {
	    if ( current_case == otherw_case ) {
		/* deal with othew_case later */
		last_case = current_case;
		current_case = current_case->next;
	    } else {
		/* throw out case and move on to next */
		if( last_case == NULL ) {
		    selectnode->un.select.bdcase = current_case->next;
		} else {
		    last_case->next = current_case->next;
		}
		temp = current_case->next;
		current_case->next = NULL;
		node_dispose ( current_case );
		current_case = temp;
	    }

	} else {

	    /* Insert new representation of expr list.
	     * All compares are explicitly tied together by
	     * one big T_UNARY_OR */
	    current_case->un.bdcase.expr =
		    node_create_fi( T_UNARY_OR, current_case );
	    current_case->un.bdcase.expr->un.unary.expr = cclist;
	    current_case->un.bdcase.expr =
		    node_eval( current_case->un.bdcase.expr );
	    /* now make a concat list of the exprs from all cases */
	    if ( allist == NULL )
		allist = all_place = node_dup(current_case->un.bdcase.expr );
	    else {
		all_place->conc = node_dup( current_case->un.bdcase.expr );
		all_place = all_place->conc;
	    }
            last_case = current_case;
	    current_case = current_case->next;
	}
    }

    /* Replace condition for otherwise */
    if ( (otherw_count == 0) && !mkotherw ) {
	node_dispose(allist);
    } else {
	if ( otherw_count == 0 ) {
	    for ( current_case = selectnode->un.select.bdcase;
		    current_case->next != NULL;
		    current_case = current_case->next ) {
	    }
	    otherw_case = node_create_fi( T_CASE, selectnode );
	    current_case->next = otherw_case;
	}
	NODE_DISPOSE( otherw_case->un.bdcase.expr );
	notnode = node_create_fi( T_NOT, otherw_case );
	otherw_case->un.bdcase.expr = notnode;
	notnode->un.unary.expr = node_create_fi( T_UNARY_OR, otherw_case );
	notnode->un.unary.expr->un.unary.expr = allist;
	otherw_case->un.bdcase.expr =
		node_eval( otherw_case->un.bdcase.expr );
    }
    /* take out select.expr, signifying this routine processing has been done */
    NODE_DISPOSE( selectnode->un.select.expr );
    selectnode->un.select.expr = NULL;
}


/*
 * This function returns a EQL node tying together its two arguments
 */
node *comparexpr( left,right )
node *left,*right;
{
    node *temp;

    temp = node_create_fi( T_EQL, right );
    temp->un.binary.lexpr = left;
    temp->un.binary.rexpr = right;
    return( temp );
}
