#define symtab_h

#define GLOBAL_SYMTAB_SIZE	2499
#define LOCAL_SYMTAB_SIZE	2499

typedef struct symbol_table_struct symbol_table;
typedef struct symbol_struct symbol;
typedef struct node_struct node;

struct symbol_struct {
    int type;				/* node type for the object */
    char *name;
    union symbol_union {

	/* T_IS_INPUT_PORT, T_IS_OUTPUT_PORT, T_IS_STATE, or T_IS_META
	 *  
         *  By definition, T_IS_INPUT_PORT and T_IS_OUTPUT_PORT are
         *  global symbols (i.e., is_local == 0).
         *
         *  local_def has meaning if (type == T_IS_STATE && is_local == 1); 
	 *  it is a pointer into the global symbol table which is the mapping 
	 *  for the local symbol.  This is used by inline()
	 */
	struct {
		int nobit;			/* no declared bit number */
	    int lobit, hibit;		/* declared lo and hi bit numbers */
	    int is_local;		/* is a local symbol */
	    int is_temp;		/* is a temporary state */
	    symbol *local_def;		/* symbol to replace local symbol */
	    struct revision_struct *versions;
	} state;

	/* T_IS_SYNONYM */
	struct {
	    int lobit, hibit;		/* declared lo and hi bit numbers */
	    symbol *syn;		/* pointer to synonym entry */
	} synonym;

	/* T_IS_CONSTANT */
	struct {
	    int const_num;		/* declared constant value */
	    int const_hbit;		/* declared constant width (MSB is 0) */
	} constant;

	/* T_IS_MACRO */
	struct {
	    char *value;		/* string replacement macro */
	    int starting_line;
	} macro;

	/* T_IS_ROUTINE */
	struct {
	    node *body;			/* routine definition */
	    symbol_table *locals;	/* local variables and arguments */
	    node *formal_args;		/* linked list of arguments */
	    int hbit, lbit;
	    int active;			/* 1 if currently "active" */
	    int call_count;		/* counts the calls */
	} routine;

	/* T_LABEL */
	struct {
	    node *def;
	} label;

    } u;
    struct symbol_struct *next;
};


struct symbol_table_struct {
    int size;
    struct symbol_struct **data;
};

extern symbol_table *global_syms, *local_syms, *reserved_syms;

#define VALUE u.state.lobit		/* hack, reuse for meta var */

symbol *symtab_lookup();
symbol_table *symtab_init();

void symtab_discard();
void symtab_info();
