/*
 * Symbolic encoding program for compiling a symbolic
 * description into a binary representation.  Target
 * is multi-level logic synthesis
 *
 * History:
 *
 * Bill Lin
 * University of California, Berkeley
 * Comments to billlin@ic.Berkeley.EDU
 *
 * Copyright (c) 1989 Bill Lin, UC Berkeley CAD Group
 *     Permission is granted to do anything with this
 *     code except sell it or remove this message.
 */

#include "jedi.h"
#include "options.h"

int parse_options();	/* forward declaration */

extern FILE *infp;			/* jedi.c */
extern FILE *outfp;			/* jedi.c */

optionStruct optionList[] = {
    { OPT_DESC, 0, "\n" },
    { "h", "",  "print help message" },
    { "e", "option", "specify encoding option" },
    { OPT_CONT, 0,  "\n    r:  random encoding" },
    { OPT_CONT, 0,  "\n    h:  one hot encoding" },
    { OPT_CONT, 0,  "\n    d:  dynamic random encoding" },
    { OPT_CONT, 0,  "\n    s:  straightforward mapping" },
    { OPT_CONT, 0,  "\n    i:  input dominant algorithm" },
    { OPT_CONT, 0,  "\n    o:  output dominant algorithm" },
    { OPT_CONT, 0,  "\n    c:  coupled dominant algorithm" },
    { OPT_CONT, 0,  "\n    y:  modified output dominant algorithm" },
    { "s", "length", "code lenght for state assignment" },
    { "x", "",  "expand state code" },
    { "c", "",  "use group based embedding algorithm" },
    { "p", "",  "output in PLA format" },
    { "g", "",  "general symbolic encoding format" },
    { OPT_OARG, "input", "file to be processed (default stdin)" },
    { OPT_DESC, 0, "\n" },
    { OPT_DESC, 0, HEADER },
    { OPT_DESC, 0, "\n" },
    { 0, 0, 0 }
};

parse_options(argc, argv)
int argc;
char **argv;
{
    extern int optind;
    extern char *optarg;
    int c;

    /* 
     * set defaults 
     */
    beginningStates = 1;
    endingStates = 10;
    startingTemperature = 1000;
    maximumTemperature = INFINITY;

    bitsFlag = FALSE;
    addDontCareFlag = TRUE;
    kissFlag = TRUE;
    verboseFlag = FALSE;
    sequentialFlag = FALSE;
    clusterFlag = FALSE;
    srandomFlag = FALSE;
    drandomFlag = FALSE;
    variationFlag = TRUE;
    oneplaneFlag = FALSE;
    hotFlag = FALSE;
    expandFlag = FALSE;
    plaFlag = FALSE;

    weightType = OUTPUT;

    /*
     * parse options
     */
    while ((c=optGetOption(argc, argv)) != EOF) {
	switch(c) {
	case 'h':
	    optUsage();
	    break;
	case 'e':
	    if (!strcmp(optarg, "s")) {
		sequentialFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "r")) {
		srandomFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "d")) {
		drandomFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "h")) {
		hotFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "i")) {
		weightType = INPUT;
		variationFlag = FALSE;
		break;
	    } else if (!strcmp(optarg, "o")) {
		weightType = OUTPUT;
		variationFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "c")) {
		weightType = COUPLED;
		variationFlag = TRUE;
		break;
	    } else if (!strcmp(optarg, "y")) {
		weightType = OUTPUT;
		variationFlag = FALSE;
		break;
	    } else {
		optUsage();
		break;
	    }
	case 's':
	    bitsFlag = TRUE;
	    code_length = atoi(optarg);
	    break;
	case 'x':
	    expandFlag = TRUE;
	    break;
	case 'c':
	    clusterFlag = TRUE;
	    break;
	case 'p':
	    plaFlag = TRUE;
	    break;
	case 'g':
	    kissFlag = FALSE;
	    break;
	default:
	    optUsage();
	}
    }

    /*
     * check for sufficient arguments
     */
    if (optind != argc && optind != argc-1) {
	optUsage();
    }

    if (optind == argc) {
	infp = stdin;
    } else {
	infp = my_fopen(argv[argc-1], "r");
    }
    outfp = stdout;
} /* end of parse_options */
