#define WORD_LENGTH 32
#define ALL_ZEROS   0
#define ALL_ONES    ~0

#define EXTRACT_BIT(word,pos) (((word) & (1 << (pos))) != 0)
#define ATPG_SLOT simulation
#define ATPG_GET(node) (int) node->ATPG_SLOT
#define ATPG_SET(node, val) node->ATPG_SLOT = (char *) val

typedef void (*VoidFN)();

typedef struct sim_node_struct sim_node_t;
struct sim_node_struct {
  node_t *node;
  int    id;
  int    n_cubes;
  int    n_inputs;
  int    **function;    /* function[cube][input] is 0, 1 or 2 */
  int    value;		/* output value of node */
  int    or_output_mask;
  int    and_output_mask;
  int    *input_uids;
  int    **inputs;	/* inputs[0] is neg inputs; inputs[1] is inputs, 
			   inputs[2] is always 0xffffffff */
  int    *or_input_masks;
  int    *and_input_masks;
  VoidFN eval;
};

typedef struct saf_struct saf_t;
struct saf_struct {
  node_t *node;
  node_t *fanin;
  int    node_id;	/* uid used in simulation */
  int    value;		/* 0 or 1 */
  int    input;		/* which input is faulty; if -1, the output is faulty */
  int    covered;	/* initially 0; set to 1 if covered by a pattern */
  int    *pattern;	/* which pattern covers it */
  int    bit;           /* need by rtg to derive pattern */
  int    redundant;	/* initially 0; set to 1 if fault is redundant */
};

typedef struct {
    node_t *node;
    int    *input;
} critical_rec_t;
 
/* satisfy.c */
extern int  satisfy();

/* com_atpg.c */
extern void free_pattern();
extern void concat_lists();

/* faul_sim.c */
extern lsList rtp_generate();
extern lsList fsim_vector();
extern void   print_fault();

/* sim_extract.c */
extern sim_node_t *sim_extract_network();
extern array_t    *network_dfs_complete();
extern array_t    *create_fanout_list();
extern void       sim_free_nodes();

/* fault_list.c */
extern lsList gen_fault_list();
extern void   print_fault_list();

/* interpret.c */
extern void fsim_network();
extern void fast_fsim_network();
extern void gen_random_pattern();
extern void clear_tfo();
extern void print_array2();

/* fault_extract.c */
extern void fault_extract();
extern void add_fault_clauses();
extern void clear_node_table();

/* non_local_imp.c */
extern void dominators();
extern void nl_impl();
extern void nl_bad_clauses();
extern void app_all_doms();

/* extract.c */
extern void extract_and2();
extern void extract_or2();
extern void add_clause_1();
extern void add_clause_2();
extern void add_clause_3();
extern void check_node();
extern void imply_node();
extern void dump_node_bindings();
extern int  extract();
extern int  *copy_array();
extern int  *extract_imp_vars();
extern int  get_variable();
extern int  next_clause();
extern void dump_clauses();
