#include "port.h"
#include "satisfy.h"

#define USE_DYN_EQ  0
#define USE_DYN_IM  0

/* sat.c: example of how to call sat_solve with different search
 * strategies. */

static int 
sat_weight1 (var_info,which_phase)
sat_var_info *var_info;
int *which_phase;
{
    /* Larrabee strategy 1 (p. 43 of thesis): */
    *which_phase = 1;
    return var_info->pos.n_implies_0 + var_info->neg.n_implies_0;
}

static int 
sat_weight2 (var_info,which_phase)
sat_var_info *var_info;
int *which_phase;
{
    /* Larrabee strategy 2 (p. 43 of thesis): */
    *which_phase = 0;
    return var_info->pos.n_implies_0 + var_info->neg.n_implies_0;
}
    
static int 
sat_weight3 (var_info,which_phase)
sat_var_info *var_info;
int *which_phase;
{
    int weight = 0;

    /* Larrabee strategy 3 (p. 44 of thesis): */
    *which_phase = 1;
    weight += (var_info->pos.n_implies_0) << 6;
    weight += (var_info->pos.n_satisfies) + (var_info->neg.n_satisfies);
    return weight;
}

static int 
sat_weight4 (var_info,which_phase)
sat_var_info *var_info;
int *which_phase;
{
    int weight = 0;
    /* Herve' phase ordering */
    *which_phase = (var_info->pos.n_trans_implies >
            var_info->neg.n_trans_implies) ? 1 : 0;
    weight += (var_info->pos.n_implies_0) << 6;
    weight += (var_info->pos.n_satisfies) + (var_info->neg.n_satisfies);
    return weight;
}

/* 0=cannot be satisfied, 1=satisfiable */
int 
satisfy (clauses, n, vars, values, n_vars, flag)
int *clauses;       /*i binary cnf in format described above    */
unsigned n;         /*i length of array clauses         */
int *vars;          /*i variables to return values for      */
int *values;        /*o satisfying assignment if result is 1    */
unsigned n_vars;    /*i number of entries in vars & values      */
int flag;           /*i debugging flag, 0=no debugging output   */
{
    int rc, max_v, *p;
    unsigned i;
    sat_options  options;
    sat_pass     passes[5];
    sat_pass    *pass;

    /* Figure out the highest variable. */
    for (max_v=0,p=clauses,i=n; i--; p++)
        if (*p > max_v) max_v = *p;

    options.trace_flag = flag;
    options.out_file   = "saved_clauses";
    options.append_out = 0;
    options.n_init_nli = 1;
    options.n_pass     = 4;
    options.passes     = passes;

    pass = options.passes;

    pass->backtrack_lim = max_v;
    pass->weight_fcn  = sat_weight4;
    pass->use_dyn_eq  = USE_DYN_EQ;
    pass->use_dyn_nli = USE_DYN_IM;
    pass++;

    pass->backtrack_lim = max_v;
    pass->weight_fcn  = sat_weight1;
    pass->use_dyn_eq  = USE_DYN_EQ;
    pass->use_dyn_nli = USE_DYN_IM;
    pass++;

    pass->backtrack_lim = max_v;
    pass->weight_fcn  = sat_weight2;
    pass->use_dyn_eq  = USE_DYN_EQ;
    pass->use_dyn_nli = USE_DYN_IM;
    pass++;

    pass->backtrack_lim = max_v;
    pass->weight_fcn  = sat_weight3;
    pass->use_dyn_eq  = USE_DYN_EQ;
    pass->use_dyn_nli = USE_DYN_IM;
    pass++;

    rc = sat_solve (clauses,n,vars,values,n_vars,&options);
    return rc;
}
