#include "mis.h"
#include "atpg.h"
#include "atpg_int.h"
#include "satisfy.h"

extern int  atpg_variable_count;
static int  apply_atpg();

/* ARGSUSED */
int
com_red_removal(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int        result = 0;
    lsList     fault_list, tested_faults;
    st_table   *table;
    sim_node_t *sim_nodes;
    
    if (argc > 1) {
	goto usage;
    }

    decomp_tech_network(*network, INFINITY, INFINITY);
    network_sweep(*network);
    table = st_init_table(st_ptrcmp, st_ptrhash);
    sim_nodes = sim_extract_network(*network, table);
    fault_list = gen_fault_list(*network, table);
    tested_faults = rtp_generate(*network, table, sim_nodes, fault_list, 20, 0);

    /* loop until no redundancy */
    result = apply_atpg(*network, &tested_faults, fault_list, 
			table, sim_nodes, 0);
    while (result == 0) {
	table = st_init_table(st_ptrcmp, st_ptrhash);
	sim_nodes = sim_extract_network(*network, table);
	fault_list = gen_fault_list(*network, table);
        result = apply_atpg(*network, &tested_faults, fault_list, 
			    table, sim_nodes, 1);
    }
    return 0;
usage:
    fprintf(misout, "usage: red_rem\n");
    return 1;
}

/* routine for test generation 
 * flag = 0 indicates test vectors are not simulated first
 * flag = 1 indicates simulate test vectors first 
 */
static int
apply_atpg(network, tested_faults, flist, table, sim_nodes, flag)
network_t       *network;
lsList          *tested_faults;
lsList          flist;
st_table        *table;
sim_node_t      *sim_nodes;
{
    lsGeneric   userData;
    lsHandle    itemHandle;
    lsList      incr_tested, temp_list;
    lsGen       gen;
    array_t     *master_constraint_clauses = array_alloc(int, 0), 
                *constraint_clauses, *non_local_clauses;
    saf_t       *fnode;
    unsigned    n, n_vars;
    int         pi_seen = FALSE, i, result, *values, *marker, *vars, *clauses;
    node_t      *const_node;

    values = ALLOC(int, network_num_pi(network));

    /* first simulate test vectors */
    if (flag) {
	temp_list = lsCreate();
	gen = lsStart(*tested_faults);
	while (lsNext(gen, (lsGeneric*)&fnode, &itemHandle) == LS_OK) {
	    for (i = 0; i < network_num_pi(network); i ++) {
		 values[i] = ((fnode->pattern[i] == 0)? 0 : ~0);
	    }
	    incr_tested = fsim_vector(network, table, sim_nodes, flist, values);
	    concat_lists(temp_list, incr_tested);
	}
	lsFinish(gen);
	lsDestroy(*tested_faults, free_pattern);
        *tested_faults = temp_list;
    }

    atpg_variable_count = 1;
    result = 1;
    if (extract(master_constraint_clauses, network) != 1) {
        array_free(master_constraint_clauses);
        return 1;
    }

    /* Do Non-local implications */
    non_local_clauses = array_alloc(int, 0);
    nl_impl(network, non_local_clauses);

    /* test pattern generation loop */
    while (lsFirstItem(flist, (lsGeneric*)&fnode, &itemHandle) == LS_OK) {
        constraint_clauses = array_dup(master_constraint_clauses);
        fault_extract(network, constraint_clauses, fnode, 1,
                      &marker, 1, 1);
	array_append(constraint_clauses, non_local_clauses);
	nl_bad_clauses(constraint_clauses, non_local_clauses, marker);
        FREE(marker);

        vars = extract_imp_vars(network, &n_vars);
        n = array_n(constraint_clauses);
        clauses = copy_array(constraint_clauses);
        array_free(constraint_clauses);
        result = satisfy(clauses, n, vars, values, n_vars, 0);
        FREE(vars);
        FREE(clauses);

        switch (result) {
	case 0: /* redundant fault */
	    const_node = node_constant(fnode->value);
	    if (node_function(fnode->node) != NODE_PI) {
		if ((fnode->fanin) == NULL) {
		    node_replace(fnode->node, const_node);
		}
		else {
		    if (node_num_fanout(fnode->fanin) == 1) {
			node_replace(fnode->fanin, const_node);
		    }
		    else {
			network_add_node(network, const_node);
			node_patch_fanin(fnode->node, fnode->fanin, const_node);
		    }
		}
	    }
	    else {
		assert(fnode->fanin == NULL);
		while (node_num_fanout(fnode->node) != 0) {
		    network_add_node(network, const_node);
		    node_patch_fanin(node_get_fanout(fnode->node, 0), 
				     fnode->node, const_node);
		}
		pi_seen = TRUE;
	    }
	    network_sweep(network);
	    lsRemoveItem(itemHandle, &userData); 
	    break;
	case 2: /* aborted fault */
	    lsRemoveItem(itemHandle, &userData); 
	    break;
	case 1:
	    for (i = 0; i < network_num_pi(network); i++) {
		if (values[i] == 1) {
		    values[i] = ~0;
		}
		else {
		    values[i] = 0;
		}
	    }
	    incr_tested = fsim_vector(network, table, sim_nodes, flist, values);
	    concat_lists(*tested_faults, incr_tested);
	    break;
	default:
	    break;
	}
	if (result == 0 && !pi_seen) {
	    break;
	}
    }
    
    sim_free_nodes(sim_nodes, table);
    st_free_table(table);
    array_free(master_constraint_clauses);    
    array_free(non_local_clauses);
    lsDestroy(flist, free);
    clear_node_table();
    FREE(values);

    return result; 
}
