#include "mis.h"
#include "atpg_int.h"
#include <math.h> 

void add_clause_1(array, var1)
array_t *array;
int var1;
{
  array_insert_last(int, array, var1);
  array_insert_last(int, array, 0);
}

/* This routine adds critical clauses as defined by Larrabee. Critical Table
 * contains information that is used to make sure that a fanin of a gate in the 
 * active path is
 * not in the transitive fanout of the fault site. */
void 
add_critical_clauses(array, critical_table)
array_t *array;
st_table *critical_table;
{
    st_generator *gen;
    node_t *node;
    critical_rec_t *ptr;
    int ct, ct2;
    node_t *fanin;

    st_foreach_item(critical_table, gen, (char **) &node, (char **) &ptr) {
    node = ptr->node;
    switch(node_function(node)) {
        case NODE_AND:
            for (ct=0; ct < node_num_fanin(node); ct++) {
                if (ptr->input[ct] != 0) {
                    for (ct2 = 0; ct2 < node_num_fanin(node) ; ct2++) {
                        if ((ct != ct2) && (ptr->input[ct2] == 0)) {
                            fanin = node_get_fanin(node, ct2);
                            switch(node_input_phase(node, fanin)) {
                                case POS_UNATE:
                                    add_clause_2(array, -(ptr->input[ct]), get_variable(fanin));
                                    break;
                                case NEG_UNATE:
                                    add_clause_2(array, -(ptr->input[ct]), -get_variable(fanin));
                                    break;
                            }
                        }
                    }
                }
            }
            break;
        case NODE_OR:
            for (ct=0; ct < node_num_fanin(node); ct++) {
                if (ptr->input[ct] != 0) {
                    for (ct2 = 0; ct2 < node_num_fanin(node) ; ct2++) {
                        if ((ct != ct2) && (ptr->input[ct2] == 0)) {
                            fanin = node_get_fanin(node, ct2);
                            switch(node_input_phase(node, fanin)) {
                                case POS_UNATE:
                                    add_clause_2(array, -(ptr->input[ct]), -get_variable(fanin));
                                    break;
                                case NEG_UNATE:
                                    add_clause_2(array, -(ptr->input[ct]), get_variable(fanin));
                                    break;
                            }
                        }
                    }
                }
            }
            break;
    }
    }
}
        
/* This routine adds special critical clauses that involves fanins that are
in the transitive fanout of the fault site. Further details are discussed in 
the 290ls report. */
void 
add_imp1_clauses(array, critical_table, total_nodes)
array_t *array;
st_table *critical_table;
int total_nodes;
{
    st_generator    *gen;
    node_t            *node;
    critical_rec_t    *ptr;
    int                ct, ct2;
    node_t            *fanin, *fanin1;

    st_foreach_item(critical_table, gen, (char **) &node, (char **) &ptr) {
        node = ptr->node;
        switch(node_function(node)) {
            case NODE_AND:
                for (ct=0; ct < node_num_fanin(node); ct++) {
                    if (ptr->input[ct] != 0) {
                        for (ct2 = 0; ct2 < node_num_fanin(node) ; ct2++) {
                            if ((ct != ct2) && (ptr->input[ct2] != 0)) {
                                fanin = node_get_fanin(node, ct2);
                                fanin1 = node_get_fanin(node, ct);
                                switch(node_input_phase(node, fanin)) {
                                    case POS_UNATE:
                                        add_clause_3(array, -(ptr->input[ct]), -get_variable(fanin1), get_variable(fanin));
                                        add_clause_3(array, -(ptr->input[ct]), -(total_nodes + get_variable(fanin1)), total_nodes + get_variable(fanin));
                                        break;
                                    case NEG_UNATE:
                                        add_clause_3(array, -(ptr->input[ct]), -get_variable(fanin1), -get_variable(fanin));
                                        add_clause_3(array, -(ptr->input[ct]), -(total_nodes + get_variable(fanin1)), -(total_nodes + get_variable(fanin)));
                                        break;
                                }
                            }
                        }
                    }
                }
                break;
            case NODE_OR:
                for (ct=0; ct < node_num_fanin(node); ct++) {
                    if (ptr->input[ct] != 0) {
                        for (ct2 = 0; ct2 < node_num_fanin(node) ; ct2++) {
                            if ((ct != ct2) && (ptr->input[ct2] != 0)) {
                                fanin = node_get_fanin(node, ct2);
                                fanin1 = node_get_fanin(node, ct);
                                switch(node_input_phase(node, fanin)) {
                                    case POS_UNATE:
                                        add_clause_3(array, -(ptr->input[ct]), get_variable(fanin1), -get_variable(fanin));
                                        add_clause_3(array, -(ptr->input[ct]), (total_nodes + get_variable(fanin1)), -(total_nodes + get_variable(fanin)));
                                        break;
                                    case NEG_UNATE:
                                        add_clause_3(array, -(ptr->input[ct]), get_variable(fanin1), get_variable(fanin));
                                        add_clause_3(array, -(ptr->input[ct]), (total_nodes + get_variable(fanin1)), (total_nodes + get_variable(fanin)));
                                        break;
                                }
                            }
                        }
                    }
                }
                break;
        }
        free(ptr->input);
        free(ptr);  
    }
}

/* Used to mark the transitive fanout of the circuit. This routine is 
not used if active clauses are to be added to clauses to be satisfied. */
void 
mark_transitive_fanout(marker, node, total_nodes)
int marker[];
node_t *node;
int total_nodes;
{
    node_t *fanout;
    lsGen gen;
    int var;

    var = get_variable(node);
    if (marker[var] == 0) {
        marker[var] = total_nodes + var;
        foreach_fanout(node, gen, fanout) {
            mark_transitive_fanout(marker, fanout, total_nodes);
        }
    }
}

void 
update_critical_table(critical_table, act, source_node, node)
st_table *critical_table;
int act;
node_t *source_node, *node;
{
    critical_rec_t *ptr;
    int i;

    if (st_lookup(critical_table, (char *)node, (char **) &ptr) == 1)
        ptr->input[node_get_fanin_index(node, source_node)] = act;
    else {
        ptr = (critical_rec_t *) malloc(sizeof(critical_rec_t));
        ptr->node = node;
        ptr->input = (int *) calloc(node_num_fanin(node), sizeof(int));
        for (i= 0 ; i < node_num_fanin(node) ; i++)
            ptr->input[i] = 0;
        ptr->input[node_get_fanin_index(node, source_node)] = act;
        st_insert(critical_table, (char *) node, (char *) ptr);
    }
}

void 
mark_transitive_fanout_and_add_active_clauses(array, tail, ct, marker, node, total_nodes, critical_table, critical)
array_t *array;
int tail;
int *ct;
int marker[];
node_t *node;
int total_nodes;
st_table *critical_table;
int critical;
{
    node_t *fanout;
    lsGen gen;
    int var, act;
    int ct1, start;

    var = get_variable(node);
    act = 3 * total_nodes + var;
    if (tail != 0) {
        add_clause_2(array, -tail, act);
    }
    if (marker[var] == 0) {
        marker[var] = total_nodes + var;
        add_clause_3(array, -act, var, marker[var]);
        add_clause_3(array, -act, -var, -marker[var]);
        switch(node_num_fanout(node)) {
            case 0:
                break;
            case 1:
                if (critical == 1) {
                    update_critical_table(critical_table, act, node, node_get_fanout(node, 0));
                }
                mark_transitive_fanout_and_add_active_clauses(array, act, ct, marker, node_get_fanout(node, 0), total_nodes, critical_table, critical);
                break;
            default:
                array_insert_last(int, array, -act);
                start = *ct;
                for (ct1=1; ct1<=node_num_fanout(node); ct1++) {
                    *ct = *ct + 1;
                    array_insert_last(int, array, *ct);
                }
                array_insert_last(int, array, 0);
                ct1 = 0;
                foreach_fanout(node, gen, fanout) {
                    ct1++;
                    if (critical == 1) {
                        update_critical_table(critical_table, start+ct1, node, fanout);
                    }
                    add_clause_2(array, -(start+ct1), act);
                    mark_transitive_fanout_and_add_active_clauses(array, start+ct1, ct, marker, fanout, total_nodes, critical_table, critical);
                }
        }
    }
}

void 
clear_marker(marker, total)
int marker[];
int total;
{
    int i;

    for (i=0; i<=total; i++)
        marker[i] = 0;
}

void 
print_marker(marker, total_nodes)
int *marker;
int total_nodes;
{
    int i;

    for (i=0 ; i <= total_nodes; i++)
        fprintf(misout, "%d ", marker[i]);
    fprintf(misout, "\n");
}


int 
add_special_case(fnode, out, in)
saf_t *fnode;
int *out, *in;
{

    if (fnode->fanin == NULL) {
        *out = 0;
        *in = 0;
    }
    else {
        *in = get_variable(fnode->fanin); 
        *out = get_variable(fnode->node);
    }
} 

int 
add_special_case_and_special_clause(array, ct, total_nodes, fnode, out, in, critical_table, critical)
array_t *array;
int *ct;
int total_nodes;
saf_t *fnode;
int *out, *in;
st_table *critical_table;
int critical;
{
    int act;

    if (fnode->fanin == NULL) {
        *out = 0;
        *in = 0;
        add_clause_1(array, 3*total_nodes + get_variable(fnode->node));
    }
    else {
        *in = get_variable(fnode->fanin); 
        *out = get_variable(fnode->node);
        act = 3*total_nodes + *out;
        *ct = *ct + 1;
        add_clause_2(array, -(*ct), act);
        add_clause_3(array, -(*ct), *in, total_nodes+*in);
        add_clause_3(array, -(*ct), -(*in), -(total_nodes+*in));
        add_clause_1(array, *ct);
        if (critical == 1) {
            update_critical_table(critical_table, act, fnode->fanin, fnode->node);
        }
    }
}

void 
add_and_clauses(array, and1, and2, out, out_p) 
array_t *array;
int out, out_p, and1, and2;
{
    add_clause_2(array, - and1, out);
    add_clause_2(array, - and1, - out_p);
    add_clause_3(array, and1, - out,  out_p);
    add_clause_2(array, -and2, - out);
    add_clause_2(array, - and2, out_p);
    add_clause_3(array, and2,  out, - out_p);
}

void 
add_or_clauses(array, total_nodes, ct)
array_t *array;
int total_nodes, ct;
{
    int i;

    for(i = (2 * total_nodes + 1); i <= (2 * total_nodes + ct); i++) {
          add_clause_2(array, 2*total_nodes+ct+1, -i);
    } 
    array_insert_last(int, array, -(2*total_nodes+ct+1));
    for(i = (2 * total_nodes + 1); i <= (2 * total_nodes + ct); i++) {
          array_insert_last(int, array, i);
    } 
    array_insert_last(int, array, 0);
    array_insert_last(int, array, (2*total_nodes+ct+1));
    array_insert_last(int, array, 0);
}

void 
add_xor_clauses(network, array, marker, total_nodes)
network_t *network;
array_t *array;
int *marker;
int total_nodes;
{
    int ct;
    lsGen gen;
    node_t *po;
    int out, out_p;
 
    ct = 1;
    foreach_primary_output(network, gen, po) {
        out = get_variable(po);
        out_p = marker[out];
        if (out_p != 0) {
            add_and_clauses(array, 2 * total_nodes + ct, 2 * total_nodes + ct + 1, out, out_p);
            ct = ct + 2;
        }
    }
    add_or_clauses(array, total_nodes, ct - 1);
}
        
void 
add_stuck_clause(array, fnode, total_nodes)
array_t *array;
saf_t *fnode;
int total_nodes;
{
    int var;

    if (fnode->fanin == NULL) {
        var = get_variable(fnode->node);
    }
    else {
        var = get_variable(fnode->fanin);
    }
    if (fnode->value == 0) {
        add_clause_1(array, var);
        add_clause_1(array, - (total_nodes + var));
    }
    else {
        add_clause_1(array, -var);
        add_clause_1(array,  (total_nodes + var));
    }
}
    
void
fault_extract(network, constraint_clauses, fnode, active_clauses, marker, critical, imp1)
network_t *network;
array_t *constraint_clauses;
saf_t *fnode;
int active_clauses;
int **marker;
int critical, imp1;
{
    int total_nodes;
    array_t *active_constraint_clauses;
    node_t *node;
    int out, in;
    int ct=0;
    st_table *critical_table;
   
    total_nodes = network_num_pi(network) + network_num_internal(network) + network_num_po(network);
    *marker = (int *) calloc(total_nodes+1, sizeof(int));
    clear_marker(*marker, total_nodes);
    node = fnode->node;
    if (active_clauses == 1) {
        if (critical == 1) {
            critical_table = st_init_table(st_ptrcmp, st_ptrhash);
	}
        active_constraint_clauses = array_alloc(int, 0);
        ct = 4*total_nodes;
        mark_transitive_fanout_and_add_active_clauses(active_constraint_clauses, 0, &ct, *marker, node, total_nodes, critical_table, critical);
        add_special_case_and_special_clause(active_constraint_clauses, &ct, total_nodes, fnode, &out, &in, critical_table, critical);
    }
    else {
        mark_transitive_fanout(*marker, node, total_nodes);
        add_special_case(fnode, &out, &in); 
    }
    add_fault_clauses(constraint_clauses, *marker, total_nodes, out, in);
    add_xor_clauses(network, constraint_clauses, *marker, total_nodes);
    add_stuck_clause(constraint_clauses, fnode, total_nodes);
    if (active_clauses == 1) {
        array_append(constraint_clauses, active_constraint_clauses);
        if (critical == 1) {
            add_critical_clauses(constraint_clauses, critical_table);
            if (imp1 == 1)
                add_imp1_clauses(constraint_clauses, critical_table, total_nodes);
            st_free_table(critical_table);
        }
    }
    array_free(active_constraint_clauses);
}

void 
print_stats(array)
array_t *array;
{
    int ct1, ct2=0, total=0, one=0, two=0, three=0, four=0;

    for(ct1=0;ct1 < array_n(array); ct1++) {
        if (array_fetch(int, array, ct1) == 0) {
            total++;
            if (ct2 == 1) one++;
            if (ct2 == 2) two++;
            if (ct2 == 3) three++;
            if (ct2 == 4) four++;
            ct2 = 0;
        } 
        else
            ct2++;
    }
    fprintf(misout, "%d: 1:%d 2:%d 3:%d 4:%d\n", total, one, two, three, four);
}

int 
check_clause(array, marker, position)
array_t *array;
int *marker;
int position;
{
    int exit, current_variable;

    exit = FALSE;
    current_variable = abs(array_fetch(int, array, position));
    while ((exit == FALSE) && (current_variable != 0)) {
        if (marker[current_variable] != 0)
            exit = TRUE;
        position++;
        current_variable = abs(array_fetch(int, array, position));
    }
    return exit;
}    

void 
copy_clause(array, marker, position, total_nodes, out, in)
array_t *array;
int *marker;
int position, total_nodes, out, in;
{
    int current_variable;
    
    current_variable = array_fetch(int, array, position);
    if (abs(current_variable) == out)
        marker[in] = total_nodes + in;
    while (current_variable != 0) {
        if (current_variable < 0) {
            if (marker[- current_variable] != 0)
                current_variable = - marker[- current_variable];
        }
        else {
            if (marker[current_variable] != 0)
                current_variable =  marker[ current_variable];
        }
        array_insert_last(int, array, current_variable);
        position++;
        current_variable = array_fetch(int, array, position);
    }
    marker[in] = 0;
    array_insert_last(int, array, 0);
}
 
int 
next_clause(array, position)
array_t *array;
int position;
{
    while (array_fetch(int, array, position) != 0)
        position++;
    return(position+1);
}
    
void 
add_fault_clauses(array, marker, total_nodes, out, in)
array_t *array;
int *marker;
int total_nodes;
int out, in;
{
    int position;
    int length;

    position = 0;
    length = array_n(array);
    while (position < length) {
        if (check_clause(array, marker, position) == TRUE) {
            copy_clause(array, marker, position, total_nodes, out, in);
        }
        position = next_clause(array, position);
    }
}
