#include "mis.h"
#include "atpg_int.h"

int nl_count = 0;

void 
nl_impl(network, array)
network_t *network;
array_t *array;
{
    lsGen gen;
    node_t *node;

    /*    printf("Doing non-local implications.\n"); */
    nl_count = 0;
    foreach_node(network, gen, node) {
        if (node_num_fanout(node) > 1) {

/* Just do non-local implications for nodes with fanout greater than one */

            check_node(network, array, node);
        }
    }
    fprintf(misout, "There are %d non_local implications.\n", nl_count);
}


void check_node(network, array, node)
network_t *network;
array_t *array;
node_t *node;
{
    node_t *fanout, *node1;
    lsGen gen;

/* Use ATPG_SLOT to do implications in the network */

    foreach_node(network, gen, node1) {
	ATPG_SET(node1, 2);
    }

/* Do implications with node set to zero */

    ATPG_SET(node, 0);
    foreach_fanout(node, gen, fanout) {
        imply_node(array,  get_variable(node), fanout);
    }
    foreach_node(network, gen, node1) {
	ATPG_SET(node1, 2);
    }

/* Do implications with node set to one */

    ATPG_SET(node, 1);
    foreach_fanout(node, gen, fanout) {
        imply_node(array, - get_variable(node), fanout);
    }
    ATPG_SET(node, 2);
}

void imply_node(array, source_var, node)
array_t *array;
int source_var;
node_t *node;
{
    int j, i;
    int status, satisfy;
    node_t *fanout, *fanin;
    node_literal_t literal;
    node_cube_t cube;
    lsGen gen;

    if (ATPG_GET(node) == 2) {
        status = 0;

/* Go through each cube and see if any of the cubes are satisfied. */

        for (i = node_num_cube(node) - 1; i >= 0; i--) {
            cube = node_get_cube(node, i);
            satisfy = 1;
            foreach_fanin(node, j, fanin) {
                literal = node_get_literal(cube, j);
                switch(literal) {
                case ONE:
                    if (ATPG_GET(fanin) == 0) {
                        satisfy = 0;
                    } 
                    else {
                        if ((ATPG_GET(fanin) == 2) && (satisfy != 0)) {
                            satisfy = 2;
                        }
                    }
                    break;
                case ZERO:
                    if (ATPG_GET(fanin) == 1) {
                        satisfy = 0;
                    } 
                    else {
                        if ((ATPG_GET(fanin) == 2) && (satisfy != 0)){
                            satisfy = 2; 
                        }
                    }
                    break;
                case TWO:
                    break;
                default:
                    fail("bad literal");
                }
            }
            if (satisfy == 1) {
                status = 1;
            } 
            else {
                if ((satisfy == 2) && (status != 1)) {
                    status = 2;
                }
            }
        }
        if (status == 1) {
	    ATPG_SET(node, 1);
            if (node_function(node) == NODE_AND) {
/* Add a non-local implication */
                add_clause_2(array, source_var, get_variable(node));
                nl_count++;
            }
        } 
        else {
            if (status == 0) {
		ATPG_SET(node, 0);
                if (node_function(node) == NODE_OR) {
/* Add a non-local implication */
                    add_clause_2(array, source_var, - get_variable(node)); 
                    nl_count++;
                }
            }
        }
        if (ATPG_GET(node) != 2) {
/* Continue implications */
            foreach_fanout(node, gen, fanout) {
                imply_node(array, source_var, fanout);
            }
        }
    }    
}

int check_clause2(array, marker, position)
array_t *array;
int *marker;
int position;
{
    int exit, current_variable;

    exit = TRUE;
    current_variable = abs(array_fetch(int, array, position));
    while ((exit==TRUE) && (current_variable != 0)) {
        if (marker[current_variable] == 0) {
            exit = FALSE;
        }
        position++;
        current_variable = abs(array_fetch(int, array, position));
    }
    return(exit);
}


void copy_clause2(array, non_local_clauses, marker, position)
array_t *array, *non_local_clauses;
int *marker;
int position;
{
    int current_variable;

    current_variable = array_fetch(int, non_local_clauses, position);
    while (current_variable != 0) {
        if (current_variable < 0) {
            if (marker[- current_variable] != 0) {
                current_variable = - marker[- current_variable];
            }
        } 
        else {
            if (marker[current_variable] != 0) {
                current_variable =  marker[ current_variable];
            }
        }
        array_insert_last(int, array, current_variable);
        position++;
        current_variable = array_fetch(int, non_local_clauses, position);
    }
    array_insert_last(int, array, 0);
}


void 
nl_bad_clauses(array, non_local_clauses, marker)
array_t *array;
array_t *non_local_clauses;
int *marker;
{
    int position;
    int length;

    position = 0;
    length = array_n(non_local_clauses);
    while (position < length) {
        if (check_clause2(non_local_clauses, marker, position) == TRUE){
            copy_clause2(array, non_local_clauses, marker, position);
        }
        position = next_clause(non_local_clauses, position);
    }
} 

int 
verify_dom(current_node, pot_dominator)
node_t *current_node, *pot_dominator;
{
    int result=0, i;

    if (current_node != pot_dominator) {
        if (node_function(current_node) == NODE_PO) {
            result = 1;
        } 
        else {
            i = 0;
            while ((result == 0) && (i < node_num_fanout(current_node))) {
                result = verify_dom(node_get_fanout(current_node, i), pot_dominator);
                i++;
            }
        }
    }
    return(result);
}

void 
recursive_mark(node)
node_t *node;
{
    lsGen gen;
    node_t *fanout;

    if (ATPG_GET(node) == 0) {
        ATPG_SET(node, 1);
    } 
    else {
        if (ATPG_GET(node) == 1) {
            ATPG_SET(node, 2);
        }
    }

    if (ATPG_GET(node) < 2) {
        foreach_fanout(node, gen, fanout) {
            recursive_mark(fanout);
        }
    }
}

void 
app_dom(dominator_table, source_node, node)
st_table *dominator_table;
node_t *source_node, *node;
{
    array_t *array;

    if (st_lookup(dominator_table, (char *) source_node, (char **) &array) == 1) {
        array_insert_last(int, array, (int)node);
    } 
    else {
        array = array_alloc(int, 0);
        array_insert_last(int, array, (int)node);
        st_insert(dominator_table, (char *) source_node, (char *) array);
    }
}

void 
add_dom(network, source_node, dominator_table, ct)
network_t *network;
node_t *source_node;
st_table *dominator_table;
int *ct;
{
    node_t *node;
    lsGen gen;

    foreach_node(network, gen, node) {
        ATPG_SET(node, 0);
    }
    recursive_mark(source_node);
    foreach_node(network, gen, node) {
        if (ATPG_GET(node) > 1) {
            if (verify_dom(source_node, node) == 0) {
                app_dom(dominator_table, source_node, node);
                (*ct)++;
            }
        }
    }
} 

void 
print_dom(dominator_table)
st_table *dominator_table;
{
    st_generator *gen;
    node_t *source_node;
    array_t *array;
    int ct;

    st_foreach_item(dominator_table, gen, (char **) &source_node, (char **) &array) {
        fprintf(misout, "%s is dominated by:\n", node_long_name(source_node));
        for (ct = 0; ct < array_n(array); ct++) {
            fprintf(misout, "%s\n", node_long_name((node_t *)array_fetch(int, array, ct)));
        }
    }
}

void 
dominators(network, dominator_table)
network_t *network;
st_table *dominator_table;
{
    lsGen gen;
    node_t *node;
    int ct=0;

    fprintf(misout, "Doing dominators.\n");
    foreach_node(network, gen, node) {
        if (node_num_fanout(node) > 1) {
            add_dom(network, node, dominator_table, &ct);
        }
    } 
    fprintf(misout, "%d dominators were detected.\n", ct);
}

void 
app_all_doms(array, dominator_table, marker, total_nodes)
array_t *array;
st_table *dominator_table;
int *marker;
int total_nodes;
{
    st_generator *gen;
    node_t *source_node;
    array_t *array1;
    int ct, act1, act2;

    st_foreach_item(dominator_table, gen, (char **) &source_node, (char **) &array1) {
        act1 = get_variable(source_node);
        if (marker[act1] != 0) {
            for (ct = 0; ct < array_n(array1); ct++) {
                act2 = get_variable(array_fetch(int, array1, ct));
                if (marker[act2] != 0) {
                    add_clause_2(array, -(3*total_nodes + act1), 3*total_nodes + act2);
                }
            }    
        }
    }
}
