#include "mis.h"
#include "bdd_int.h"

/*
 * Takes the given order and returns the bdd_manager with the matching
 * order.  Matching is defined as having a 1-1 correspondence between the
 * contents of the order_lists.
 *
 * If order_list is NIL or contains no nodes, returns the first bdd_manager.
 *
 * If no matching order is found, a new bdd_manager is created.
 */
static bdd_manager *
bdd_match_order(node,order_list)
node_t *node;
array_t *order_list;
{
    bdd_manager *bdd;
    int order_len,i;
    array_t *bdd_order;
    node_t *on,*bn;
    lsGen gen;
    network_t *network = node_network(node);
    lsList bdd_list = network_bdd_list(network);

    gen = lsStart(bdd_list);
    order_len = array_n(order_list);
    if (order_len == 0) {
        if (lsNext(gen,(lsGeneric *) &bdd,LS_NH) == LS_OK) {
	    (void) lsFinish(gen);
	    return(bdd);
	}
    }
    else {
        while (lsNext(gen,(lsGeneric *) &bdd,LS_NH) == LS_OK) {
	    bdd_order = bdd->order;
	    if (array_n(bdd_order) != order_len) {
	        continue;
	    }
	    i = 0;
	    do {
	        on = array_fetch(node_t *,order_list,i);
		bn = array_fetch(node_t *,bdd_order,i);
	    } while (on == bn && ++i < order_len);
	    if (on == bn) {
	        (void) lsFinish(gen);
		return(bdd);			/* found a match */
	    }
	}
    }
    (void) lsFinish(gen);
    bdd = bdd_init(network,order_list);
    return(bdd);
}

bdd_t *
bdd_constant(phase,order_list)
int phase;
array_t *order_list;
{
    bdd_manager *bdd;
    node_t *node;

    if (order_list == NIL(array_t) || array_n(order_list) == 0) {
        fail("bdd_constant: bad order_list");
    }
    node = array_fetch(node_t *,order_list,0);
    bdd = bdd_match_order(node,order_list);

    if (phase == 0) {
        return(bdd_zero(bdd));
    }
    return(bdd_one(bdd));
}

bdd_t *
bdd_construct(node,order_list,locality)
node_t *node;
array_t *order_list;
int locality;
{
    bdd_manager *bdd;

    if (order_list == NIL(array_t)) {
        fail("bdd_construct: invalid order_list");
    }

    bdd = bdd_match_order(node,order_list);
    return(bdd_do_construct(node,bdd,locality));
}

