#include "mis.h"
#include "bdd_int.h"

static void
get_branches(g,g1,g0)
bdd_node *g,**g1,**g0;
{
    bdd_node *G = BDD_REGULAR(g);

    *g1 = G->T;
    *g0 = G->E;
    if (BDD_IS_COMPLEMENT(g)) {
        *g1 = BDD_NOT(*g1);
	*g0 = BDD_NOT(*g0);
    }
}

static int
check_cube(g,bdd)
bdd_node *g;
bdd_manager *bdd;
{
    bdd_node *g1,*g0,*zero;
    
    if (g == BDD_ONE(bdd)) {
        return(1);
    }
    zero = BDD_ZERO(bdd);
    get_branches(g,&g1,&g0);

    if (g0 == zero) {
        return(check_cube(g1,bdd));
    }
    if (g1 == zero) {
        return(check_cube(g0,bdd));
    }
    return(0);
}

static bdd_node *
bdd_do_cofactor(f,g,bdd)
bdd_node *f,*g;
bdd_manager *bdd;
{
    bdd_node *one,*zero,*g1,*g0,*f1,*f0,*t,*e,*v,*r;
    unsigned int topf,topg;
    bdd_table *table;

    if (g == (one = BDD_ONE(bdd))) {
        return(f);
    }

    topf = BDD_REGULAR(f)->id;
    topg = BDD_REGULAR(g)->id;
    zero = BDD_ZERO(bdd);
    if (topf >= topg) {
        get_branches(g,&g1,&g0);
    }
    if (topf <= topg) {
       get_branches(f,&f1,&f0);
    }
    if (topf > topg) {        
	if (g0 == zero) {
	    return(bdd_do_cofactor(f,g1,bdd));
	}
	return(bdd_do_cofactor(f,g0,bdd));
    }
    if (topf == topg) {
        if (g0 == zero) {
	    return(bdd_do_cofactor(f1,g1,bdd));
	}
	return(bdd_do_cofactor(f0,g0,bdd));
    }

    t = bdd_do_cofactor(f1,g,bdd);
    bdd_ref(t);

    e = bdd_do_cofactor(f0,g,bdd);
    bdd_ref(e);

    table = bdd->table;
    v = bdd_find_or_add(table,topf,one,zero);
    r = ITE(bdd,v,t,e);

    bdd_recursive_deref(t,table);	/* t & e not necassarily part of r */
    bdd_recursive_deref(e,table);
    return(r);
}
        

bdd_t *
bdd_cofactor(f,g)
bdd_t *f,*g;
{
    bdd_t *h = ALLOC(bdd_t,1);
    bdd_node *gn;
    bdd_manager *bdd;

    if (f == NIL(bdd_t) || g == NIL(bdd_t)) {
        fail("bdd_cofactor: invalid BDD");
    }
    bdd = f->bdd;
    if (bdd != g->bdd) {
        fail("bdd_cofactor: different bdd managers");
    }
    gn = g->node;
    if (gn == BDD_ZERO(bdd)) {
        fail("bdd_cofactor: cofactor wrt zero not defined");
    }
    if (check_cube(gn,bdd) == 0) {
        fail("bdd_cofactor: cofactor only defined wrt single cube");
    }
    bdd_ref(h->node = bdd_do_cofactor(f->node,gn,bdd));
    h->bdd = bdd;
    return(h);
}


static bdd_node *
bdd_do_compose(f,v,g,bdd)
bdd_node *f,*g;
unsigned int v;
bdd_manager *bdd;
{
    bdd_node *f1,*f0,*r,*t,*e;
    bdd_table *table;
    unsigned int topf;

    topf = BDD_REGULAR(f)->id;
    if (topf > v) {
        return(f);
    }
    get_branches(f,&f1,&f0);

    t = bdd_do_compose(f1,v,g,bdd);
    bdd_ref(t);
    e = bdd_do_compose(f0,v,g,bdd);
    bdd_ref(e);

    table = bdd->table;
    if (topf == v) {
        r = ITE(bdd,g,t,e);
    }
    else {
        r = bdd_find_or_add(table,topf,BDD_ONE(bdd),BDD_ZERO(bdd));
	r = ITE(bdd,r,t,e);
    }
    bdd_recursive_deref(t,table);	/* t & e not necessarily part of r */
    bdd_recursive_deref(e,table);

    return((BDD_IS_COMPLEMENT(f) == 0) ? r : BDD_NOT(r));
}

bdd_t *
bdd_compose(f,g,h)
bdd_t *f,*g,*h;
{
    bdd_t *e = ALLOC(bdd_t,1);
    bdd_manager *bdd;
    bdd_node *n;

    if (f == NIL(bdd_t) || g == NIL(bdd_t) || h == NIL(bdd_t)) {
        fail ("bdd_compose: invalid BDD");
    }
    bdd = f->bdd;
    if (bdd != g->bdd || bdd != h->bdd) {
        fail("bdd_compose: different bdd managers");
    }

    n = g->node;
    if (BDD_IS_COMPLEMENT(n) == 1 || n->T != BDD_ONE(bdd) || n->E != BDD_ZERO(bdd)) {
        fail("bdd_compose: second argument not a variable");
    }

    e->bdd = bdd;
    bdd_ref(e->node = bdd_do_compose(f->node,n->id,h->node,bdd));
    return(e);
}
