#include "mis.h"
#include "bdd_int.h"

/*
 * 'one' gets the highest number available.
 */
bdd_manager *
bdd_init(network,order_list)
network_t *network;
array_t *order_list;
{
    bdd_manager *bdd;
    int i,size;
    node_t *node;
    array_t *order;

    size = (order_list == NIL(array_t)) ? 0 : array_n(order_list);

    bdd = ALLOC(bdd_manager,1);
    bdd->table = bdd_init_table();
    bdd->vars = 0;
    bdd->one = bdd_find_or_add(bdd->table,~0,NIL(bdd_node),NIL(bdd_node));
    bdd->iton = array_alloc(node_t *,10);
    bdd->order = array_alloc(node_t *,size);
    bdd->network = network;
    bdd_ref(bdd->one);
    bdd->handle = network_add_bdd(network,bdd);
    bdd->local_order = NIL(st_table);
    
    order = bdd->order;
    for (i = 0; i < size; i++) {
        node = array_fetch(node_t *,order_list,i);
        array_insert(node_t *,order,i,node);
    }
    return(bdd);
}

/*
 * Deletes resources associated with a single bdd manager
 */
void
bdd_quit(bdd)
bdd_manager *bdd;
{
    network_t *network = bdd->network;
    lsGen gen;
    node_t *node;
    lsGeneric dummy;
    st_generator *stgen;
    char *x;
    bdd_t *f;
    
    if (bdd != NIL(bdd_manager)) {
        (void) lsRemoveItem(bdd->handle,&dummy);
        foreach_node (network,gen,node) {
	    f = bdd_at_node(node);
	    if (f != NIL(bdd_t) && f->bdd == bdd) {
	        bddDealloc(node);	/* quick delete: no DAG traversal */
	    }
	}

	if (bdd->local_order != NIL(st_table)) {
	    st_foreach_item(bdd->local_order,stgen,&x,(char **) &f) {
	        FREE(f);
	    }
	    st_free_table(bdd->local_order);
	}

        bdd_free_table(bdd->table);
	array_free(bdd->iton);
	array_free(bdd->order);
	FREE(bdd);
    }
}

/*
 * Called only by network_free.
 */
void
bdd_terminate(bdd)
bdd_manager *bdd;
{
    st_generator *stgen;
    char *x;
    bdd_t *f;

    if (bdd->local_order != NIL(st_table)) {
        st_foreach_item(bdd->local_order,stgen,&x,(char **) &f) {
	    FREE(f);
	}
	st_free_table(bdd->local_order);
    }
    bdd_free_table(bdd->table);
    array_free(bdd->iton);
    array_free(bdd->order);
    FREE(bdd);
}
