#define BDD_STATS		0

#define BDD_UNIQUE_SLOTS	127
#define BDD_CACHE_SLOTS		503
#define BDD_NOT(node)		((bdd_node *) ((int) (node) ^ 01))
#define BDD_REGULAR(node)	((bdd_node *) ((int) (node) & ~01))
#define BDD_IS_COMPLEMENT(node)	((int) (node) & 01)

#define BDD_non_constant	((bdd_node *) 1)
#define BDD_MAXREF		((unsigned short) ~0)

#define BDD_ONE(bdd)		(bdd)->one
#define BDD_ZERO(bdd)		(BDD_NOT(BDD_ONE(bdd)))

#define HEURISTIC	0
#define FANIN		1
#define RANDOM		2
#define ALAP		3
#define UNORDERED	4

#define MAX_TABLE_DENSITY	4
#define GC_PERCENT		10	/* gc when this percent are dead */
#define BDD_GROW_FACTOR		2	/* table increases on rehash */

			/* suggested by paper */
#define bdd_hash(a,b,c)		(((a) << 5) + ((b) << 7) + ((c) << 11))

typedef struct bdd_node {	/* bit0 of node ptr = 1 means complement ptr */
    unsigned short id;		/* unique id */
    unsigned short ref;		/* reference counts */
    struct bdd_node *T;		/* then */
    struct bdd_node *E;		/* else */
    struct bdd_node *next;	/* next pointer for symbol table */
} bdd_node;

typedef struct bdd_cache {
    bdd_node *f,*g,*h;		/* key for the cache */
    bdd_node *data;		/* already constructed bdd for ITE(f,g,h) */
} bdd_cache;

typedef struct bdd_table {	/* specialized bdd symbol table */
    unsigned int slots;		/* number of hash buckets */
    unsigned int keys;		/* number of data items */
    unsigned int dead;		/* number of dead nodes */
    bdd_cache *cache;		/* the cache */
    bdd_node **nodelist;	/* data */
    bdd_node **memory_list;	/* memory manager for symbol table */
    bdd_node *next_free;	/*	"	*/
    unsigned int max_keys;	/* when to rehash */
} bdd_table;

/* already typedef'd in bdd.h */

struct bdd_manager {
    bdd_table *table;		/* the specialized symbol table */
    unsigned int vars;		/* the largest unique id used so far */
    bdd_node *one;		/* constant */
    array_t *iton;		/* maps ids in bdd_node to node_t */
    array_t *order;		/* copy of order_list */
    lsHandle handle;		/* where it is in network bdd list */
    network_t *network;
    st_table *local_order;	/* for local BDDs */
};

struct bdd_t {
    bdd_node *node;		/* normal or complement pointer to node */
    bdd_manager *bdd;		/* back pointer to find 1, cache, table */
};

	/* bdd_cache.c */
extern void bdd_free_cache();
extern void bdd_cache_insert();
extern bdd_node *bdd_cache_lookup();

	/* bdd_ite.c */
extern bdd_node *ITE();
extern bdd_node *ITE_constant();

	/* bdd_order.c */
extern array_t *alap_order_nodes();
extern array_t *shuffle();
extern array_t *order_nodes();

	/* bdd_table.c */
extern bdd_table *bdd_init_table();
extern void bdd_free_table();
extern bdd_node *bdd_find_or_add();

	/* bdd_ref.c */
extern void bdd_ref();
extern void bdd_reclaim();
extern void bdd_recursive_deref();
extern void bdd_deref();

	/* com_bdd.c */
extern void bddDealloc();

#if BDD_STATS 
extern int memused,nodesallocated,nodesfreed,nodesreclaimed,nodesdropped;
extern int cachehits,cachecollisions,cacheinserts;
#endif 
