#include "mis.h"
#include "bdd_int.h"

/*
 * Replace variables with constants if possible (part of canonical form).
 */
static void
var_to_const(f,gp,hp,one)
bdd_node *f,**gp,**hp,*one;
{
    bdd_node *g = *gp;
    bdd_node *h = *hp;
    
    if (f == g) {			/* ITE(F,F,H) = ITE(F,1,H) = F + H */
        *gp = one;
    }
    else if (f == BDD_NOT(g)) {		/* ITE(F,!F,H) = ITE(F,0,H) = !F * H */
        *gp = BDD_NOT(one);
    }

    if (f == h) {			/* ITE(F,G,F) = ITE(F,G,0) = F * G */
        *hp = BDD_NOT(one);
    }
    else if (f == BDD_NOT(h)) {		/* ITE(F,G,!F) = ITE(F,G,1) = !F + G */
        *hp = one;
    }
}

/*
 * Pick unique member from equiv expressions.
 * Reduces 2 variable expressions to canonical form (p. 16,17)
 */
static int
var_to_canonical(fp,gp,hp,one,topfp,topgp,tophp)
bdd_node **fp,**gp,**hp,*one;
unsigned int *topfp,*topgp,*tophp;
{
    register bdd_node *F,*G,*H,*r,*f,*g,*h;
    register unsigned int topf,topg,toph;
    int complement,change;

    f = *fp;
    g = *gp;
    h = *hp;
    F = BDD_REGULAR(f);
    G = BDD_REGULAR(g);
    H = BDD_REGULAR(h);
    topf = F->id;
    topg = G->id;
    toph = H->id;

    change = 0;

    if (G == one) {				/* ITE(F,c,H) */
	if ((topf > toph) || (topf == toph && f > h)) {
	    r = h; h = f; f = r;		/* ITE(F,1,H) = ITE(H,1,F) */
	    if (g != one) {
	        f = BDD_NOT(f);			/* ITE(F,0,H) = ITE(!H,0,!F) */
		h = BDD_NOT(h);
	    }
	    change = 1;
	}
    }
    else if (H == one) {			/* ITE(F,G,c) */
        if ((topf > topg) || (topf == topg && f > g)) {
	    r = g; g = f; f = r;		/* ITE(F,G,0) = ITE(G,F,0) */
	    if (h == one) {
	        f = BDD_NOT(f);			/* ITE(F,G,1) = ITE(!G,!F,1) */
		g = BDD_NOT(g);
	    }	        
	    change = 1;
	}
    }
    else if (g == BDD_NOT(h)) {			/* ITE(F,G,!G) = ITE(G,F,!F) */
        if ((topf > topg) || (topf == topg && f > g)) {
	    r = f; f = g; g = r; h = BDD_NOT(r);
	    change = 1;
	}
    }

      /** adjust pointers so that the first 2 arguments to ITE are regular **/

    if (BDD_IS_COMPLEMENT(f) != 0) {	/* ITE(!F,G,H) = ITE(F,H,G) */
        f = BDD_NOT(f);
	r = g;
	g = h;
	h = r;
	change = 1;
    }

    complement = 0;
    if (BDD_IS_COMPLEMENT(g) != 0) {	/* ITE(F,!G,H) = ! ITE(F,G,!H) */
        g = BDD_NOT(g);
	h = BDD_NOT(h);
	change = 1;
	complement = 1;
    }

    if (change != 0) {
        *fp = f;
	*gp = g;
	*hp = h;
    }

    *topfp = f->id;
    *topgp = g->id;
    *tophp = BDD_REGULAR(h)->id;

    return(complement);
}

#define bdd_min(x,y)	(((x) < (y)) ? (x) : (y))

bdd_node *
ITE(bdd,f,g,h)
bdd_manager *bdd;
bdd_node *f,*g,*h;
{
    bdd_node *one,*zero;
    bdd_node *r,*Fv,*Fnv,*Gv,*Gnv,*H,*Hv,*Hnv,*t,*e;
    unsigned int topf,topg,toph,v;
    int complement;
    bdd_table *table;

/* Trivial cases */

	/** one variable cases **/

    if (f == (one = BDD_ONE(bdd))) {	/* ITE(1,G,H) = G */
        return(g);
    }
    if (f == (zero = BDD_ZERO(bdd))) {	/* ITE(0,G,H) = H */
        return(h);
    }

	/** from now on, f is known to not be a constant **/

    var_to_const(f,&g,&h,one);

	/** check remaining one variable cases **/
	
    if (g == h) {			/* ITE(F,G,G) = G */
        return(g);
    }
    if (g == one) {			/* ITE(F,1,0) = F */
        if (h == zero) {
	    return(f);
	}
    }
    else if (g == zero) {		/* ITE(F,0,1) = !F */
        if (h == one) {
	    return(BDD_NOT(f));
	}
    }
	/** from here, there is at most one constant (only G or H) **/

    complement = var_to_canonical(&f,&g,&h,one,&topf,&topg,&toph);

    	/** f & g are now regular pointers **/

	/** a shortcut: ITE(F,G,H) = (v,G,H) if F=(v,1,0), v < top(G,H) **/

    v = bdd_min(topg,toph);

    table = bdd->table;
    if (topf < v && f->T == one && f->E == zero) {
        r = bdd_find_or_add(table,f->id,g,h);
	return((complement == 0) ? r : BDD_NOT(r));
    }

/* Check cache */

    r = bdd_cache_lookup(table,f,g,h);
    if (r != NIL(bdd_node) && r != BDD_non_constant) {
        return((complement == 0) ? r : BDD_NOT(r));
    }

/* Recursive step */

    v = bdd_min(topf,v);	/* v = top_var(F,G,H) */

    if (topf == v) {
        Fv = f->T;
	Fnv = f->E;
    }
    else {
        Fv = Fnv = f;
    }
    if (topg == v) {
        Gv = g->T;
	Gnv = g->E;
    }
    else {
        Gv = Gnv = g;
    }
    if (toph == v) {
        H = BDD_REGULAR(h);
        Hv = H->T;
	Hnv = H->E;
	if (BDD_IS_COMPLEMENT(h) == 1) {
	    Hv = BDD_NOT(Hv);
	    Hnv = BDD_NOT(Hnv);
	}
    }
    else {
        Hv = Hnv = h;
    }
    
    t = ITE(bdd,Fv,Gv,Hv);
    bdd_ref(t);

    e = ITE(bdd,Fnv,Gnv,Hnv);

    r = (t == e) ? t : bdd_find_or_add(table,v,t,e);

    bdd_deref(t);	/* t is incorporated into r, only need to deref */

    bdd_cache_insert(table,f,g,h,r);

    if (complement != 0) {
        r = BDD_NOT(r);
    }

    return(r);
}

#define BDD_IS_CONSTANT(node)	(BDD_REGULAR(node) == one)

bdd_node *
ITE_constant(bdd,f,g,h)
bdd_manager *bdd;
bdd_node *f,*g,*h;
{
    bdd_node *r,*Fv,*Fnv,*Gv,*Gnv,*H,*Hv,*Hnv,*t,*e;
    bdd_node *one = BDD_ONE(bdd);
    bdd_node *zero = BDD_ZERO(bdd);
    int complement;
    unsigned int topf,topg,toph,v;
    bdd_table *table;

/* Trivial cases */

    if (f == one) {		/* ITE(1,G,H) => G */
        return(g);
    }
    if (f == zero) {		/* ITE(0,G,H) => H */
        return(h);
    }

	/* f now not a constant */

    var_to_const(f,&g,&h,one);	/* possibly convert g or h to constants */

    if (g == h) {		/* ITE(F,G,G) => G */
        return(g);
    }
    if (BDD_IS_CONSTANT(g) && BDD_IS_CONSTANT(h)) {
        return(BDD_non_constant);  /* ITE(F,1,0) or ITE(F,0,1) => non_const */
    }

    complement = var_to_canonical(&f,&g,&h,one,&topf,&topg,&toph);

/* Cache lookup */

    table = bdd->table;
    r = bdd_cache_lookup(table,f,g,h);
    if (r != NIL(bdd_node)) {
        return((complement == 0) ? r : BDD_NOT(r));
    }

/* Recursion */

    v = bdd_min(topg,toph);    
    v = bdd_min(topf,v);	/* v = top_var(F,G,H) */

    if (topf == v) {
        Fv = f->T;
	Fnv = f->E;
    }
    else {
        Fv = Fnv = f;
    }
    if (topg == v) {
        Gv = g->T;
	Gnv = g->E;
    }
    else {
        Gv = Gnv = g;
    }
    if (toph == v) {
        H = BDD_REGULAR(h);
        Hv = H->T;
	Hnv = H->E;
	if (BDD_IS_COMPLEMENT(h) == 1) {
	    Hv = BDD_NOT(Hv);
	    Hnv = BDD_NOT(Hnv);
	}
    }
    else {
        Hv = Hnv = h;
    }
    
    t = ITE_constant(bdd,Fv,Gv,Hv);
    if (!BDD_IS_CONSTANT(t)) {
        bdd_cache_insert(table,f,g,h,BDD_non_constant);
	return(BDD_non_constant);
    }
    
    e = ITE_constant(bdd,Fnv,Gnv,Hnv);
    if (!BDD_IS_CONSTANT(e) || t != e) {
        bdd_cache_insert(table,f,g,h,BDD_non_constant);
        return(BDD_non_constant);
    }
    bdd_cache_insert(table,f,g,h,t);
    if (complement != 0) {
        t = BDD_NOT(t);
    }
    return(t);
}
