#include "mis.h"
#include "bdd_int.h"

/* for order nodes */
typedef struct  info_defn{
	node_t *node;
	int level;
	int max_leaf;
} info_t ;

static int max_leaf_cmp();
static void rec_level();
static void rec_order();
static int rec_max();

array_t *
order_nodes(node_vec, pi_only)
array_t *node_vec;
int pi_only;
{
	node_t *node, *factor_node, *pi;
	network_t *network;
	array_t *dfs_array, *factor_array, *order_list, *node_list;
	st_table *info_table, *factor_table, *max_table;
	int i, j, cur_level, max_leaf;
	info_t *info;
	lsGen gen;
	char *dummy;

	/*	form dfs array	*/
	if(array_n(node_vec) == 0) return NIL(array_t);

	node = array_fetch(node_t *, node_vec, 0);
	network = node_network(node);
	dfs_array = network_dfs_from_input(network);

	/*	to get fains of any NODE_POs in node_vec	*/
	node_list = array_alloc(node_t *, 0);

	/*	initialize info_table	*/
	info_table = st_init_table(st_ptrcmp, st_ptrhash);
	for(i=0; i<array_n(node_vec); i++){
		node = array_fetch(node_t *, node_vec, i);
		if(node_function(node) == NODE_PO){
			node = node_get_fanin(node, 0);
		}
		array_insert_last(node_t *, node_list, node);
		info = ALLOC(info_t, 1);
		info->node = node;
		info->level = 0;
		info->max_leaf = 0;
		(void)st_insert(info_table, (char *)node, (char *)info);
	}

	/*	traverse dfs_array	*/
	for(i=0; i<array_n(dfs_array); i++){
		node = array_fetch(node_t *, dfs_array, i);
		if((node_function(node) != NODE_PI) && 
		(st_lookup(info_table, (char *)node, &dummy))){
			info = (info_t *)dummy;
			factor_array = factor_to_nodes(node);
			factor_table = st_init_table(st_ptrcmp, st_ptrhash);
			for(j=0; j<array_n(factor_array); j++){
				factor_node = array_fetch(node_t *, factor_array, j);
				(void)st_insert(factor_table, (char *)factor_node, (char *)1);
			}
			factor_node = array_fetch(node_t *, factor_array, 0);
			cur_level = info->level;
			(void)rec_level(factor_node, cur_level, factor_table, info_table);
			for(j=0; j<array_n(factor_array); j++){
				factor_node = array_fetch(node_t *, factor_array, j);
				node_free(factor_node);
			}
			array_free(factor_array);
			st_free_table(factor_table);
		}
	}

	array_free(dfs_array);

	/*	update max_leaf fields	*/
	max_table = st_init_table(st_ptrcmp, st_ptrhash);
	foreach_primary_input(network, gen, pi){
		if(st_lookup(info_table, (char *)pi, &dummy)){
			info = (info_t *)dummy;
			max_leaf = info->level;
			info->max_leaf = max_leaf;
			(void)st_insert(max_table, (char *)pi, (char *)max_leaf);
		}
	}

	for(i=0; i<array_n(node_list); i++){
		node = array_fetch(node_t *, node_list, i);
		(void) rec_max(node, max_table, info_table);
	}

	st_free_table(max_table);

	/*	all the bookkeeping done, go ahead and order	*/
	order_list = array_alloc(node_t *, 0);
	rec_order(node_list, pi_only, info_table, order_list);

	st_free_table(info_table);
	array_free(node_list);

	return order_list;
}

static void
rec_level(node, level, factor_table, info_table)
node_t *node;
int level;
st_table *factor_table, *info_table;
{
	int i;
	node_t *fanin;
	info_t *info;
	char *dummy;

	foreach_fanin(node, i, fanin){
		if(st_lookup(factor_table, (char *)fanin, &dummy)){
			rec_level(fanin, level+1, factor_table, info_table);
		} else {
			if(st_lookup(info_table, (char *)fanin, &dummy)){
				info = (info_t *)dummy;
				info->level = MAX(info->level, level+1);
			} else {
				info = ALLOC(info_t, 1);
				info->node = fanin;
				info->level = level+1;
				info->max_leaf = 0;
				(void)st_insert(info_table, (char *)fanin, (char *)info);
			}
		}
	}
}

static int
max_leaf_cmp(key1, key2)
char **key1, **key2;
{
	info_t *info1, *info2;

	info1 = *((info_t **) key1);
	info2 = *((info_t **) key2);
	if(info1->max_leaf < info2->max_leaf) return 1;
		else if (info1->max_leaf > info2->max_leaf) return -1;
		 	else if(info1->level > info2->level) return 1;
				else if(info1->level < info2->level) return -1;
					else return 0;
}

static void
rec_order(node_vec, pi_only, info_table, order_list)
array_t *node_vec, *order_list;
int pi_only;
st_table *info_table;
{
	array_t *info_array, *fanin_array, *node_list;
	int i,j;
	node_t *node, *fanin;
	info_t *info;
	char *dummy;

	info_array = array_alloc(info_t *, 0);
	for(i=0; i<array_n(node_vec); i++){
		node = array_fetch(node_t *, node_vec, i);
		if(st_lookup(info_table, (char *)node, &dummy)){
			info = (info_t *)dummy;
			array_insert_last(info_t *, info_array, info);
		}
	}

	array_sort(info_array, max_leaf_cmp);

	node_list = array_alloc(node_t *, 0);
	for(i=0; i<array_n(info_array); i++){
		info = array_fetch(info_t *, info_array, i);
		node = info->node;
		array_insert_last(node_t *, node_list, node);
	}
	array_free(info_array);

	for(i=0; i<array_n(node_list); i++){
		node = array_fetch(node_t *, node_list, i);
		if(st_lookup(info_table, (char *)node, &dummy)){
			if(node_function(node) != NODE_PI){ 
				fanin_array = array_alloc(node_t *, 0);
				foreach_fanin(node, j, fanin){
					array_insert_last(node_t *, fanin_array, fanin);
				}
				rec_order(fanin_array, pi_only, info_table, order_list);
				array_free(fanin_array);
			} 
			if((!pi_only) || (node_function(node) == NODE_PI))
				array_insert_last(node_t *, order_list, node);
			(void)st_delete(info_table, (char **)&node, &dummy);
			info = (info_t *)dummy;
			FREE(info);
		}
	}
	array_free(node_list);
}

static int
rec_max(node, max_table, info_table)
node_t *node;
st_table *max_table, *info_table;
{
	int max_leaf, i;
	info_t *info;
	char *dummy;
	node_t *fanin;

	if(st_lookup(max_table, (char *)node, &dummy))
		return((int)dummy);
	
	max_leaf = 0;
	foreach_fanin(node, i, fanin){
		if(st_lookup(max_table, (char *)fanin, &dummy)){
			max_leaf = MAX(max_leaf, (int)dummy);
		} else {
			max_leaf = MAX(max_leaf, rec_max(fanin, max_table, info_table)); 
		}
	}
	(void)st_lookup(info_table, (char *)node, &dummy); 
	info = (info_t *)dummy;
	info->max_leaf = max_leaf;
	(void)st_insert(max_table, (char *)node, (char *)max_leaf);
	return max_leaf;
}

array_t *
shuffle(list)
array_t *list;
{
	array_t *newlist, *mark;
	int i, list_size, next_index, count, j, index, rand_num;
	node_t *new_node;
	
	mark = array_alloc(int, 0);
	newlist = array_alloc(node_t *, 0);

	list_size = array_n(list);
	for(i = 0; i < list_size; i++){
		array_insert(int , mark, i, 0);
	}

	for(i = 0; i<list_size; i++){
		/*NOSTRICT*/
		rand_num = rand() % 32768;
		/* taking random numbers from 0 - 2**15	
		this should be portable across most machines	*/
		next_index = (rand_num / 32768.0)*(list_size - i);
		/* normalizing it and multiplying by the range	*/
		count = -1;
		for(j = 0; j<list_size; j++){
			if(array_fetch(int, mark, j) == 0) count++;
			if(count == next_index){
				array_insert(int, mark, j, 1);
				index = j;
				 break;
			}
		}
		new_node = array_fetch(node_t *, list, index);
		array_insert(node_t *, newlist, i, new_node);
	}
	array_free(mark);
	array_free(list);
	return(newlist);
}
