#include "mis.h"
#include "bdd_int.h"

void
bdd_ref(node)
bdd_node *node;
{
    bdd_node *n;


    if (node == NIL(bdd_node)) {
        return;
    }

    n = BDD_REGULAR(node);

    if (n->ref < BDD_MAXREF) {
        n->ref++;
    }
}

static void
bdd_do_reclaim(n,table)
bdd_node *n;
bdd_table *table;
{
    unsigned int ref;

    n = BDD_REGULAR(n);

    ref = n->ref;
    if (ref < BDD_MAXREF) {
        n->ref++;
    }

    if (ref == 0) {		/* used to be dead */
        table->dead--;
#if BDD_STATS
	nodesreclaimed++;
#endif
	bdd_do_reclaim(n->T,table);
	bdd_do_reclaim(n->E,table);
    }
}

/* Bring children of dead node back */
void
bdd_reclaim(n,table)
bdd_node *n;
bdd_table *table;
{
    n = BDD_REGULAR(n);
    table->dead--;
#if BDD_STATS
    nodesreclaimed++;
#endif
    bdd_do_reclaim(n->T,table);
    bdd_do_reclaim(n->E,table);
}

void
bdd_recursive_deref(n,table)
bdd_node *n;
bdd_table *table;
{
    int ref;

    n = BDD_REGULAR(n);
    ref = n->ref;

    if (ref == 0) {
        fail("bdd_recursive_deref: dereferencing something with no ref count");
    }
    if (ref == 1) {		/* this node will go dead */
        table->dead++;
    }
    if (ref != BDD_MAXREF) {
        ref = --n->ref;
    }
    
    if (ref == 0) {
#if BDD_STATS
	nodesdropped++;
#endif
	bdd_recursive_deref(n->T,table);
	bdd_recursive_deref(n->E,table);
    }
}

void
bdd_deref(node)
bdd_node *node;
{
    bdd_node *n = BDD_REGULAR(node);

    assert(n->ref > 0);
    if (n->ref != BDD_MAXREF) 
	--(n->ref);
}
