#include "mis.h"
#include "port.h"
#include "bdd_int.h"

/*ARGSUSED*/
static enum st_retval
bdd_table_delete(key,val,arg)
char *key,*val,*arg;
{
    FREE(key);
    return(ST_DELETE);
}

/*
 * Orders the nodes in list2 so that the corresponding nodes in both lists
 * have the same names.
 */
static int
same_order(list1,list2p)
array_t *list1,**list2p;
{
    array_t *list2,*out_list;
    int size,i,bad,index;
    st_table *tab;
    char *n1,*n2;
    node_t *node;

    list2 = *list2p;
    size = array_n(list1);

    if (size != array_n(list2)) {
        return(0);
    }

    tab = st_init_table(strcmp,st_strhash);
    out_list = array_alloc(node_t *,0);

    for (i = 0; i < size; i++) {
        n1 = util_strsav(node_long_name(array_fetch(node_t *,list1,i)));
	(void) st_insert(tab,n1,(char *) i);
    }

    bad = 0;
    for (i = 0; i < size; i++) {
        node = array_fetch(node_t *,list2,i);
	n2 = node_long_name(node);
	if (st_lookup(tab,n2,(char **) &index) == 1) {
	    array_insert(node_t *,out_list,index,node);
	}
	else {
	    bad = 1;
	    break;
	}
    }

    (void) st_foreach(tab,bdd_table_delete,NIL(char));
    st_free_table(tab);

    if (bad != 0) {
        array_free(out_list);
	return(0);
    }
    array_free(list2);
    *list2p = out_list;
    return(1);
}

int
bdd_verify(net1,net2)
network_t *net1,*net2;
{
    array_t *po_list1,*po_list2,*pi_list1,*pi_list2,*temp;
    lsGen gen;
    node_t *po,*ipo,*n1,*n2;
    bdd_manager *bdd;
    bdd_t *f1,*f2;
    bdd_node *fn1,*fn2;
    int i,stat;
    
    if (net1 == net2) {
        return(1);
    }

    if (net1 == NIL(network_t) || net2 == NIL(network_t)) {
        return(0);
    }
    po_list1 = array_alloc(node_t *,0);
    foreach_primary_output (net1,gen,po) {
	array_insert_last(node_t *,po_list1,po);
    }

    po_list2 = array_alloc(node_t *,0);
    foreach_primary_output (net2, gen, po) {
	array_insert_last(node_t *,po_list2,po);
    }

    if (same_order(po_list1,&po_list2) == 0) {
        (void) fprintf(miserr,
	      "The name/number of outputs of the two networks don't match.\n");
	array_free(po_list1);
	array_free(po_list2);
	return(0);
    }

	for(i=0; i<array_n(po_list1); i++){
		po = array_fetch(node_t *, po_list1, i);
		ipo = node_get_fanin(po, 0);
		array_insert(node_t *, po_list1, i, ipo);
		po = array_fetch(node_t *, po_list2, i);
		ipo = node_get_fanin(po, 0);
		array_insert(node_t *, po_list2, i, ipo);
	}

    /*
    (void) fprintf(misout,"Ordering first network\n");
    */

    pi_list1 = order_nodes(po_list1,1);

    /*
    (void) fprintf(misout,"Ordering second network\n");
    */

    pi_list2 = order_nodes(po_list2,1);

    if (same_order(pi_list1,&pi_list2) == 0) {
        (void) fprintf(miserr,
	       "The name/number of inputs of the two networks don't match.\n");
	stat = 0;
    }
    else {
        bdd = bdd_init(net1,NIL(array_t));

	temp = bdd->order;

	stat = 1;
	for (i = array_n(po_list1) - 1; i >= 0; i--) {
	    n1 = array_fetch(node_t *,po_list1,i);
	    n2 = array_fetch(node_t *,po_list2,i);

	    bdd->order = pi_list1;			/* hokey */
	    f1 = bdd_do_construct(n1,bdd,0);

	    bdd->order = pi_list2;			/* hokey */
	    f2 = bdd_do_construct(n2,bdd,0);

	    fn1 = f1->node;
	    fn2 = f2->node;
            if (fn1 != fn2) {
	        (void) fprintf(miserr,
	    		"Networks differ on (at least) primary output %s\n",
			node_name(n1));
		bdd_node_free(n1);
		bdd_node_free(n2);
		stat = 0;
		break;
	    }

	    bdd_node_free(n1);	/* frees f1 & f2 */
	    bdd_node_free(n2);
	}
	foreach_node (net2,gen,n2) {		/* hokey */
	    f2 = bdd_at_node(n2);
	    if (f2 != NIL(bdd_t) && f2->bdd == bdd) {
	        bddDealloc(n2);
	    }
	}
	bdd->order = temp;
	bdd_quit(bdd);
    }
    array_free(pi_list1);
    array_free(pi_list2);
    array_free(po_list1);
    array_free(po_list2);

    return(stat);
}



