#include "mis.h"
#include "port.h"
#include "com_int.h"

network_t *backup_network;

static int autoexec;		/* indicates currently in autoexec */
static jmp_buf env;

static int com_dispatch();
static int apply_alias();
static char *split_line();
static int check_shell_escape();


/* ARGSUSED */
static SIGNAL_FN
sigterm(sig, code, scp)
{
    (void) signal(SIGINT, SIG_IGN);	/* ignore further ctl-c */
    longjmp(env, 1);
}



com_execute(network, command)
network_t **network;
char *command;
{
    int status, argc;
    char *commandp, **argv;

    (void) signal(SIGINT, SIG_IGN);
    commandp = command;
    do {
	if (check_shell_escape(commandp, &status)) break;
	commandp = split_line(commandp, &argc, &argv);
	status = apply_alias(&argc, &argv);
	if (status == 0) {
	    status = com_dispatch(network, argc, argv);
	}
	com_free_argv(argc, argv);
    } while (status == 0 && *commandp != '\0');
    return status;
}

static int
com_dispatch(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int status;
    char *value;
    command_descr_t *descr;

    if (argc == 0) {		/* empty command */
	return 0;
    }

    if (! avl_lookup(command_table, argv[0], &value)) {
	(void) fprintf(miserr, "unknown command '%s'\n", argv[0]);
	return 1;
    }

    descr = (command_descr_t *) value;
    if (setjmp(env)) {
	/* return from control-c -- restore the network */
	if (descr->changes_network) {
	    *network = backup_network;
	    backup_network = NIL(network_t);
	}
	return 1;
    }

    if (descr->changes_network) {
	if (backup_network != NIL(network_t)) {
	    network_free(backup_network);
	}
	backup_network = network_dup(*network);
    }

#ifndef _IBMR2
    (void) signal(SIGINT, sigterm);
#endif
    status = (*descr->command_fp)(network, argc, argv);

    /* automatic execution of arbitrary command after each command */
    /* usually this is a passive command ... */
    if (status == 0 && ! autoexec) {
	if (avl_lookup(flag_table, "autoexec", &value)) {
	    autoexec = 1;
	    status = com_execute(network, value);
	    autoexec = 0;
	}
    }

    (void) signal(SIGINT, SIG_IGN);
    return status;
}

static int
apply_alias(argc, argv)
int *argc;
char ***argv;
{
    int i, loop;
    char *value;
    alias_descr_t *alias;

    for(loop = 0; loop < 20; loop++) {
	if (*argc == 0) return 0;
	if (! avl_lookup(alias_table, (*argv)[0], &value)) {
	    return 0;
	}
	alias = (alias_descr_t *) value;

	FREE((*argv)[0]);
	*argv = REALLOC(char *, *argv, *argc + alias->argc);
	for(i = *argc-1; i >= 1; i--) {
	    (*argv)[i + alias->argc - 1] = (*argv)[i];
	}
	for(i = 0; i < alias->argc; i++) {
	    (*argv)[i] = util_strsav(alias->argv[i]);
	}
	*argc += alias->argc - 1;
    }

    (void) fprintf(miserr, "error -- alias loop\n");
    return 1;
}

static char *
split_line(command, argc, argv)
char *command;
int *argc;
char ***argv;
{
    register char *p, *start, c;
    register int i, j;
    register char *new_arg;
    array_t *argv_array;
    int single_quote, double_quote;

    argv_array = array_alloc(char *, 5);

    p = command;
    for(;;) {
	/* skip leading white space */
	while (isspace(*p)) {
	    p++;
	}

	/* skip until end of this token */
	single_quote = double_quote = 0;
	for(start = p; (c = *p) != '\0'; p++) {
	    if (c == ';' || c == '#' || isspace(c)) {
		if (! single_quote && ! double_quote) {
		    break;
		}
	    }
	    if (c == '\'') {
		single_quote = ! single_quote;
	    }
	    if (c == '"') {
		double_quote = ! double_quote;
	    }
	}
	if (single_quote || double_quote) {
	    (void) fprintf(miserr, "ignoring unbalanced quote ...\n");
	}
	if (start == p) break;

	new_arg = ALLOC(char, p - start + 1);
	j = 0;
	for(i = 0; i < p - start; i++) {
	    c = start[i];
	    if (c != '\'' && c != '\"') {
		new_arg[j++] = isspace(c) ? ' ' : start[i];
	    }
	}
	new_arg[j] = '\0';
	array_insert_last(char *, argv_array, new_arg);
    }

    *argc = array_n(argv_array);
    *argv = array_data(char *, argv_array);
    array_free(argv_array);
    if (*p == ';') {
	p++;
    } else if (*p == '#') {
	for(; *p != 0; p++) ;		/* skip to end of line */
    }
    return p;
}

static int
check_shell_escape(p, status)
char *p;
int *status;
{
    while (isspace(*p)) {
	p++;
    }
    if (*p == '!') {
	*status = system(p+1);
	return 1;
    }
    return 0;
}
