#include "mis.h"
#include "com_int.h"


/* ARGSUSED */
com_time(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    static long last_time = 0;
    long time;

    if (argc != 1) goto usage;

    time = util_cpu_time();
    (void) fprintf(misout, 
	"elapse: %2.1f seconds, total: %2.1f seconds\n", 
	(time - last_time) / 1000.0, time / 1000.0);
    last_time = time;
    return 0;

usage:
    (void) fprintf(miserr, "usage: time");
    return 1;
}

/* ARGSUSED */
com_usage(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    if (argc != 1) {
	(void) fprintf(miserr, "usage: usage -- give CPU usage statistics");
	return 1;
    }

    util_print_cpu_stats(misout);
    return 0;
}

/* ARGSUSED */
com_echo(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i;

    for(i = 1; i < argc; i++) {
       (void) fprintf(misout, "%s ", argv[i]);
    }
    (void) fprintf(misout, "\n");
    return 0;
}


/* ARGSUSED */
com_quit(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    return -1;
}


/* ARGSUSED */
static int
com_infinite_loop(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    *network = 0;
    for (;;) ;
}


/* ARGSUSED */
static int
com_save_image(network, argc, argv)
network_t *network;
int argc;
char **argv;
{
    FILE *fp;
    char *file1, *file2;

    if (argc != 2) {
	(void) fprintf(stderr, "usage: save filename\n");
	return 1;
    }

    /* get current executable name by searching the path ... */
    file1 = util_path_search(program_name);
    if (file1 == 0) {
	(void) fprintf(miserr, "cannot locate current executable\n");
	return 1;
    }

    /* users name for the new executable -- perform tilde-expansion */
    fp = com_open_file(argv[1], "w", &file2, /* silent */ 0);
    if (fp == NULL) return 1;
    if (fp != stdout) (void) fclose(fp);

    if (! util_save_image(file1, file2)) {
	(void) fprintf(stderr, "error occured during save ...\n");
	return 1;
    }
    FREE(file1);
    FREE(file2);
    return 0;
}

/* ARGSUSED */
com_which(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    FILE *fp;
    char *filename;

    if (argc != 2) return 1;

    fp = com_open_file(argv[1], "r", &filename, 0);
    if (fp != 0) {
	(void) fprintf(misout, "%s\n", filename);
	(void) fclose(fp);
    }
    FREE(filename);
    return 0;
}

/* ARGSUSED */
com_best(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    static int bst_count = -1;
    static network_t *bst_network;
    int count = 0;
    node_t *node;
    lsGen gen;

    foreach_node(*network, gen, node)
        if (node->type == INTERNAL)
            count += factor_num_literal(node);

    if (bst_count < 0) {
	bst_network = network_dup(*network);
	bst_count = count;
    }
    else if (count <= bst_count) {
	network_free(bst_network);
	bst_network = network_dup(*network);
	bst_count = count;
    } else {
	network_free(*network);
	*network = network_dup(bst_network);
    }
    return 0;
}

init_command()
{
    char *path;

    command_table = avl_init_table(strcmp);
    flag_table = avl_init_table(strcmp);
    alias_table = avl_init_table(strcmp);

    com_add_command("alias", com_alias, 0);
    com_add_command("echo", com_echo, 0);
    com_add_command("help", com_help, 0);
    com_add_command("quit", com_quit, 0);
    com_add_command("save", com_save_image, 0);
    com_add_command("source", com_source, 1);
    com_add_command("_iloop", com_infinite_loop, 1);
    com_add_command("undo", com_undo, 0);
    com_add_command("set", com_set_variable, 0);
    com_add_command("unalias", com_unalias, 0);
    com_add_command("unset", com_unset_variable, 0);
    com_add_command("time", com_time, 0);
    com_add_command("usage", com_usage, 0);
    com_add_command("_which", com_which, 0);
    com_add_command("_best", com_best, 1);

    /* set the default open_path */
    path = ALLOC(char, strlen(mis_library()) + 10);
    (void) sprintf(path, ".:%s", mis_library());
    (void) avl_insert(flag_table, util_strsav("open_path"), path);
}


end_command()
{
    avl_free_table(flag_table, free, free);
    avl_free_table(command_table, (void (*)()) 0, com_command_free);
    avl_free_table(alias_table, (void (*)()) 0, com_alias_free);
    if (backup_network != NIL(network_t)) {
	network_free(backup_network);
    }
    error_cleanup();
}
