#include "mis.h"
#include "com_int.h"

static int end_loop();

com_source(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, echo, prompt, silent, status, interactive, quit_count, lp_count = -1;
    int lp_lits, lp_time;
    char *prompt_string, *real_filename, line[MAX_STR];
    FILE *fp;

    interactive = silent = prompt = echo = lp_lits = lp_time = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "ipsxlt")) != EOF) {
	switch(c) {
	case 'i':		/* a hack to distinguish EOF from stdin */
	    interactive = 1;
	    break;
	case 'p':
	    prompt = 1;
	    break;
	case 's':
	    silent = 1;
	    break;
	case 'x':
	    echo = 1;
	    break;
	case 'l':
	    lp_lits = 1;
	    break;
	case 't':
	    lp_time = 1;
	    break;
	default:
	    goto usage;
	}
    }

    if (argc - optind != 1) goto usage;

    do {
    lp_count ++; /* increment the loop counter */

    fp = com_open_file(argv[optind], "r", &real_filename, silent);
    if (fp == NULL) {
	FREE(real_filename);
	return ! silent;	/* error return if not silent */
    }

    quit_count = 0;
    do {
	if (prompt) {
	    prompt_string = com_get_flag("prompt");
	    if (prompt_string == NIL(char)) prompt_string = "misII> ";
	    (void) fprintf(stdout, "%s", prompt_string);
	}

	/* clear errors -- e.g., EOF reached from stdin */
	clearerr(fp);

	/* read another command line */
	if (fgets(line, MAX_STR, fp) == NULL) {
	    if (interactive) {
	        if (quit_count++ < 5) {
		    (void) fprintf(miserr, "\nUse \"quit\" to leave misII.\n");
		    continue;
		}
		status = -1;		/* fake a 'quit' */
	    } else {
		status = 0;		/* successful end of 'source' ; loop? */
	    }
	    break;
	}
	quit_count = 0;

	if (echo) {
	    (void) fprintf(misout, "%s", line);
	}

	status = com_execute(network, line);
    } while (status == 0);

    if (fp != stdin) {
	if (status != 0 && status != -1) {
	    (void) fprintf(miserr, "aborting 'source %s'\n", real_filename);
	}
	(void) fclose(fp);
        FREE(real_filename);
    }

    } while ((status == 0) && (lp_lits || lp_time) && ((lp_count <= 0) || !end_loop(network, lp_lits, lp_time)));

    return status;

usage:
    (void) fprintf(miserr, "source [-psx] filename\n");
    (void) fprintf(miserr, "\t-p supply prompt before reading each line\n");
    (void) fprintf(miserr, "\t-s silently ignore nonexistant file\n");
    (void) fprintf(miserr, "\t-x echo each line as it is executed\n");
    (void) fprintf(miserr, "\t-l loop while literal count decreases\n");
    (void) fprintf(miserr, "\t-t loop for amount of time\n");
    return 1;
}

/* ARGSUSED */
static int
end_loop(network, lp_lits, lp_time)
network_t **network;
int lp_lits, lp_time;
{
    static int prev_count = -1;
    static network_t *prev_network;
    int count = 0;
    node_t *node;
    lsGen gen;

    if (lp_lits) {

	foreach_node(*network, gen, node)
	    if (node->type == INTERNAL)
	        count += factor_num_literal(node);
	if (prev_count < 0) {
	    prev_network = network_dup(*network);
	    prev_count = count;
	    return 0;
	}
	else if (count < prev_count) {
	    network_free(prev_network);
	    prev_network = network_dup(*network);
	    prev_count = count;
	    return 0;
	} else {
	    network_free(*network);
	    *network = network_dup(prev_network);
	    return 1;
	}
    }

    return 1;
}
