#include "mis.h"
#include "util.h"
#include "delay_int.h"
#include <math.h>

static int delay_sort_flag;
extern array_t *com_get_true_nodes(); /* Should disappear */

static int
delay_sort(node1_p, node2_p)
char **node1_p, **node2_p;
{
    node_t *node1 = *(node_t **) node1_p;
    node_t *node2 = *(node_t **) node2_p;
    double diff;
    delay_time_t delay1, delay2;

    switch(delay_sort_flag) {
    case 0:
	delay1 = delay_arrival_time(node1);
	delay2 = delay_arrival_time(node2);
	diff = MAX(delay1.rise, delay1.fall) - MAX(delay2.rise, delay2.fall);
	break;
    case 1:
	delay1 = delay_required_time(node1);
	delay2 = delay_required_time(node2);
	diff = MAX(delay2.rise, delay2.fall) - MAX(delay1.rise, delay1.fall);
	break;
    case 2:
	delay1 = delay_slack_time(node1);
	delay2 = delay_slack_time(node2);
	diff = MAX(delay2.rise, delay2.fall) - MAX(delay1.rise, delay1.fall);
	break;
    case 3:
	diff = DELAY(node1)->load - DELAY(node2)->load;
	break;
    default:
	fail("bad delay sort flag");
	/* NOTREACHED */
    }

    if (diff > 0) {
	return -1;
    } else if (diff == 0) {
	return 0;
    } else {
	return 1;
    }
}

#define GET_DEFAULT(parm) delay_get_default_parameter(net, parm)

static int
com_print_delay_constraints(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    int i;
    lsGen gen;
    node_t *fanout, *node;
    network_t *net = *network;

    (void) fprintf(misout, "\t\tA setting of -1.0 means unset\n");
    
    (void) fprintf(misout, "Default settings:\n\t\tinput arrival=(%5.2f %5.2f)\n",
	    GET_DEFAULT(DELAY_DEFAULT_ARRIVAL_RISE),
	    GET_DEFAULT(DELAY_DEFAULT_ARRIVAL_FALL));
    (void) fprintf(misout, "\t\toutput required=(%5.2f %5.2f)\n",
	    GET_DEFAULT(DELAY_DEFAULT_REQUIRED_RISE),
	    GET_DEFAULT(DELAY_DEFAULT_REQUIRED_FALL));
    (void) fprintf(misout, "\t\tinput drive=(%5.2f %5.2f)\n",
	    GET_DEFAULT(DELAY_DEFAULT_DRIVE_RISE),
	    GET_DEFAULT(DELAY_DEFAULT_DRIVE_FALL));
    (void) fprintf(misout, "\t\toutput load=%5.2f\n",
	    GET_DEFAULT(DELAY_DEFAULT_OUTPUT_LOAD));
    delay_print_wire_loads(net, misout);
    
    node_vec = com_get_nodes(net, argc, argv);

    for(i = 0; i < array_n(node_vec); ++i) {
	node = array_fetch(node_t *, node_vec, i);
	if(node_function(node) == NODE_PI) {
	    (void) fprintf(misout,
"Settings for input %s:\tdrive=(%5.2f %5.2f)\tarrival=(%5.2f %5.2f)\n",
		    node_long_name(node),
		    delay_get_parameter(node, DELAY_DRIVE_RISE),
		    delay_get_parameter(node, DELAY_DRIVE_FALL),
		    delay_get_parameter(node, DELAY_ARRIVAL_RISE),
		    delay_get_parameter(node, DELAY_ARRIVAL_FALL));
	} else {
	    foreach_fanout(node, gen, fanout) {
		if(node_function(fanout) == NODE_PO) {
		    (void) fprintf(misout,
"Settings for output %s:\tload= %5.2f\trequired=(%5.2f %5.2f)\n",
			    node_long_name(fanout),
			    delay_get_parameter(fanout, DELAY_OUTPUT_LOAD),
			    delay_get_parameter(fanout, DELAY_REQUIRED_RISE),
			    delay_get_parameter(fanout, DELAY_REQUIRED_FALL));
		}
	    }
	}
    }
    array_free(node_vec);
    return 0;
}
	   
   

static int
com_print_delay(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i, c, j, aflag, rflag, sflag, lflag, some_flag, print_max;
    array_t *node_vec;
    node_t *node;
    delay_time_t arrival, required, slack;
    delay_model_t model;

    aflag = 0;
    rflag = 0;
    sflag = 0;
    lflag = 0;
    some_flag = 0;
    print_max = INFINITY;
    delay_sort_flag = -1;
    model = DELAY_MODEL_LIBRARY;
    getopt_reset();
    while ((c = getopt(argc, argv, "alm:p:rs")) != EOF) {
	switch(c) {
	case 'a':
	    some_flag = 1;
	    aflag = 1;
	    if (delay_sort_flag == -1) delay_sort_flag = 0;
	    break;
	case 'm':
	    if (strcmp(optarg, "unit") == 0) {
		model = DELAY_MODEL_UNIT;
	    } else if (strcmp(optarg, "unit-fanout") == 0) {
		model = DELAY_MODEL_UNIT_FANOUT;
	    } else if (strcmp(optarg, "library") == 0) {
		model = DELAY_MODEL_LIBRARY;
	    } else if (strcmp(optarg, "mapped") == 0) {
		model = DELAY_MODEL_MAPPED;
	    } else {
		goto usage;
	    }
	    break;
	case 'l':
	    some_flag = 1;
	    lflag = 1;
	    if (delay_sort_flag == -1) delay_sort_flag = 3;
	    break;
	case 'p':
	    print_max = atoi(optarg);
	    break;
	case 'r':
	    some_flag = 1;
	    rflag = 1;
	    if (delay_sort_flag == -1) delay_sort_flag = 1;
	    break;
	case 's':
	    some_flag = 1;
	    sflag = 1;
	    if (delay_sort_flag == -1) delay_sort_flag = 2;
	    break;
	default:
	    goto usage;
	}
    }

    node_vec = com_get_true_nodes(*network, argc-optind+1, argv+optind-1);
    if (array_n(node_vec) < 1) goto usage;

    switch (model) {
    case DELAY_MODEL_LIBRARY:
	if (lib_network_is_mapped(*network)) {
	    (void) fprintf(misout, " ... using library delay model\n");
	    if (! delay_trace(*network, DELAY_MODEL_LIBRARY)) {
		(void) fprintf(miserr, "%s", error_string());
		array_free(node_vec);
		return 1;
	    }
	} else {
	    (void) fprintf(miserr, 
		"network not mapped, cannot use library delay model\n");
	    array_free(node_vec);
	    return 1;
	}
	break;
     case DELAY_MODEL_MAPPED:
	if (lib_network_is_mapped(*network)) {
	    (void) fprintf(misout,
			   "...network mapped, using library delay model\n");
	    if (! delay_trace(*network, DELAY_MODEL_LIBRARY)) {
		(void) fprintf(miserr, "%s", error_string());
		array_free(node_vec);
		return 1;
	    }
	} else {
	    (void) fprintf(miserr, 
		"...network not mapped, using mapped delay model\n");
	    if (! delay_trace(*network, DELAY_MODEL_MAPPED)) {
		(void) fprintf(miserr, "%s", error_string());
		array_free(node_vec);
		return 1;
	    }
	}
	break;


    case DELAY_MODEL_UNIT_FANOUT:
	(void) fprintf(misout, " ... using unit delay model (with fanout)\n");
	if (! delay_trace(*network, DELAY_MODEL_UNIT_FANOUT)) {
	    (void) fprintf(miserr, "%s", error_string());
	    array_free(node_vec);
	    return 1;
	}
	break;

    case DELAY_MODEL_UNIT:
	(void) fprintf(misout, " ... using unit delay model (levels)\n");
	if (! delay_trace(*network, DELAY_MODEL_UNIT)) {
	    (void) fprintf(miserr, "%s", error_string());
	    array_free(node_vec);
	    return 1;
	}
	break;

    default:
	array_free(node_vec);
	goto usage;		/* shouldn't happen ? */
    }
	

    if (delay_sort_flag != -1) {
	array_sort(node_vec, delay_sort);
    }

    for(i = 0, j = 0; (j < print_max) && (i < array_n(node_vec)); i++) {
	node = array_fetch(node_t *, node_vec, i);

	/*
	 * Print data on the primary output only once. When multiple
	 * primary outputs share the same internal node, we will print
	 * the primary outputs as well as the internal node
	 */
	if (argc == optind && node_function(node) == NODE_PO &&
		node_num_fanout(node_get_fanin(node,0)) == 1) continue;
	else j++;

	arrival = delay_arrival_time(node);
	required = delay_required_time(node);
	slack = delay_slack_time(node);

	(void) fprintf(misout, "%-10s: ", node_name(node));
	if (! some_flag || aflag) {
	    if (arrival.rise < -(INFINITY/2) || arrival.fall < -(INFINITY/2)){
		(void) fprintf(misout, "arrival=(-INFINITY -INFINITY) ");
	    } else {
		(void) fprintf(misout, "arrival=(%5.2f %5.2f) ", 
		    arrival.rise, arrival.fall);
	    }
	}
	if (! some_flag || rflag) {
	    if (required.rise < -(INFINITY/2) || required.fall < -(INFINITY/2)){
		(void) fprintf(misout, "required=(0 0) ");
	    } else {
	    (void) fprintf(misout, "required=(%5.2f %5.2f) ", 
		required.rise, required.fall);
	    }
	}
	if (! some_flag || sflag) {
	    if (slack.rise  > (INFINITY/2) || slack.fall  > (INFINITY/2)){
		(void) fprintf(misout, "slack=(INFINITY INFINITY) ");
	    } else {
		(void) fprintf(misout, "slack=(%5.2f %5.2f) ", 
		    slack.rise, slack.fall);
	    }
	}
	if (lflag) {
	    (void) fprintf(misout, "load=%6.3f", DELAY(node)->load);
	}
	(void) fprintf(misout, "\n");
    }
    array_free(node_vec);
    return 0;

usage:
    (void) fprintf(miserr, 
	"print_delay [-alrs] [-m model] [-p n] n1 n2 ...\n");
    (void) fprintf(miserr, "    -a\t\tprint arrival times\n");
    (void) fprintf(miserr, "    -l\t\tprint output loads\n");
    (void) fprintf(miserr, "    -r\t\tprint required times\n");
    (void) fprintf(miserr, "    -s\t\tprint slack times\n");
    (void) fprintf(miserr,
    	"    -m model\tchoose delay model (unit, unit-fanout, library, mapped)\n");
    (void) fprintf(miserr, "    -p n\tonly print top 'n' values\n");
    return 1;
}

enum parm_enum {ATIME, RTIME, LOAD, DRIVE, DEFAULT_RTIME, DEFAULT_ATIME,
		    DEFAULT_LOAD, DEFAULT_DRIVE, DEFAULT_WIRE};
#define INPUTS 1
#define OUTPUTS 2

static int
com_set_delay(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i, c, set = 0, nodetype = 0, found = 0;
    array_t *node_vec;
    node_t *node;
    delay_param_t rise_parm, fall_parm;
    double parm_val;
    enum parm_enum set_parameter;

    
    getopt_reset();
    while ((c = getopt(argc, argv, "?a:r:d:l:D:A:R:L:W:S:")) != EOF) {
	switch(c) {
	case 'r':
	    set_parameter = RTIME;
	    ++set;
	    found = 1;
	    nodetype = OUTPUTS;
	    break;
	case 'a':
	    set_parameter = ATIME;
	    ++set;
	    found = 1;
	    nodetype = INPUTS;
	    break;
	case 'd':
	    set_parameter = DRIVE;
	    ++set;
	    found = 1;
	    nodetype = INPUTS;
	    break;
	case 'l':
	    set_parameter = LOAD;
	    ++set;
	    found = 1;
	    nodetype = OUTPUTS;
	    break;
	case 'D':
	case 'A':
	case 'R':
	case 'L':
	case 'W':
	case 'S':
	    if((parm_val = atof(optarg)) == HUGE) goto usage;
	    switch(c) {
	    case 'D':
		delay_set_default_parameter(*network, DELAY_DEFAULT_DRIVE_RISE, parm_val);
		delay_set_default_parameter(*network, DELAY_DEFAULT_DRIVE_FALL, parm_val);
		break;
	    case 'A':
		delay_set_default_parameter(*network, DELAY_DEFAULT_ARRIVAL_RISE, parm_val);
		delay_set_default_parameter(*network, DELAY_DEFAULT_ARRIVAL_FALL, parm_val);
		break;
	    case 'R':
		delay_set_default_parameter(*network, DELAY_DEFAULT_REQUIRED_RISE, parm_val);
		delay_set_default_parameter(*network, DELAY_DEFAULT_REQUIRED_FALL, parm_val);
		break;
	    case 'L':
		delay_set_default_parameter(*network, DELAY_DEFAULT_OUTPUT_LOAD, parm_val);
		break;
	    case 'W':
		delay_set_default_parameter(*network, DELAY_ADD_WIRE_LOAD, parm_val);
		break;
	    case 'S':
		delay_set_default_parameter(*network, DELAY_WIRE_LOAD_SLOPE, parm_val);
		break;
	    }
	    break;
	default:
	    goto usage;
	}

	if(set > 1) goto usage;
	
	if(found)
	    if((parm_val = atof(optarg)) == HUGE) goto usage;
	
	found = 0;
    }

    node_vec = com_get_true_nodes(*network, argc-optind+1, argv+optind-1);

    if(set) {
	if(array_n(node_vec) == 0) goto usage;

	for(i = 0; i < array_n(node_vec); ++i) {
	    node = array_fetch(node_t *, node_vec, i);
	    if(nodetype == INPUTS) {
		if(node_function(node) != NODE_PI) goto usage;
	    } else if(nodetype == OUTPUTS) {
		if(node_function(node) != NODE_PO) goto usage;
	    } else {
		goto usage;
	    }
	}
	
	switch(set_parameter) {
	case LOAD:
	    if(nodetype == INPUTS) goto usage;
	    rise_parm = fall_parm = DELAY_OUTPUT_LOAD;
	    break;
	case DRIVE:
	    if(nodetype == OUTPUTS) goto usage;
	    rise_parm = DELAY_DRIVE_RISE;
	    fall_parm = DELAY_DRIVE_FALL;
	    break;
	case RTIME:
	    if(nodetype == INPUTS) goto usage;
	    rise_parm = DELAY_REQUIRED_RISE;
	    fall_parm = DELAY_REQUIRED_FALL;
	    break;
	case ATIME:
	    if(nodetype == OUTPUTS) goto usage;
	    rise_parm = DELAY_ARRIVAL_RISE;
	    fall_parm = DELAY_ARRIVAL_FALL;
	    break;
	default:
	    abort(); /* not reached */
	}
	

	for(i = 0; i < array_n(node_vec); ++i) {
	    node = array_fetch(node_t *, node_vec, i);
	    delay_set_parameter(node, rise_parm, parm_val);
	    delay_set_parameter(node, fall_parm, parm_val);
	}
    }
	
    array_free(node_vec);
    return 0;

usage:
    (void) fprintf(miserr, 
	"set_delay [-a|r|l|d f]  [-A f] [-R f] [-D f] [-L f] [-W f] [-S f] [o1 o2 ... | i1 i2 ...]\n");
    (void) fprintf(miserr, "    -a f\t\tset arrival times to f\n");
    (void) fprintf(miserr, "    -r f\t\tset required times to f\n");
    (void) fprintf(miserr, "    -l f\t\tset loads on primary outputs to f\n");
    (void) fprintf(miserr, "    -d f\t\tset drives from primary inputs to f\n");
    (void) fprintf(miserr, "    -A f\t\tset default arrival time to f\n");
    (void) fprintf(miserr, "    -R f\t\tset default required time to f\n");
    (void) fprintf(miserr, "    -D f\t\tset default input drive to f\n");
    (void) fprintf(miserr, "    -L f\t\tset default output to f\n");
    (void) fprintf(miserr, "    -W f\t\tSet the next wire load to f\n");
    (void) fprintf(miserr, "    -S f\t\tSet the wire load slope to f\n");
    (void) fprintf(miserr, "specify at most one of a,r,l,d\n");
    (void) fprintf(miserr, "i1...in a vector of primary inputs\n");
    (void) fprintf(miserr, "o1...on a vector of primary outputs\n");
    return 1;
}


init_delay()
{
    com_add_command("print_delay", com_print_delay, /* changes-network */ 0);
    com_add_command("set_delay", com_set_delay, /* changes-network */ 0);
    com_add_command("constraints", com_print_delay_constraints,
		    /* changes-network */ 0);

    node_register_daemon(DAEMON_ALLOC, delay_alloc);
    node_register_daemon(DAEMON_FREE, delay_free);
    node_register_daemon(DAEMON_DUP, delay_dup);
    node_register_daemon(DAEMON_INVALID, delay_invalid);

}


end_delay()
{
}
