#include "mis.h"
#include "util.h"
#include "extract_int.h"


static int
com_cube_extract(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    lsGen gen;
    sm_matrix *cube_literal_matrix;
    int thres, use_best_subcube, c;
    node_t *node;

    use_best_subcube = 0;
    use_complement = 0;		/* global */
    cube_extract_debug = 0;	/* global */
    thres = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "bcdft:v:")) != EOF) {
	switch(c) {
	case 'b':
	    use_best_subcube = 1;
	    break;
	case 'c':
	    use_complement = 1;
	    break;
	case 'd':
	    cube_extract_debug = 1;
	    break;
	case 'f':
	    use_best_subcube = 2;
	    break;
	case 't':
	    thres = atoi(optarg);
	    break;
	case 'v':
	    cube_extract_debug = atoi(optarg);
	    break;
	default:
	    goto usage;
	}
    }
    if (argc - optind != 0) goto usage;
		
    ex_setup_globals(*network, 0);

    /* create the cube-literal matrix */
    cube_literal_matrix = sm_alloc();
    foreach_node(*network, gen, node) {
	switch(node_function(node)) {
	case NODE_PI: 
	case NODE_PO:
	case NODE_0: 
	case NODE_1:
	case NODE_INV:
	case NODE_BUF:
	    break;

	case NODE_AND:
	case NODE_OR:
	case NODE_COMPLEX:
	    ex_node_to_sm(node, cube_literal_matrix);
	    break;
	}
    }

    /* extract cubes */
    (void) sparse_cube_extract(cube_literal_matrix, thres, use_best_subcube);

    free_value_cells(cube_literal_matrix);
    sm_free(cube_literal_matrix);
    ex_free_globals(0);
    return 0;

usage:
    (void) fprintf(miserr, 
	"usage: gcx [-bcdf] [-t thresh]\n");
    (void) fprintf(miserr, 
	"    -b\t\tuses best subcube (rather than pingpong)\n");
    (void) fprintf(miserr, 
	"    -c\t\tuses complement during division\n");
    (void) fprintf(miserr, 
	"    -d\t\tenables debug\n");
    (void) fprintf(miserr, 
	"    -f\t\tuses factored literals for value\n");
    (void) fprintf(miserr, 
	"    -t thres\tset threshold value for extraction\n");
    return 1;
}

static int
com_kernel_extract(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    sm_matrix *func;
    node_t *node;
    int use_best_subkernel, use_all_kernels, use_overlap, one_pass;
    int need_duplicate, value, thres, c, index;
    lsGen gen;

    one_pass = 0;
    use_all_kernels = 0;
    use_best_subkernel = 0;
    use_complement = 0;		/* global */
    kernel_extract_debug = 0;	/* global */
    use_overlap = 0;
    need_duplicate = 0;
    thres = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "1abcdfot:v:")) != EOF) {
	switch(c) {
	case '1':
	    one_pass = 1;
	    break;
	case 'a':
	    use_all_kernels = 1;
	    break;
	case 'b':
	    use_best_subkernel = 1;
	    break;
	case 'c':
	    use_complement = 1;
	    break;
	case 'd':
	    kernel_extract_debug = 1;
	    break;
	case 'f':
	    use_best_subkernel = 2;
	    break;
	case 'o':
	    use_overlap = 1;
	    need_duplicate = 1;
	    break;
	case 't':
	    thres = atoi(optarg);
	    break;
	case 'v':
	    kernel_extract_debug = atoi(optarg);
	    break;
	default:
	    goto usage;
	}
    }

    if (argc - optind != 0) goto usage;
		
    do {
	ex_setup_globals(*network, need_duplicate);

	kernel_extract_init();
	foreach_node(*network, gen, node) {
	    switch(node_function(node)) {
	    case NODE_PI:
	    case NODE_PO:
	    case NODE_0:
	    case NODE_1:
	    case NODE_INV:
	    case NODE_BUF:
		break;

	    case NODE_AND:
	    case NODE_OR:
	    case NODE_COMPLEX:
		func = sm_alloc();
		ex_node_to_sm(node, func);
		index = nodeindex_indexof(global_node_index, node);
		kernel_extract(func, index, use_all_kernels);
		free_value_cells(func);
		sm_free(func);
		break;
	    }
	}

	if (use_overlap) {
	    value = overlapping_kernel_extract(thres, use_best_subkernel);
	} else {
	    value = sparse_kernel_extract(thres, use_best_subkernel);
	}
	kernel_extract_free();

	ex_free_globals(need_duplicate);
    } while (! one_pass && value > 0);

    return 0;

usage:
    (void) fprintf(miserr, 
	"usage: gkx [-1abcdfo] [-t thresh]\n");
    (void) fprintf(miserr, 
	"    -1\t\tone pass over network\n");
    (void) fprintf(miserr, 
	"    -a\t\tuse all kernels (rather than level 0)\n");
    (void) fprintf(miserr, 
	"    -b\t\tuse best subkernel (rather than pingpong)\n");
    (void) fprintf(miserr, 
	"    -c\t\tuse complement during division\n");
    (void) fprintf(miserr, 
	"    -d\t\tenables debug\n");
    (void) fprintf(miserr, 
	"    -f\t\tuse factored literals for value\n");
    (void) fprintf(miserr, 
	"    -o\t\tallow overlapping factors\n");
    (void) fprintf(miserr, 
	"    -t n\tset threshold value for extraction\n");
    return 1;
}

/* ARGSUSED */
static int
print_kernel(kernel, co_kernel, state)
node_t *kernel;
node_t *co_kernel;
char *state;
{
    (void) fprintf(misout, "(");
    node_print_rhs(misout, co_kernel);
    (void) fprintf(misout, ") * (");
    node_print_rhs(misout, kernel);
    (void) fprintf(misout, ")\n");
    node_free(kernel);
    node_free(co_kernel);
    return 1;
}


static int
com_print_kernel(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    node_t *node;
    array_t *node_vec;
    int c, i, subkernel, subkernel_level;

    subkernel = 0;
    subkernel_level = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "as")) != EOF) {
	switch(c) {
	case 'a':
	    subkernel_level = 1;
	    break;
	case 's':
	    subkernel = 1;
	    break;
	default:
	    goto usage;
	}
    }

    node_vec = com_get_nodes(*network, argc-optind+1, argv+optind-1);

    for(i = 0; i < array_n(node_vec); i++) {
	node = array_fetch(node_t *, node_vec, i);
	if (node->type == PRIMARY_INPUT) continue;
	if (node->type == PRIMARY_OUTPUT) continue;
	if (subkernel) {
	    (void) fprintf(misout, "Subkernels of %s\n", node_name(node));
	    ex_subkernel_gen(node, print_kernel, subkernel_level, NIL(char));
	} else {
	    (void) fprintf(misout, "Kernels of %s\n", node_name(node));
	    ex_kernel_gen(node, print_kernel, NIL(char));
	}
    }
    array_free(node_vec);
    return 0;

usage:
    (void) fprintf(miserr, "usage: print_kernel [-as] n1 n2 ...\n");
    (void) fprintf(miserr, 
	"    -a\t\tgenerate kernels of all levels\n");
    (void) fprintf(miserr, 
	"    -s\t\tgenerate subkernels rather than kernels\n");
    return 1;
}

static int
com_gdiv(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    node_t *node, *divisor;
    array_t *node_vec;
    int i;

    node_vec = com_get_nodes(*network, argc, argv);
    if (array_n(node_vec) < 1) goto usage;

    for(i = 0; i < array_n(node_vec); i++) {
	node = array_fetch(node_t *, node_vec, i);
	if (node->type == PRIMARY_INPUT) continue;
	if (node->type == PRIMARY_OUTPUT) continue;
	if (strcmp(argv[0], "_gdiv") == 0) {
	    (void) fprintf(misout, "good divisor of %s is: ", node_name(node));
	    divisor = ex_find_divisor(node, 0, 1);
	} else {
	    (void) fprintf(misout, "quick divisor of %s is: ", node_name(node));
	    divisor = ex_find_divisor_quick(node);
	}
	if (divisor == 0) {
	    (void) fprintf(misout, " (no multiple-cube divisors)\n");
	} else {
	    node_print_rhs(misout, divisor);
	    (void) fprintf(misout, "\n");
	    node_free(divisor);
	}
    }
    array_free(node_vec);
    return 0;

usage:
    (void) fprintf(miserr, "usage: %s n1 n2 ...\n", argv[0]);
    return 1;
}

init_extract()
{
    com_add_command("gkx", com_kernel_extract, 1);
    com_add_command("gcx", com_cube_extract, 1);
    com_add_command("print_kernel", com_print_kernel, 1);
    com_add_command("_gdiv", com_gdiv, 0);
    com_add_command("_qdiv", com_gdiv, 0);
}

end_extract()
{
    sm_cleanup();
}
