#include "genlib_int.h"

static int check_internal_phase();
static int set_functions();


static int 
leaf_compare(t1, t2)
char **t1, **t2;
{
    tree_node_t *tree1 = *(tree_node_t **) t1;
    tree_node_t *tree2 = *(tree_node_t **) t2;

    return strcmp(tree1->name, tree2->name);
}



int
gl_convert_gate_to_blif(name, area, function, pins, use_nor)
char *name;
double area;
function_t *function;
pin_info_t *pins;
int use_nor;
{
    pin_info_t *pin, *pinnext;
    tree_node_t **list, **leafs;
    int i, j, nlist, nleafs;
    FILE *fpout = gl_out_file;

    if (! check_internal_phase(function->tree, 0)) {
	return 0;
    }

    if (function->tree->type == ZERO_NODE || function->tree->type == ONE_NODE) {
	/* special case for constants */
	nlist = 1;
	list = ALLOC(tree_node_t *, 1);
	list[0] = gl_dup_tree(function->tree);

    } else {
	/* make the tree strictly alternate AND/OR */
	gl_make_well_formed(function->tree);

	/* generate all nand or nor gate forms */
	nlist = gl_nand_gate_forms(function->tree, &list, use_nor);

	/* set all tree nodes to a specific type */
	for(i = 0; i < nlist; i++) {
	    set_functions(list[i], use_nor);
	}
    }

    /* get the leaf names */
    nleafs = gl_get_unique_leaf_pointers(function->tree, &leafs);
    qsort((char *) leafs, nleafs, sizeof(tree_node_t *), leaf_compare);

    for(i = 0; i < nlist; i++) {
	gl_assign_node_names(list[i]);
	FREE(list[i]->name);
	list[i]->name = util_strsav(function->name);

	(void) fprintf(fpout, "# (%d of %d) %s-FORM of ", 
			    i+1, nlist, use_nor ? "NOR" : "NAND");
	gl_print_tree(fpout, function->tree);
	(void) fprintf(fpout, "\n");
	(void) fprintf(fpout, ".model %s\n", name);

	gl_write_blif(fpout, list[i], use_nor);

	(void) fprintf(fpout, ".area %3.2f\n", area);

	for(pin = pins; pin != 0; pin = pin->next) {
	    if (strcmp(pin->name, "*") == 0) {
		if (pin != pins || pin->next != 0) {	/* only 1 allowed */
		    read_error("improper use of pin wildcard '*'");
		    return 0;
		}
		for(j = 0; j < nleafs; j++) {
		    (void) fprintf(fpout,
			".delay %s %s %4.3f %4.3f %4.3f %4.3f %4.3f %4.3f\n",
			leafs[j]->name, pin->phase, pin->value[0], 
			pin->value[1], pin->value[2], pin->value[3],
			pin->value[4], pin->value[5]);
		}
	    } else {
		for(j = 0; j < nleafs; j++) {
		    if (strcmp(pin->name, leafs[j]->name) == 0) {
			break;
		    }
		}
		if (j == nleafs) {
		    read_error("pin (in delay info) not found");
		    return 0;
		}
		(void) fprintf(fpout,
		    ".delay %s %s %4.3f %4.3f %4.3f %4.3f %4.3f %4.3f\n",
		    pin->name, pin->phase, pin->value[0], 
		    pin->value[1], pin->value[2], pin->value[3],
		    pin->value[4], pin->value[5]);
	    }
	}
	(void) fprintf(fpout, ".end\n\n");

	gl_free_tree(list[i]);
    }
    FREE(list);
    FREE(leafs);

    /* free memory used */
    for(pin = pins; pin != 0; pin = pinnext) {
	pinnext = pin->next;
	FREE(pin->name);
	FREE(pin->phase);
	FREE(pin);
    }
    gl_free_tree(function->tree);
    FREE(function->name);
    FREE(function);
    FREE(name);
    return 1;
}

static int
check_internal_phase(tree, level)
tree_node_t *tree;
int level;
{
    int i;

    if (tree->nsons != 0) {
	for(i = 0; i < tree->nsons; i++) {
	    if (! check_internal_phase(tree->sons[i], level+1)) {
		return 0;	/* error in some child */
	    }
	}
	/* if (level > 0 && tree->phase != 1) {
	 *   read_error("inversion allowed only at root and leaves of function");
	 *   return 0;
	 * }
	 */
	if (level > 0 && (tree->type == ZERO_NODE || tree->type == ONE_NODE)) {
	    read_error("0 and 1 are allowed only at the root of a function");
	    return 0;
	}
	    
    }
    return 1;
}


static int
set_functions(tree, use_nor)
tree_node_t *tree;
int use_nor;
{
    int i;

    for(i = 0; i < tree->nsons; i++) {
	set_functions(tree->sons[i], use_nor);
    }
    tree->type = use_nor ? NOR_NODE : NAND_NODE;
}
