#include "port.h"
#include "mis.h"
#include "graphics.h"

/*
 * Converted to X11 by Rick L. Spickelmier
 */

typedef enum plot_action_enum plot_action_t;
enum plot_action_enum {
    PLOT_PLOT, PLOT_KILL
};

static st_table *table; 
void     	plot_network();

/* ARGSUSED */
int
com_plot(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    plot_action_t action;
    int c;
    char name[40];

    action = PLOT_PLOT;
    (void) strcpy(name, network_name(*network));
    getopt_reset();
    while ((c = getopt(argc, argv, "n:k")) != EOF) {
	switch (c) { 
	case 'n':
	    (void) strcpy(name, optarg);
	    break;
	case 'k':
	    action = PLOT_KILL;
	    break;
	default:
	    (void) fprintf(misout, "usage: plot [-n name] [-k]\n");
	    (void) fprintf(misout, "    -n\tname\tDisplay name\n");
	    (void) fprintf(misout, "    -k\t\tKill the display\n");
	    return 1;
	}
    }

    if (argc - optind > 0) {
	(void) fprintf(misout, "usage: plot [-n name] [-k]\n");
	return 1;
    }

    plot_network(*network, name, action);
    return 0;
}

init_graphics()
{
    table = st_init_table(strcmp, st_strhash);

    com_add_command("plot", com_plot, 0);
}

end_graphics()
{
    char *d1;
    char *d2;
    st_generator *gen;
    plot_t *ploter;

    st_foreach_item(table, gen, &d1, &d2) {
	ploter = (plot_t *) d2;
	FREE(d1);
	graphics_kill(ploter);
    }
    st_free_table(table);
}

void
plot_network(network, name, action)
network_t *network;
char *name;
plot_action_t action;
{
    plot_t *ploter;
    char *dummy, pname[40], *sname;
    int num;
    lsGen gen;
    node_t *np;

    if (action == PLOT_KILL) {
	if (st_delete(table, &name, &dummy)) {
	    ploter = (plot_t *) dummy; 
	    FREE(name);
	    graphics_kill(ploter);
	} else {
	    (void) fprintf(misout, "%s: No window with this name!\n", name);
	}
	return;
    }

    num = 0;
    foreach_node(network, gen, np) {
	if (np->type == INTERNAL) {
	    num += factor_num_literal(np);
	}
    }
    (void) sprintf(pname, "%s:  %d  %d", name, network_num_internal(network), num);
		      
    if (st_lookup(table, name, &dummy)) {
	ploter = (plot_t *) dummy; 
	sname = name;
    } else {
	ploter = NIL(plot_t); 
	sname = util_strsav(name);
    }

    if ((ploter = graphics_plot(ploter, network, pname,
	                        NIL(st_table), NIL(st_table), NIL(st_table),
				NIL(st_table), NIL(st_table), NIL(st_table),
				NIL(st_table), NIL(st_table))) != NIL(plot_t)) {
	(void) st_insert(table, sname, (char *) ploter);
    }
}

