/* RCS Info: $Revision: 1.3 $ on $Date: 89/06/01 16:35:11 $
 *           $Source: /octtools/release-3.0/src/misII/misII/graphics/RCS/draw.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "graphics_int.h"

static void find_limits();

#ifndef NO_X

dg_plot *
dg_draw(window, graph, type, arg)
    dg_plot *window;
    dg_graph *graph;
    dg_outtype type;
    char *arg;
{
    int maxlev, minlev, maxvert, minvert;
    dg_plot *win;
    lsGen gen;
    dg_arc *arc;
    dg_node *node;

    dg_level(graph);
    dg_make_uniform(graph);
    dg_place(graph);

    dg_setdev(type);

    find_limits(graph, &maxlev, &minlev, &maxvert, &minvert);
    maxvert += (maxvert - minvert) / 12;
    minvert -= (maxvert - minvert) / 12;

    win = dg_open(window, arg, - maxlev, minvert, - minlev, maxvert,
            graph->title);

    if (win == (dg_plot *) 0) {
	return (dg_plot *) 0;
    }

    dg_foreach_node(graph, gen, node)
        dg_calc_sizes(win, node->shape, node->font, 
            node->numlines, node->lines, &node->width, &node->height);

    dg_foreach_arc(graph, gen, arc) {
        if (arc->from->type == DG_DUMMY)
            continue;
        dg_draw_arc(win, - arc->from->level, 
                arc->from->vertpos,
                - arc->to->level, arc->to->vertpos,
                arc->color, arc->style, arc->label,
                (arc->to->type != DG_DUMMY),
                (arc->to->type == DG_DUMMY),
                arc->from->shape, arc->from->width,
                arc->from->height, arc->to->shape,
                arc->to->width, arc->to->height);
        while (arc->to->type == DG_DUMMY) {
            assert(num_fanouts(arc->to) == 1);
            (void) lsFirstItem(arc->to->fanouts, (lsGeneric *) &arc, LS_NH);
            dg_draw_arc(win, - arc->from->level, 
                arc->from->vertpos,
                - arc->to->level, arc->to->vertpos,
                arc->color, arc->style, (char *) NULL,
                (arc->to->type != DG_DUMMY),
                (arc->to->type == DG_DUMMY),
                arc->from->shape, arc->from->width,
                arc->from->height, arc->to->shape,
                arc->to->width, arc->to->height);
        }
    }

    dg_foreach_node(graph, gen, node) {
        dg_draw_node(win, - node->level, 
                node->vertpos, node->shape,
                node->color, node->font, node->numlines,
                node->lines, node->width, node->height);
    }

    dg_update(win);

    return ((char *) win);
}

void
dg_kill(win)
    char *win;
{
    dg_close(win);
    return;
}

static void
find_limits(graph, maxlev, minlev, maxvert, minvert)
    dg_graph *graph;
    int *maxlev, *minlev;
    int *maxvert, *minvert;
{
    int first = FALSE;
    lsGen gen;
    dg_node *node;

    dg_foreach_node(graph, gen, node) {
        if (!first) {
            *maxlev = node->level;
            *minlev = node->level;
            *maxvert = node->vertpos + node->grids_high / 2;
            *minvert = node->vertpos - node->grids_high / 2;
            first = TRUE;
        } else {
            if (*maxlev < node->level)
                *maxlev = node->level;
            if (*minlev > node->level)
                *minlev = node->level;
            if (*maxvert < node->vertpos + node->grids_high / 2)
                *maxvert = node->vertpos + node->grids_high / 2;
            if (*minvert > node->vertpos - node->grids_high / 2)
                *minvert = node->vertpos - node->grids_high / 2;
        }
    }
    return;
}

#else /* NO_X */

/* ARGSUSED */
dg_plot *
dg_draw(window, graph, type, arg)
    dg_plot *window;
    dg_graph *graph;
    dg_outtype type;
    char *arg;
{
    (void) fprintf(stderr, "Error: dg_draw(): no graphics available\n");
}

/* ARGSUSED */
void
dg_kill(win)
    char *win;
{
    (void) fprintf(stderr, "Error: dg_kill(): no graphics available\n");
}

#endif /* NO_X */
