#include "mis.h"
#include "graphics_int.h"

/*
 * Converted to X11 by Rick L. Spickelmier
 */

#ifndef NO_X

plot_t *
graphics_plot(window, network, label, colors, fonts, shapes, names, arccolors,
        arcstyles, arcnames, drawit)
    plot_t *window;
    network_t *network;
    char *label;
    st_table *colors;
    st_table *fonts;
    st_table *shapes;
    st_table *names;
    st_table *arccolors;
    st_table *arcstyles;
    st_table *arcnames;
    st_table *drawit;
{
    node_t *node, *fanout;
    char buf[512];
    lsGen gen, gen2;
    dg_graph *graph;
    dg_node *gn, *gn1;
    dg_arc *arc;
    st_table *tab;
    dg_nodeshape shape;
    dg_linestyle style;
    dg_nodetype type;
    char *color, *font, *text;
    char *arccolor, *arctext;
    int numlines;
    char *s;
    char **lines;

    tab = st_init_table(st_ptrcmp, st_ptrhash);

    graph = dg_alloc(label ? label : network_name(network));

    foreach_node(network, gen, node) {
        if (drawit && !st_lookup(drawit, (char *) node, &s))
            continue;
        gn = dg_create_node(graph);
        switch (node_function(node)) {
            case NODE_0:
            case NODE_1:
            text = (node_function(node) == NODE_0) ? "0" : "1";
            type = DG_INPUT;
            color = DEF_CONSTANT_COLOR;
            shape = DEF_CONSTANT_SHAPE;
            font = DEF_CONSTANT_FONT;
            break;

            case NODE_PI:
            text = node_name(node);
            type = DG_INPUT;
            color = DEF_INPUT_COLOR;
            shape = DEF_INPUT_SHAPE;
            font = DEF_INPUT_FONT;
            break;

            case NODE_INV:
            case NODE_BUF:
            case NODE_AND:
            case NODE_OR:
            case NODE_COMPLEX:
            text = node_name(node);
            type = DG_INTERNAL;
            color = DEF_FUNCTION_COLOR;
            shape = DEF_FUNCTION_SHAPE;
            font = DEF_FUNCTION_FONT;
            break;


            case NODE_PO:
            text = node_name(node);
            type = DG_OUTPUT;
            color = DEF_OUTPUT_COLOR;
            shape = DEF_OUTPUT_SHAPE;
            font = DEF_OUTPUT_FONT;
            break;
        }

        if (colors)
            (void) st_lookup(colors, (char *) node, &color);
        if (fonts)
            (void) st_lookup(fonts, (char *) node, &font);
        if (shapes)
            (void) st_lookup(shapes, (char *) node,
                    (char **) &shape);
        if (names)
            (void) st_lookup(names, (char *) node, &text);
        
        lines = makeLines(text, &numlines);

        dg_set_node_props(gn, type, color, shape, font, numlines,
                lines);

        freeLines(lines, numlines);

        (void) st_insert(tab, (char *) node, (char *) gn);
    }

    /* Now toss in all the arcs. */
    foreach_node(network, gen, node) {
        if (!st_lookup(tab, (char *) node, (char **) &gn1))
            continue;
        foreach_fanout(node, gen2, fanout) {
            if (!st_lookup(tab, (char *) fanout, (char **) &gn))
                continue;
            arc = dg_create_arc(gn1, gn);

            arccolor = DEF_ARC_COLOR;
            style = DEF_ARC_STYLE;
            arctext = DEF_ARC_LABEL;
            (void) sprintf(buf, "%s %s", node_name(node),
                    node_name(fanout));
            if (arccolors)
                (void) st_lookup(arccolors, buf, &arccolor);
            if (arcstyles)
                (void) st_lookup(arccolors, buf,
                        (char **) &style);
            if (arcnames)
                (void) st_lookup(arcnames, buf, &arctext);
            
            dg_set_arc_props(arc, arccolor, style, arctext);
        }
    }

    return (dg_draw(window, graph, DG_XWINDOW, NIL(char)));
}

void
graphics_kill(win)
    plot_t *win;
{
    dg_kill((dg_plot *) win);
    return;
}

#else /* NO_X */

plot_t *
graphics_plot(window, network, label, colors, fonts, shapes, names, arccolors,
        arcstyles, arcnames, drawit)
    plot_t *window;
    network_t *network;
    char *label;
    st_table *colors;
    st_table *fonts;
    st_table *shapes;
    st_table *names;
    st_table *arccolors;
    st_table *arcstyles;
    st_table *arcnames;
    st_table *drawit;
{
    (void) fprintf(miserr, "Error: graphics_plot(): no graphics available\n");

    return (NULL);
}

void
graphics_kill(win)
    plot_t *win;
{
    (void) fprintf(miserr, "Error: graphics_kill(): no graphics available\n");

    return;
}

#endif /* NO_X */

