/* RCS Info: $Revision: 1.2 $ on $Date: 89/06/01 16:35:22 $
 *           $Source: /octtools/release-3.0/src/misII/misII/graphics/RCS/level.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "graphics_int.h"

extern void assign_levels();

/* Level the graph.  Outputs are level 0. */

void
dg_level(graph)
    dg_graph *graph;
{
    lsGen gen;
    dg_node *node;

    dg_foreach_node(graph, gen, node) {
        node->level = -1;

        if (node->type == DG_UNKNOWN) {
            if (num_fanouts(node) == 0)
                node->type = DG_OUTPUT;
            else if (num_fanins(node) == 0)
                node->type = DG_INPUT;
            else
                node->type = DG_INTERNAL;
        } else {
            if ((node->type == DG_INPUT) &&
                    (num_fanins(node) != 0)) {
                (void) fprintf(stderr, "Warning: node not an input\n");
            } else if ((node->type == DG_OUTPUT) && 
                    (num_fanouts(node) != 0)) {
                (void) fprintf(stderr, "Warning: node not an output\n");
            }
        }
    }

    dg_foreach_node(graph, gen, node)
        if (num_fanouts(node) == 0) {
            node->level = 0;
            assign_levels(node);
        }

    return;
}

static void
assign_levels(node)
    dg_node *node;
{
    lsGen gen;
    dg_arc *arc;
    dg_node *fanin;

    dg_foreach_fanin(node, gen, arc) {
        fanin = arc->from;
        if (fanin->level <= node->level)
            fanin->level = node->level + 1;
        assign_levels(fanin);
    }

    return;
}
