/* RCS Info: $Revision: 1.2 $ on $Date: 89/06/01 16:35:33 $
 *           $Source: /octtools/release-3.0/src/misII/misII/graphics/RCS/solvemat.c,v $
 * Copyright (c) 1987 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * This program calculates the coefficients used for the linear multistep
 * method of a given order.
 *
 * Converted to X11 by  Rick L. Spickelmier
 */

#define u3b     0
#define u3b5    0
#define gcos    0
#define gould   0
#define pdp11   0
#define u370    0

#include <math.h>
#include "port.h"
#include "utility.h"

#define MINVAL    1e-20

dg_solvemat(matrix, vector, size)
    double **matrix;
    double *vector;
    int size;
{
    int i, j, k;
    int *order;
    double mult;
    double *rhs;

/*
(void) printf("matrix:\n");
for (i = 0; i < size; i++) {
for (j = 0; j < size; j++)
(void) printf("%lg ", matrix[i][j]);
(void) printf("\n");
}
*/
    order = ALLOC(int, size);
    for (i = 0; i < size; i++)
        order[i] = i;

    /* Forward substitution. */
    for (i = 0; i < size; i++) {
        /* Check matrix[i][i] to make sure it's an ok pivot. */
        if (fabs(matrix[i][i]) < MINVAL) {
            for (j = i + 1; j < size; j++)
                if (matrix[i][j] > MINVAL)
                    break;
            if (j == size) {
                (void) printf("help, can't solve matrix\n");
                return;
            }

            /* Swap columns i and j... */
            k = order[i];
            order[i] = order[j];
            order[j] = k;

            for (k = 0; k < size; k++) {
                mult = matrix[k][i];
                matrix[k][i] = matrix[k][j];
                matrix[k][j] = mult;
            }
        }

        mult = matrix[i][i];
        for (k = 0; k < size; k++)
            matrix[i][k] /= mult;
        vector[i] /= mult;

        for (j = i + 1; j < size; j++) {
            mult = matrix[j][i];
            for (k = i; k < size; k++)
                matrix[j][k] -= mult * matrix[i][k];
            vector[j] -= mult * vector[i];
        }
    }

    /* Back substitution. */
    for (i = size - 1; i > 0; i--) {
        for (j = i - 1; j >= 0; j--) {
            vector[j] -= matrix[j][i] * vector[i];
            matrix[j][i] = 0;
        }
    }

    /* All done, now put the RHS back in order... */
    rhs = ALLOC(double, size);
    for (i = 0; i < size; i++)
        rhs[i] = vector[i];
    for (i = 0; i < size; i++)
        vector[order[i]] = rhs[i];
    
    FREE(rhs);
    FREE(order);
    
    /* Hmm. */
    return;
}

