#include "mis.h"
#include "io_int.h"

#define MAX_WORD 1024
#define GETC(c, fp)         \
    (c = getc(fp),          \
    c == '\n' ? read_lineno++ :    \
    (c == '\\' ?            \
    (c = getc(fp), c == '\n' ? (c = ' ', read_lineno++) : 0) : 0), \
     c)

static int
com_read(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i, c, append, single_output, status;
    char *fname, *error, *real_filename;
    network_t *new_network;
    FILE *fp;

    append = 0;
    single_output = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "as")) != EOF) {
    switch(c) {
    case 'a':
        append = 1;
        break;
    case 's':
        single_output = 1;
        break;
    default:
        goto usage;
    }
    }

    if (argc - optind == 0) {
    fname = "-";
    } else if (argc - optind == 1) {
    fname = argv[optind];
    } else {
    goto usage;
    }

    fp = com_open_file(fname, "r", &real_filename, /* silent */ 0);
    if (fp == NULL) {
    FREE(real_filename);
    return 1;
    }

    error_init();
    read_register_filename(real_filename);
    new_network = 0;

    if (strcmp(argv[0], "read_blif") == 0) {
    i = read_blif(fp, &new_network);
    if (i == EOF) {
        (void) fprintf(miserr, "read_blif: no network found\n");
        new_network = 0;
    } else if (i != 1) {
        new_network = 0;
    }

    } else if (strcmp(argv[0], "read_eqn") == 0) {
    new_network = read_eqn(fp);

    } else if (strcmp(argv[0], "read_pla") == 0) {
    new_network = mis_read_pla(fp, single_output);
    if (new_network == 0) {
        (void) fprintf(miserr, "read_pla: error reading PLA\n");
    }

    } else {
    fail("com_read: called with bad argv[0]");
    }

    error = error_string();
    if (error[0] != '\0') {
    (void) fprintf(miserr, "%s", error);
    }
    if (new_network != 0){
        if (strcmp(argv[0], "read_blif") == 0) {
        read_dc_blif(fp, &new_network);
        }
    }

    if (new_network != 0) {
    if (append) {
        error_init();
        if (! network_append(*network, new_network)) {
        (void) fprintf(miserr, "%s", error_string());
        }
        network_free(new_network);
    } else {
        network_free(*network);
        *network = new_network;
    }
    network_reset_short_name(*network);
    status = 0;
    } else {
    status = 1;
    }

    FREE(real_filename);
    if (fp != stdin) (void) fclose(fp);
    return status;

usage:
    (void) fprintf(miserr, "usage: %s [-a] [file]\n", argv[0]);
    (void) fprintf(miserr, "    -a\t\tappend to current network\n");
    if (strcmp(argv[0], "read_pla") == 0) {
    (void) fprintf(miserr, "    -s\t\tread PLA as single-output\n");
    }
    return 1;
}


static int
com_write(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, short_name, net_list, status;
    FILE *fp;
    network_t *dc_network;

    short_name = 0;
    net_list = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "sn")) != EOF) {
    switch(c) {
    case 's':
        short_name = 1;
        break;
    case 'n':
        net_list = 1;
        break;
    default:
        goto usage;
    }
    }

    /* open the destination file */
    if (argc - optind == 0) {
    fp = com_open_file("-", "w", NIL(char *), /* silent */0);
    } else if (argc - optind == 1) {
    fp = com_open_file(argv[optind], "w", NIL(char *), /* silent */0);
    } else {
    goto usage;
    }
    if (fp == NULL) return 1;

    /* allow error reporting */
    error_init();
    dc_network= network_dc_network(*network);

    if (strcmp(argv[0], "write_blif") == 0) {
    write_blif(fp, *network, short_name, net_list);
    if (dc_network != NIL (network_t)){
        io_fputc_break(fp, '\n');
        io_fputs_break(fp, ".exdc ");  
        io_fputc_break(fp, '\n');
        write_blif(fp, dc_network, short_name, net_list);
        io_fputc_break(fp, '\n');
    }
    status = 1;
    } else if (strcmp(argv[0], "write_lif") == 0) {
    write_lif(fp, *network, short_name);
    if (dc_network != NIL (network_t)){
        io_fputc_break(fp, '\n');
        io_fputs_break(fp, ".exdc ");  
        io_fputc_break(fp, '\n');
        write_lif(fp, dc_network, short_name);
        io_fputc_break(fp, '\n');
    }
    status = 1;
    } else if (strcmp(argv[0], "write_eqn") == 0) {
    write_eqn(fp, *network, short_name);
    if (dc_network != NIL (network_t)){
        io_fputc_break(fp, '\n');
        io_fputs_break(fp, ".exdc ");  
        io_fputc_break(fp, '\n');
        write_eqn(fp, dc_network, short_name);
        io_fputc_break(fp, '\n');
    }
    status = 1;
    } else if (strcmp(argv[0], "write_pla") == 0) {
    write_pla(fp, *network);
    status = 1;
    } else if (strcmp(argv[0], "write_bdnet") == 0) {
    status = write_bdnet(fp, *network);
    } else {
    (void) fprintf(miserr, " ... how did I get here ?\n");
    return 1;
    }

    if (fp != stdout) {
    (void) fclose(fp);
    }

    if (! status) {
    (void) fprintf(miserr, "%s", error_string());
    return 1;
    }
    return 0;

usage:
    (void) fprintf(miserr, "usage: %s [-s] [filename]\n", argv[0]);
    (void) fprintf(miserr, "    -s\t\tuse network short-names\n");
    if (strcmp(argv[0], "write_blif") == 0) {
    (void) fprintf(miserr, "    -n\t\tuse net-list format when possible\n");
    }
    return 1;
}

static int
com_print(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    network_t *net;
    node_t *node;
    array_t *node_vec;
    int c, i, negative, print_dc;

    negative = 0;
    print_dc = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "nd")) != EOF) {
    switch(c) {
    case 'n':
        negative = 1;
        break;
    case 'd':
        print_dc = 1;
        break;
    default:
        goto usage;
    }
    }

    if (print_dc) {
    net = network_dc_network(*network);
    if (net == NIL(network_t)) {
        return 0;
    }
    }
    else 
    net = *network;
    node_vec = com_get_nodes(net, argc-optind+1, argv+optind-1);
    if (array_n(node_vec) < 1) goto usage;

    for(i = 0; i < array_n(node_vec); i++) {
    node = array_fetch(node_t *, node_vec, i);
    if (negative) {
        node_print_negative(misout, node);
    } else {
        node_print(misout, node);
    }
    }
    array_free(node_vec);
    return 0;

usage:
    (void) fprintf(miserr, "usage: print [-nd] n1 n2 ...\n");
    return 1;
}

/* ARGSUSED */
com_print_stats(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    network_t *net;
    node_t *p, *node;
    int i, c, pi, po, nodes, lits, lits_fac, factored_lits, print_dc;
    lsGen gen;
    array_t *node_vec;

    factored_lits = 0;
    print_dc = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "fd")) != EOF) {
    switch(c) {
    case 'f':
        factored_lits = 1;
        break;
    case 'd':
        print_dc = 1;
        break;
    default:
        goto usage;
    }
    }

    if (print_dc) {
    net = network_dc_network(*network);
    if (net == NIL(network_t)) {
        return 0;
    }
    }
    else
       net = *network;

    if (argc - optind >= 1) {
    node_vec = com_get_nodes(net, argc-optind+1, argv+optind-1);
    if (array_n(node_vec) < 1) goto usage;
    for(i = 0; i < array_n(node_vec); i++) {
        node = array_fetch(node_t *, node_vec, i);
        if (node->type == PRIMARY_INPUT) {
        (void) fprintf(misout, "%-10s (primary input)\n", 
            node_name(node));
        } else if (node->type == PRIMARY_OUTPUT) {
        (void) fprintf(misout, "%-10s (primary output)\n", 
            node_name(node));
        } else {
        (void) fprintf(misout, "%-10s %d terms, %d literals\n", 
            node_name(node), node_num_cube(node), 
            node_num_literal(node));
        }
    }
    array_free(node_vec);

    } else {
    pi = network_num_pi(net);
    po = network_num_po(net);
    nodes = network_num_internal(net);

    lits = 0;
    lits_fac = 0;
    foreach_node(net, gen, p) {
        lits += node_num_literal(p);
        if (factored_lits) {
        lits_fac += factor_num_literal(p);
        }
    
    }

    (void) fprintf(misout, 
        "%-14s\tpi=%2d\tpo=%2d\tnodes=%3d\tlits(sop)=%4d",
        network_name(net), pi, po, nodes, lits);

    if (factored_lits) {
        (void) fprintf(misout, "\tlits(fac)=%4d", lits_fac);
    }
    (void) fprintf(misout, "\n");
    }
    return 0;

usage:
    (void) fprintf(miserr, "print_stats [-fd] [n1 n2 ...]\n");
    return 1;
}

static int
com_print_io(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    network_t *net;
    lsGen gen;
    node_t *node, *p;
    array_t *node_vec;
    int c, i, j, print_dc;

    print_dc = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "d")) != EOF) {
    switch(c) {
    case 'd':
        print_dc = 1;
        break;
    default:
        goto usage;
    }
    }

    if (print_dc) {
    net = network_dc_network(*network);
    if (net == NIL(network_t)) {
        return 0;
    }
    }
    else
       net = *network;

    if (argc - optind >= 1) {
    node_vec = com_get_nodes(net, argc-optind+1, argv+optind-1);
    for(i = 0; i < array_n(node_vec); i++) {
        node = array_fetch(node_t *, node_vec, i);
        (void) fprintf(misout, "%s\n", node_name(node));
        (void) fprintf(misout, "   inputs:  ");
        foreach_fanin(node, j, p) {
        (void) fprintf(misout, " %s", node_name(p));
        }
        (void) fprintf(misout, "\n");
        (void) fprintf(misout, "   outputs: ");
        foreach_fanout(node, gen, p) {
        (void) fprintf(misout, " %s", node_name(p));
        }
        (void) fprintf(misout, "\n");
    }
    array_free(node_vec);
    }
    else {
    (void) fprintf(misout, "primary inputs:  ");
    foreach_primary_input(net, gen, p) {
        (void) fprintf(misout, " %s", node_name(p));
    }
    (void) fprintf(misout, "\n");
    (void) fprintf(misout, "primary outputs: ");
    foreach_primary_output(net, gen, p) {
        (void) fprintf(misout, " %s", node_name(p));
    }
    (void) fprintf(misout, "\n");
    }
    return 0;
usage:
    (void) fprintf(miserr, "print_io [-d]\n");
    return 1;
}

/* ARGSUSED */
static int
com_chname(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    if (argc != 1) {
    (void) fprintf(miserr, "usage: chname\n");
    return 1;
    }

    if (name_mode == LONG_NAME_MODE) {
    name_mode = SHORT_NAME_MODE;
    (void) fprintf(misout, "changing to short-name mode\n");
    } else {
    name_mode = LONG_NAME_MODE;
    (void) fprintf(misout, "changing to long-name mode\n");
    }
    return 0;
}

static int
com_reset_name(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, short_names, long_names;

    short_names = long_names = 1;
    getopt_reset();
    while ((c = getopt(argc, argv, "sl")) != EOF) {
    switch(c) {
    case 's':
        long_names = 0;
        break;
    case 'l':
        short_names = 0;
        break;
    default:
        goto usage;
    }
    }
    if (argc != optind) goto usage;

    if (short_names) {
    network_reset_short_name(*network);
    }
    if (long_names) {
    network_reset_long_name(*network);
    }
    return 0;

usage:
    (void) fprintf(miserr, "usage: reset_names [-sl]\n");
    (void) fprintf(miserr, "\t-s reset only short names\n");
    (void) fprintf(miserr, "\t-s reset only long, made-up names\n");
    return 1;
}

/* ARGSUSED */
static int
com_print_altname(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    node_t *node;
    int i;

    if (argc < 2) {
    (void) fprintf(miserr, "usage: print_altname n1 n2 ...\n");
    return 1;
    }

    node_vec = com_get_nodes(*network, argc, argv);
    for(i = 0; i < array_n(node_vec); i++) {
    node = array_fetch(node_t *, node_vec, i);
    (void) fprintf(misout, "%s = %s\n", node->name, node->short_name);
    }
    array_free(node_vec);
    return 0;
}

/* ARGSUSED */
com_bdsyn(old_network, argc, argv)
network_t **old_network;        /* never used ... */
int argc;
char **argv;
{
    lsGen gen;
    int i, min_node;
    network_t *network;
    node_t *node;

    if (argc == 1) {
    min_node = 1;
    } else if (argc == 2) {
    min_node = 0;
    } else {
    (void) fprintf(miserr, "usage: bdsyn [1]\n");
    (void) fprintf(miserr, "\t\t1 implies don't minimize each node\n");
    return 1;
    }

    error_init();
    read_register_filename("(stdin)");
    network = 0;
    do {
    i = read_blif(stdin, &network);
    if (i == EOF) {
        break;
    } else if (i != 1) {
        (void) fprintf(miserr, "%s", error_string());
        return 1;
    }

    (void) network_collapse(network);
    if (min_node) {
        foreach_node(network, gen, node) {
        if (node->type == INTERNAL) {
            (void) node_simplify_replace(node, 
                    NIL(node_t), NODE_SIM_ESPRESSO);
        }
        }
    }
    write_blif_for_bdsyn(stdout, network);
    network_free(network);
    (void) fflush(stdout);
    } while (! feof(stdin));

    return 0;
}

init_io()
{
    com_add_command("bdsyn", com_bdsyn, 0);
    com_add_command("read_blif", com_read, 1);
    com_add_command("read_pla", com_read, 1);
    com_add_command("read_eqn", com_read, 1);
    com_add_command("write_bdnet", com_write, 0);
    com_add_command("write_blif", com_write, 0);
    com_add_command("write_eqn", com_write, 0);
    com_add_command("write_lif", com_write, 0);
    com_add_command("write_pla", com_write, 0);
    com_add_command("print", com_print, 0);
    com_add_command("print_io", com_print_io, 0);
    com_add_command("print_stats", com_print_stats, 0);
    com_add_command("chng_name", com_chname, 0);
    com_add_command("reset_name", com_reset_name, 0);
    com_add_command("print_altname", com_print_altname, 0);
    com_add_command("write_slif", com_write_slif, 0);
}

end_io()
{
    read_register_filename(NIL(char));
}
