#include "mis.h"
#include "io_int.h"


int read_lineno;
char *read_filename;

static int is_in_list();


void
read_error(s)
char *s;
{
    char buffer[20];

    if (read_filename == 0) {
	error_append(s);
    } else {
	error_append("\"");
	error_append(read_filename);
	error_append("\", line ");
	(void) sprintf(buffer, "%d", read_lineno);
	error_append(buffer);
	error_append(": ");
	error_append(s);
    }
    error_append("\n");
}


void
read_register_filename(fname)
char *fname;
{
    read_lineno = 1;
    if (read_filename != NIL(char)) {
	FREE(read_filename);
    }
    read_filename = fname ? util_strsav(fname) : 0;
}


node_t *
read_find_or_create_node(network, s)
network_t *network;
char *s;
{
    node_t *node;

    if ((node = network_find_node(network, s)) == 0) {
	node = node_alloc();
	node->name = util_strsav(s);
	network_add_node(network, node);
    }
    return node;
}


void
read_check_io_list(network, po_list)
network_t *network;
lsList po_list;
{
    int relax_i, relax_o;
    node_t *node;
    lsGen gen;
    char errmsg[1024];

    /* Check for nodes not driven or not faning out */
    relax_i = network_num_pi(network) == 0;	/* relax if no inputs given */ 
    relax_o = lsLength(po_list) == 0;		/* relax if no outputs given */
    foreach_node(network, gen, node) {
	if (node->F == 0 && node->type == INTERNAL) {
	    if (relax_i) {
		network_change_node_type(network, node, PRIMARY_INPUT);
	    } else {
		(void) sprintf(errmsg, "node '%s' is not driven (zero assumed)",
		    node->name);
		read_error(errmsg);
		node_replace(node, node_constant(0));
	    }
	} else if (node_num_fanout(node) == 0 && ! is_in_list(po_list, node)) {
	    if (relax_o) {
		lsNewEnd(po_list, (lsGeneric) node, LS_NH);
	    } else {
		(void) sprintf(errmsg, "node '%s' does not fanout", node->name);
		read_error(errmsg);
	    }
	}
    }
}


void
read_hack_outputs(network, po_list)
network_t *network;
lsList po_list;
{
   lsGen gen;
   node_t *node;

    /* hack the primary outputs -- create the PO nodes */
    lsForeachItem(po_list, gen, node) {
	(void) network_add_primary_output(network, node);
    }
}


static int
is_in_list(list, node)
lsList list;
node_t *node;
{
    node_t *p;
    lsGen gen;

    lsForeachItem(list, gen, p) {
        if (p == node) {
	    lsFinish(gen);
	    return 1;
	}
    }
    return 0;
}


void
read_filename_to_netname(network, filename)
network_t *network;
char *filename;
{
    char *name, *base;

    base = strrchr(filename, '/');
    name = (base == 0) ? filename : base + 1;
    network_set_name(network, name);
}
