#include "mis.h"


static void write_instance_name();

write_bdnet(fp, network)
FILE *fp;
network_t *network;
{
    lsGen gen;
    char *flag, *path, *viewname;
    node_t *node, *fanin;
    int i;

    if (! lib_network_is_mapped(network)) {
	error_append("write_bdnet: network not mapped, cannot write netlist\n");
	return 0;
    }

    path = com_get_flag("OCT-CELL-PATH");
    viewname = com_get_flag("OCT-CELL-VIEW");
    if (viewname == 0) viewname = "physical";

    (void) fprintf(fp, "MODEL \"%s\";\n", network_name(network));
    if ((flag = com_get_flag("OCT-TECHNOLOGY")) != 0) {
	(void) fprintf(fp, "    TECHNOLOGY %s;\n", flag);
    }
    if ((flag = com_get_flag("OCT-VIEWTYPE")) != 0) {
	(void) fprintf(fp, "    VIEWTYPE %s;\n", flag);
    }
    if ((flag = com_get_flag("OCT-EDITSTYLE")) != 0) {
	(void) fprintf(fp, "    EDITSTYLE %s;\n", flag);
    }
    (void) fprintf(fp, "\n");

    if ( network_num_pi(network) > 0 ) {
	(void) fprintf(fp, "INPUT");
	foreach_primary_input(network, gen, node) {
	    (void) fprintf(fp, "\n\t\"%s\"\t:\t\"%s\"", node->name, node->name);
	}
	(void) fprintf(fp, ";\n\n");
    }

    if ( network_num_po(network) > 0 ) {
	(void) fprintf(fp, "OUTPUT");
	foreach_primary_output(network, gen, node) {
	    (void) fprintf(fp, "\n\t\"%s\"\t:\t\"%s\"", 
			   node->name, node_get_fanin(node, 0)->name);
	}
	(void) fprintf(fp, ";\n\n");
    }

    if ( network_num_internal(network) > 0 ) {
	foreach_node(network, gen, node) {
	    if (node->type == INTERNAL) {
		write_instance_name(fp, path, viewname, node);
		foreach_fanin(node, i, fanin) {
		    (void) fprintf(fp, "\t\"%s\"\t:\t", 
				   lib_gate_pin_name(lib_gate_of(node), i, /*inflag*/ 1));
		    (void) fprintf(fp, "\"%s\";\n", fanin->name);
		}
		(void) fprintf(fp, "\t\"%s\"\t:\t", 
			       lib_gate_pin_name(lib_gate_of(node), 0, /*inflag*/ 0));
		(void) fprintf(fp, "\"%s\";\n\n", node->name);
	    }
	}
    }

    (void) fprintf(fp, "ENDMODEL;\n");
    return 1;
}


static void
write_instance_name(fp, path, viewname, node)
FILE *fp;
char *path;
char *viewname;
node_t *node;
{
    char *gate, *view;

    gate = util_strsav(lib_gate_name(lib_gate_of(node)));
    view = strchr(gate, ':');
    if (view == 0) {
	view = viewname; 
    } else {
	*view++ = '\0';
    }
    if (path == 0 || gate[0] == '/' || gate[0] == '~') {
	(void) fprintf(fp, "INSTANCE \"%s\":\"%s\"\n", gate, view);
    } else {
	(void) fprintf(fp, "INSTANCE \"%s/%s\":\"%s\"\n", path, gate, view);
    }
    FREE(gate);
}
