#include "mis.h"
#include "io_int.h"


static void do_write_blif();
static void write_blif_delay();
static int get_delay_param();


void
write_blif_for_bdsyn(fp, network)
FILE *fp;
network_t *network;
{
    io_fput_params("\\\n", 32000);
    do_write_blif(fp, network, 0, 0);
}


void
write_blif(fp, network, short_flag, netlist_flag)
FILE *fp;
network_t *network;
int short_flag;
int netlist_flag;
{
    io_fput_params("\\\n", 78);
    do_write_blif(fp, network, short_flag, netlist_flag);
}


static void
do_write_blif(fp, network, short_flag, netlist_flag)
register FILE *fp;
network_t *network;
int short_flag;
int netlist_flag;
{
    node_t *p;
    lsGen gen;

    io_fputs_break(fp, ".model ");
    io_fputs_break(fp, network_name(network));
    io_fputc_break(fp, '\n');

    io_fputs_break(fp, ".inputs");
    foreach_primary_input(network, gen, p) {
	io_fputc_break(fp, ' ');
	io_write_name(fp, p, short_flag);
    }
    io_fputc_break(fp, '\n');

    io_fputs_break(fp, ".outputs");
    foreach_primary_output(network, gen, p) {
	io_fputc_break(fp, ' ');
	io_write_name(fp, p, short_flag);
    }
    io_fputc_break(fp, '\n');

    write_blif_delay(fp, network);

    foreach_node(network, gen, p) {
	if (netlist_flag && lib_gate_of(p) != 0) {
	    io_write_gate(fp, p, short_flag);
	} else {
	    io_write_node(fp, ".names", p, short_flag);
	}
    }

    io_fputs_break(fp, ".end\n");
}

static void
write_blif_delay(fp, network)
FILE *fp;
network_t *network;
{
    delay_time_t delay;
    node_t *pi, *po;
    lsGen gen;
    char buffer[1024];

    if (get_default_delay_param(network, DELAY_DEFAULT_ARRIVAL_RISE, &delay)) {
	(void) sprintf(buffer, 
	    ".default_input_arrival %4.5f %4.5f\n", delay.rise, delay.fall);
	io_fputs_break(fp, buffer);
    }
    if (get_default_delay_param(network, DELAY_DEFAULT_REQUIRED_RISE, &delay)) {
	(void) sprintf(buffer, 
	    ".default_output_required %4.5f %4.5f\n", delay.rise, delay.fall);
	io_fputs_break(fp, buffer);
    }
    if (get_default_delay_param(network, DELAY_DEFAULT_DRIVE_RISE, &delay)) {
	(void) sprintf(buffer, 
	    ".default_input_drive %4.5f %4.5f\n", delay.rise, delay.fall);
	io_fputs_break(fp, buffer);
    }
    if (get_default_delay_param(network, DELAY_DEFAULT_OUTPUT_LOAD, &delay)) {
	(void) sprintf(buffer, 
	    ".default_output_load %4.5f\n", delay.rise);
	io_fputs_break(fp, buffer);
    }
    if (get_default_delay_param(network, DELAY_WIRE_LOAD_SLOPE, &delay)) {
	(void) sprintf(buffer,".wire_load_slope %4.2f\n", delay.rise);
	io_fputs_break(fp, buffer);
    }
    delay_print_blif_wire_loads(network, fp);

    foreach_primary_input(network, gen, pi) {
	if (get_delay_param(pi, DELAY_ARRIVAL_RISE, &delay)) {
	    (void) sprintf(buffer, ".input_arrival %s %4.5f %4.5f\n", 
		pi->name, delay.rise, delay.fall);
	    io_fputs_break(fp, buffer);
	}
	if (get_delay_param(pi, DELAY_DRIVE_RISE, &delay)) {
	    (void) sprintf(buffer, ".input_drive %s %4.5f %4.5f\n", 
		pi->name, delay.rise, delay.fall);
	    io_fputs_break(fp, buffer);
	}
    }

    foreach_primary_output(network, gen, po) {
	if (get_delay_param(po, DELAY_REQUIRED_RISE, &delay)) {
	    (void) sprintf(buffer, ".output_required %s %4.5f %4.5f\n", 
		po->name, delay.rise, delay.fall);
	    io_fputs_break(fp, buffer);
	}
	if (get_delay_param(po, DELAY_OUTPUT_LOAD, &delay)) {
	    (void) sprintf(buffer, ".output_load %s %4.5f\n", 
		po->name, delay.rise);
	    io_fputs_break(fp, buffer);
	}
    }
}

static int
get_default_delay_param(network, param, delay)
network_t *network;
delay_param_t param;
delay_time_t *delay;
{
    switch(param) {
    case DELAY_DEFAULT_ARRIVAL_RISE:
	delay->rise = delay_get_default_parameter(network, DELAY_DEFAULT_ARRIVAL_RISE);
	delay->fall = delay_get_default_parameter(network, DELAY_DEFAULT_ARRIVAL_FALL);
	break;
    case DELAY_DEFAULT_DRIVE_RISE:
	delay->rise = delay_get_default_parameter(network, DELAY_DEFAULT_DRIVE_RISE);
	delay->fall = delay_get_default_parameter(network, DELAY_DEFAULT_DRIVE_FALL);
	break;
    case DELAY_DEFAULT_REQUIRED_RISE:
	delay->rise = delay_get_default_parameter(network, DELAY_DEFAULT_REQUIRED_RISE);
	delay->fall = delay_get_default_parameter(network, DELAY_DEFAULT_REQUIRED_FALL);
	break;
    case DELAY_DEFAULT_OUTPUT_LOAD:
	delay->rise = delay_get_default_parameter(network, DELAY_DEFAULT_OUTPUT_LOAD);
	delay->fall = 0.0;
	break;
    case DELAY_WIRE_LOAD_SLOPE:
	delay->rise = delay_get_default_parameter(network, DELAY_WIRE_LOAD_SLOPE);
	delay->fall = 0.0;
	break;
    }

    if (delay->rise == DELAY_NOT_SET || delay->fall == DELAY_NOT_SET){
	return 0;
    }
    return 1;
}

static int
get_delay_param(node, param, delay)
node_t *node;
delay_param_t param;
delay_time_t *delay;
{
    switch(param) {
    case DELAY_DEFAULT_ARRIVAL_RISE:
	delay->rise = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_ARRIVAL_RISE);
	delay->fall = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_ARRIVAL_FALL);
	break;
    case DELAY_DEFAULT_DRIVE_RISE:
	delay->rise = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_DRIVE_RISE);
	delay->fall = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_DRIVE_FALL);
	break;
    case DELAY_DEFAULT_REQUIRED_RISE:
	delay->rise = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_REQUIRED_RISE);
	delay->fall = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_REQUIRED_FALL);
	break;
    case DELAY_DEFAULT_OUTPUT_LOAD:
	delay->rise = delay_get_parameter(NIL(node_t), 
	    DELAY_DEFAULT_OUTPUT_LOAD);
	delay->fall = 0;
	break;

    case DELAY_ARRIVAL_RISE:
	delay->rise = delay_get_parameter(node, DELAY_ARRIVAL_RISE);
	delay->fall = delay_get_parameter(node, DELAY_ARRIVAL_FALL);
	break;
    case DELAY_DRIVE_RISE:
	delay->rise = delay_get_parameter(node, DELAY_DRIVE_RISE);
	delay->fall = delay_get_parameter(node, DELAY_DRIVE_FALL);
	break;
    case DELAY_REQUIRED_RISE:
	delay->rise = delay_get_parameter(node, DELAY_REQUIRED_RISE);
	delay->fall = delay_get_parameter(node, DELAY_REQUIRED_FALL);
	break;
    case DELAY_OUTPUT_LOAD:
	delay->rise = delay_get_parameter(node, DELAY_OUTPUT_LOAD);
	delay->fall = 0;
	break;
    }

    if (delay->rise == DELAY_NOT_SET) return 0;
    if (delay->fall == DELAY_NOT_SET) return 0;
    return 1;
}
