#include "mis.h"
#include "io_int.h"


static void write_sop();


void
write_eqn(fp, network, short_name)
FILE *fp;
network_t *network;
int short_name;
{
    lsGen gen;
    node_t *node;

    io_fput_params("\n", 78);

    io_fputs_break(fp, "INORDER =");
    foreach_primary_input(network, gen, node) {
	io_fputc_break(fp, ' ');
	io_write_name(fp, node, short_name);
    }
    io_fputs_break(fp, ";\n");

    io_fputs_break(fp, "OUTORDER =");
    foreach_primary_output(network, gen, node) {
	io_fputc_break(fp, ' ');
	io_write_name(fp, node, short_name);
    }
    io_fputs_break(fp, ";\n");

    foreach_node(network, gen, node) {
	write_sop(fp, node, short_name);
    }
}


static void
write_sop(fp, node, short_name)
FILE *fp;
node_t *node;
int short_name;
{
    register int i, j;
    int first_cube, first_lit;
    node_cube_t cube;
    node_literal_t literal;
    node_t *fanin;

    /* never print inputs ... */
    if (node->type == PRIMARY_INPUT) return;

    /* print outputs only if feed by inputs, or if feed by > 1 node */
    if (node->type == PRIMARY_OUTPUT) {
	fanin = node_get_fanin(node, 0);
	if (fanin->type == INTERNAL && 
		io_po_fanout_count(fanin, NIL(node_t *)) == 1) {
	    return;
	}
    }

    io_write_name(fp, node, short_name);
    io_fputs_break(fp, " = ");

    switch(node_function(node)) {
    case NODE_0:
	io_fputs_break(fp, "0;\n");
	return;
    case NODE_1:
	io_fputs_break(fp, "1;\n");
	return;
    case NODE_PO:
	io_write_name(fp, node_get_fanin(node, 0), short_name);
	io_fputs_break(fp, ";\n");
	return;
    }

    first_cube = 1;
    for(i = node_num_cube(node)-1; i >= 0; i--) {
	if (! first_cube) {
	    io_fputs_break(fp, " + ");
	}
	cube = node_get_cube(node, i);

	first_lit = 1;
	foreach_fanin(node, j, fanin) {
	    literal = node_get_literal(cube, j);
	    switch(literal) {
	    case ZERO:
	    case ONE:
		if (! first_lit) {
		    io_fputc_break(fp, '*');
		}
		if (literal == ZERO) {
		    io_fputc_break(fp, '!');
		}
		io_write_name(fp, fanin, short_name);
		first_lit = 0;
		break;
	    }
	}
	first_cube = 0;
    }
    io_fputs_break(fp, ";\n");
}
