#include "mis.h"
#include "io_int.h"


static void
write_cover(fp, p)
FILE *fp;
node_t *p;
{
    register pset last, ps;
    register int c, i;

    if (p->type == PRIMARY_OUTPUT) {
	io_fputs_break(fp, "1 1\n");
    } else {
	foreach_set(p->F, last, ps) {
	    for(i = 0; i < p->nin; i++) {
		c = "?01-"[GETINPUT(ps,i)];
		io_fputc_break(fp, c);
	    }
	    io_fputs_break(fp, " 1\n");
	}
    }
}


int
io_po_fanout_count(node, first)
node_t *node;
node_t **first;
{
    register int po_cnt;
    register node_t *p;
    lsGen gen;

    po_cnt = 0;
    foreach_fanout(node, gen, p) {
	if (p->type == PRIMARY_OUTPUT) {
	    if (first != 0 && po_cnt == 0) *first = p; 
	    po_cnt++;
	}
    }
    return po_cnt;
}


void
io_write_name(fp, node, short_flag)
FILE *fp;
node_t *node;
int short_flag;
{
    node_t *po;

    if (node->type == INTERNAL && io_po_fanout_count(node, &po) == 1) {
	io_fputs_break(fp, short_flag ? po->short_name : po->name);
    } else {
	io_fputs_break(fp, short_flag ? node->short_name : node->name);
    }
}


void
io_write_node(fp, prefix, node, short_flag)
FILE *fp;
char *prefix;
node_t *node;
int short_flag;
{
    register int i;
    node_t *fanin;

    /* never print inputs ... */
    if (node->type == PRIMARY_INPUT) return;

    /* print outputs only if feed by inputs, or if feed by > 1 node */
    if (node->type == PRIMARY_OUTPUT) {
	fanin = node_get_fanin(node, 0);
	if (fanin->type != PRIMARY_INPUT && 
			io_po_fanout_count(fanin, NIL(node_t *)) == 1) {
	    return;
	}
    }

    io_fputs_break(fp, prefix);
    foreach_fanin(node, i, fanin) {
	io_fputc_break(fp, ' ');
	io_write_name(fp, fanin, short_flag);
    }
    io_fputc_break(fp, ' ');
    io_write_name(fp, node, short_flag);
    io_fputc_break(fp, '\n');

    write_cover(fp, node);
}

void
io_write_gate(fp, node, short_flag)
FILE *fp;
node_t *node;
int short_flag;
{
    register int i;
    node_t *fanin, *po;

    /* never print inputs ... */
    if (node->type == PRIMARY_INPUT) return;

    /* print outputs only if feed by inputs, or if feed by > 1 node */
    if (node->type == PRIMARY_OUTPUT) {
	fanin = node_get_fanin(node, 0);
	if (fanin->type != PRIMARY_INPUT && 
			io_po_fanout_count(fanin, NIL(node_t *)) == 1) {
	    return;
	}
    }

    /* may happen if two or more outputs are equivalent ... */
    if (lib_gate_of(node) == 0) {
	io_write_node(fp, ".names", node, short_flag);
	return;
    }

    io_fputs_break(fp, ".gate ");
    io_fputs_break(fp, lib_gate_name(lib_gate_of(node)));
    foreach_fanin(node, i, fanin) {
	io_fputc_break(fp, ' ');
	fputs(lib_gate_pin_name(lib_gate_of(node), i, /*in*/ 1), fp);
	fputc('=', fp);
	if (fanin->type == INTERNAL && io_po_fanout_count(fanin, &po) == 1) {
	    fputs(short_flag ? po->short_name : po->name, fp);
	} else {
	    fputs(short_flag ? fanin->short_name : fanin->name, fp);
	}
    }
    io_fputc_break(fp, ' ');
    io_fputs_break(fp, lib_gate_pin_name(lib_gate_of(node), 0, /*in*/ 0));
    io_fputc_break(fp, '=');
    io_write_name(fp, node, short_flag);
    io_fputc_break(fp, '\n');
}

static int colnum;
static char *io_break_string = 0;
static int io_break_column = 32000;


/*
 * write a string to a file, keeping track of current column number;
 * force a newline to keep all lines < 80 columns (when possible)
 */

void
io_fputs_break(fp, s)
register FILE *fp;
register char *s;
{
    char *p;

    if (colnum+strlen(s) > io_break_column) {
	if (io_break_string != 0) {
	    for(p = io_break_string; *p; p++) {
		(void) putc(*p, fp);
		colnum = (*p == '\n') ? 0 : colnum + 1;
	    }
	}
    }
    for( ; *s; s++) {
	(void) putc(*s, fp);
	colnum = (*s == '\n') ? 0 : colnum + 1;
    }
}


void
io_fputc_break(fp, c)
FILE *fp;
int c;
{
    (void) putc(c, fp);
    colnum = (c == '\n') ? 0 : colnum + 1;
}


void
io_fput_params(s, count)
char *s;
int count;
{
    io_break_string = s;
    io_break_column = count;
    colnum = 0;
}
