#include "mis.h"
#include "vov.h"

extern void init_mis();
extern void end_mis();

static int main_has_restarted = 0;
static network_t *network;		/* allows for restart ... */


static int
source_misrc(network)
network_t **network;
{
    char *cmdline;
    int status1, status2, status3;

    cmdline = ALLOC(char, strlen(mis_library()) + 20);

    (void) sprintf(cmdline, "source -s %s/.misrc", mis_library());
    status1 = com_execute(network, cmdline);
    status2 = com_execute(network, "source -s ~/.misrc");
    status3 = com_execute(network, "source -s .misrc");
    FREE(cmdline);
    return status1 && status2 && status3;
}


static void
usage(prog)
char *prog;
{
    (void) fprintf(miserr, "%s\n", mis_version());
    (void) fprintf(miserr, 
"usage: %s [-sx] [-c cmd] [-f script] [-o file] [-t type] [-T type] [file]\n", 
	prog);
    (void) fprintf(miserr,
	"    -c cmd\texecute MIS commands `cmd'\n");
    (void) fprintf(miserr,
	"    -f file\texecute MIS commands from a file\n");
    (void) fprintf(miserr,
	"    -o file\tspecify output filename (default is -)\n");
    (void) fprintf(miserr,
	"    -s\t\tsuppress initial 'source %s/.misrc'\n", mis_library());
    (void) fprintf(miserr,
    "    -t type\tspecify input type (blif, eqn, oct, pla, or none)\n");
    (void) fprintf(miserr,
    "    -T type\tspecify output type (bdnet, blif, eqn, oct, pla, or none)\n");
    (void) fprintf(miserr,
	"    -x\t\tequivalent to '-t none -T none'\n");
    exit(2);
}


static int
check_type(s)
char *s;
{
    if (strcmp(s, "bdnet") == 0) {
         return 1;
    } else if (strcmp(s, "blif") == 0) {
         return 1;
    } else if (strcmp(s, "eqn") == 0) {
         return 1;
    } else if (strcmp(s, "oct") == 0) {
         return 1;
    } else if (strcmp(s, "pla") == 0) {
         return 1;
    } else if (strcmp(s, "none") == 0) {
         return 1;
    } else {
	(void) fprintf(miserr, "unknown type %s\n", s);
	return 0;
    }
}

main(argc, argv)
int argc;
char **argv;
{
    int c, status, batch, initial_source, initial_read, final_write;
    char readcmd[20], writecmd[20];
    char *cmdline, *cmdline1, *infile, *outfile;

    program_name = argv[0];

    if (main_has_restarted) {
	(void) fprintf(stderr, "Restarting frozen image ...\n");
    } else {
	main_has_restarted = 1;
	init_mis();
	network = network_alloc();
    }

    cmdline = util_strsav("");
    (void) strcpy(readcmd, "read_blif");
    (void) strcpy(writecmd, "write_blif");
    infile = "-";
    outfile = "-";
    initial_source = 1;
    initial_read = 1;
    final_write = 1;
    batch = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "c:f:o:st:T:x")) != EOF) {
	switch(c) {
	case 'c':
	    FREE(cmdline);
	    cmdline = util_strsav(optarg);
	    batch = 1;
	    break;

	case 'f':
	    FREE(cmdline);
	    cmdline = ALLOC(char, strlen(optarg) + 20);
	    (void) sprintf(cmdline, "source %s", optarg);
	    batch = 1;
	    break;

	case 'o':
	    outfile = optarg;
	    break;

	case 's':
	    initial_source = 0;
	    break;

	case 't':
	    if (check_type(optarg)) {
		if (strcmp(optarg, "none") == 0) {
		    initial_read = 0;
		} else {
		    (void) sprintf(readcmd, "read_%s", optarg);
		}
	    } else {
		usage(argv[0]);
	    }
	    batch = 1;
	    break;

	case 'T':
	    if (check_type(optarg)) {
		if (strcmp(optarg, "none") == 0) {
		    final_write = 0;
		} else {
		    (void) sprintf(writecmd, "write_%s", optarg);
		}
	    } else {
		usage(argv[0]);
	    }
	    batch = 1;
	    break;

	case 'x':
	    final_write = 0;
	    initial_read = 0;
	    batch = 1;
	    break;

	default:
	    usage(argv[0]);
	}
    }

    VOVbegin( argc, argv );
    if (! batch) {
	/* interactive use ... */
	if (argc - optind != 0) {
	    (void) fprintf(miserr, "warning -- trailing arguments ignored\n");
	}

	(void) fprintf(misout, "%s\n", mis_version());
	if (initial_source) {
	    (void) source_misrc(&network);
	}
	while (com_execute(&network, "source -ip -") != -1)
	    ;
	status = 0;

    } else {

	/* read initial network */
	if (argc - optind == 0) {
	    infile = "-";
	} else if (argc - optind == 1) {
	    infile = argv[optind];
	} else {
	    usage(argv[0]);
	}

	if (initial_source) {
	    (void) source_misrc(&network);
	}

	status = 0;
	if (initial_read) {
	    cmdline1 = ALLOC(char, strlen(infile) + 20);
	    (void) sprintf(cmdline1, "%s %s", readcmd, infile);
	    status = com_execute(&network, cmdline1);
	    FREE(cmdline1);
	}

	if (status == 0) {
	    status = com_execute(&network, cmdline);
	    if ((status == 0 || status == -1) && final_write) {
		cmdline1 = ALLOC(char, strlen(outfile) + 20);
		(void) sprintf(cmdline1, "%s %s", writecmd, outfile);
		status = com_execute(&network, cmdline1);
		FREE(cmdline1);
	    }
	}

    }

    FREE(cmdline);
    network_free(network);
    end_mis();
    VOVend(status);
}
