 /* file addinv.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:28 */
#include "mis.h"
#include "map.h"
#include "map_int.h"


static node_t *
create_inverter(nodelist, fanin)
lsList nodelist;
node_t *fanin;
{
    node_t *inv;

    inv = node_alloc();
    node_replace(inv, node_literal(fanin, 0));
    lsNewEnd(nodelist, (char *) inv, LS_NH);
    return inv;
}


static int
feeds_po(node)
node_t *node;
{
    lsGen gen;
    node_t *fanout;

    foreach_fanout(node, gen, fanout) {
	if (fanout->type == PRIMARY_OUTPUT) {
	    (void) lsFinish(gen);
	    return 1;
	}
    }
    return 0;
}

static void
add_inverters(nodelist, node)
lsList nodelist;
node_t *node;
{
    int new_inv;
    node_t *inv1, *inv2, *inv, *fanout;
    lsGen gen;

    /* Find if the node already has an inverter at the output */
    inv = NIL(node_t);
    foreach_fanout(node, gen, fanout) {
	if (node_function(fanout) == NODE_INV) {
	    if (inv != NIL(node_t)) {
		/*
		(void) fprintf(stderr, 
		    "Warning: node %s has fanout to >1 inverter\n",
		    node_name(node));
		 */
	    } else {
		inv = fanout;
	    }
	}
    }

    /* Create an inverter if none found so far */
    new_inv = 0;
    if (inv == NIL(node_t)) {
	inv = create_inverter(nodelist, node);
	new_inv = 1;
    }

    /* Handle the positive-phase fanouts (i.e., fanouts of 'node') */
    foreach_fanout(node, gen, fanout) {
	if (node_function(fanout) != NODE_INV) {
	    inv1 = create_inverter(nodelist, inv);
	    assert(node_patch_fanin(fanout, node, inv1));
	}
    }

    /* Handle the negative-phase fanouts (i.e., fanouts of original 'inv') */
    if (! new_inv) {
	foreach_fanout(inv, gen, fanout) {
	    if (node_function(fanout) != NODE_INV && 
					    fanout->type != PRIMARY_OUTPUT) {
		inv1 = create_inverter(nodelist, inv);
		inv2 = create_inverter(nodelist, inv1);
		assert(node_patch_fanin(fanout, inv, inv2));
	    }
	}
    }
}

void 
map_add_inverter(network, add_at_pipo)
network_t *network;
int add_at_pipo;
{
    lsList nodelist;
    node_t *node;
    int need_inv;
    lsGen gen;

    nodelist = lsCreate();

    foreach_node(network, gen, node) {
	switch(node_function(node)) {
	case NODE_PI:
	    need_inv = node_num_fanout(node) > 1 || add_at_pipo;
	    break;

	case NODE_PO:
	case NODE_INV:
	case NODE_0:
	case NODE_1:
	    need_inv = 0;
	    break;

	case NODE_AND:
	case NODE_OR:
	    /* special case for single-fanout outputs on primitives */
	    if (node_num_fanout(node) == 1 && feeds_po(node)) {
		need_inv = add_at_pipo;
	    } else {
		need_inv = 1;
	    }
	    break;

	default:
	    fail("map_add_inverter: what is this doing here ?");
	    /* NOTREACHED */
	}

	if (need_inv) {
	    add_inverters(nodelist, node);
	}
    }

    lsForeachItem(nodelist, gen, node) {
	network_add_node(network, node);
    }
    lsDestroy(nodelist, (void (*)()) 0);
}

void
map_remove_inverter(network, report_data)
network_t *network;
VoidFn report_data;
{
    int i, j, num_fanout;
    node_t *inv1, *inv2, *node, *fanout, *fanout1;
    node_t **fanouts;
    array_t *nodes;
    lsGen gen, gen1;

    if (report_data) {
      fprintf(misout, "before remove inv: "); 
      (*report_data)(network);
    }
    /* step 1: check for sequential inverters */
    nodes = network_dfs(network);
    for(i = 0; i < array_n(nodes); i++) {
	inv2 = array_fetch(node_t *, nodes, i);
	if (node_function(inv2) == NODE_INV) {

	    /* see if the fanin is also an inverter ... */
	    inv1 = node_get_fanin(inv2, 0);
	    if (node_function(inv1) == NODE_INV) {
		node = node_get_fanin(inv1, 0);
		num_fanout = node_num_fanout(inv2);
		fanouts = ALLOC(node_t *, num_fanout); 
		j = 0;
		foreach_fanout(inv2, gen, fanout) {
		  fanouts[j++] = fanout;
		}
		for (j = 0; j < num_fanout; j++) {
		  assert(node_patch_fanin(fanouts[j], inv2, node));
		}
		FREE(fanouts);
		network_delete_node(network, inv2);
	    }
	}
    }
    array_free(nodes);
    if (report_data) {
      fprintf(misout, "before remove inv: "); 
      (*report_data)(network);
    }
    /* step 2: check for parallel inverters */
    nodes = network_dfs_from_input(network);
    for(i = 0; i < array_n(nodes); i++) {
	node = array_fetch(node_t *, nodes, i);
	if (node_function(node) != NODE_INV) {

	    /* find if more than 2 inverters fanout from 'node' */
	    inv1 = NIL(node_t);
	    foreach_fanout(node, gen, fanout) {
		if (node_function(fanout) == NODE_INV) {
		    if (inv1 == NIL(node_t)) {
			inv1 = fanout;
		    } else {
			/* move fanouts of 'fanout' to fanouts of inv1 */
			foreach_fanout(fanout, gen1, fanout1) {
			    assert(node_patch_fanin(fanout1, fanout, inv1));
			}
			network_delete_node(network, fanout);
		    }
		}
	    }
	}
    }
    array_free(nodes);

    /* cleanup dangling nodes  (?) -- are there any ?? */
    (void) network_cleanup(network);
}
