 /* file libutil.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:43 */
#include "mis.h"
#include "map_int.h"
#include "lib_int.h"

library_t *lib_current_library;


library_t *
lib_get_library()
{
    return lib_current_library;
}


lsGen 
lib_gen_classes(library) 
library_t *library;
{
    return lsStart(library->classes);
}


lsGen 
lib_gen_gates(class) 
lib_class_t *class;
{
    return lsStart(class->gates);
}


lib_gate_t *
lib_get_gate(library, name)
library_t *library;
char *name;
{
    lib_gate_t *gate;
    lsGen gen;

    lsForeachItem(library->gates, gen, gate) {
	if (strcmp(gate->name, name) == 0) {
	    (void) lsFinish(gen);
	    return gate;
	}
    }
    return NIL(lib_gate_t);
}

lib_class_t *
lib_get_class(network, library)
network_t *network;
library_t *library;
{
    lsGen gen;
    lib_class_t *class;
    lib_gate_t *gate;
    network_t *net1;
    node_t *node, *node1;
    prim_t *prim;
    char *dummy;

    /* Special hack as usual for constants */
    if (network_num_pi(network) == 0) {
	node = node_get_fanin(network_get_po(network, 0), 0);
	lsForeachItem(library->classes, gen, class) {
	    if (network_num_pi(class->network) == 0) {
		node1 = node_get_fanin(network_get_po(class->network, 0), 0);
		if (node_equal_by_name(node1, node)) {
		    (void) lsFinish(gen);
		    return class;
		}
	    }
	}
	return NIL(lib_class_t);
    }

    net1 = network_dup(network);
    decomp_quick_network(net1);
    if (library->nand_flag) {
	decomp_tech_network(net1, /* and_limit */ 0, /* or_limit */ 2);
    } else {
	decomp_tech_network(net1, /* and_limit */ 2, /* or_limit */ 0);
    }
    add_inv_network(net1);
    if (library->add_inverter) {
	map_add_inverter(net1, /* add_at_pipo */ 0);
    }
    map_setup_network(net1);	/* needed for lib_pattern_matches() */

    class = 0;
    lsForeachItem(library->patterns, gen, prim) {
	if (lib_pattern_matches(net1, prim)) {
	    lsFirstItem(prim->gates, &dummy, LS_NH);
	    gate = (lib_gate_t *) dummy;
	    class = gate->class;
	    (void) lsFinish(gen);
	    break;
	}
    }
    network_free(net1);
    return class;
}

/*
 *  lib_class_t access functions
 */

char *
lib_class_name(class)
lib_class_t *class;
{
    lib_gate_t *gate;
    char *dummy;

    if (class == 0) return NIL(char);
    if (lsFirstItem(class->gates, &dummy, LS_NH) == LS_OK) {
	gate = (lib_gate_t *) dummy;
	return gate->name;
    }
    return NIL(char);
}


lib_class_t * 
lib_class_dual(class) 
lib_class_t *class;
{
    return class->dual_class;
}


network_t *
lib_class_network(class) 
lib_class_t *class;
{
    return class->network;
}

/*
 *  lib_gate_t access functions
 */

char *
lib_gate_name(gate) 
lib_gate_t *gate;
{
    if (gate == 0) return NIL(char);
    return gate->name;
}


char *
lib_gate_pin_name(gate, pin, inflag) 
lib_gate_t *gate; 
int pin; 
int inflag;
{
    node_t *p;

    if (gate == 0) return NIL(char);
    if (inflag) {
	p = network_get_pi(gate->network, pin);
    } else {
	p = network_get_po(gate->network, pin);
    }
    return p ? p->name : NIL(char);
}
    

double 
lib_gate_area(gate) 
lib_gate_t *gate;
{
    if (gate == 0) return 0.0;
    return gate->area;
}


lib_class_t *
lib_gate_class(gate)
lib_gate_t *gate;
{
    if (gate == 0) return NIL(lib_class_t);
    return gate->class;
}


int
lib_gate_num_in(gate)
lib_gate_t *gate;
{
    if (gate == 0) return -1;
    return network_num_pi(gate->network);
}


int
lib_gate_num_out(gate)
lib_gate_t *gate;
{
    if (gate == 0) return -1;
    return network_num_po(gate->network);
}

/*
 *  misc functions 
 */

lib_gate_t * 
lib_gate_of(node) 
node_t *node;
{
    if (MAP(node) == 0) return NIL(lib_gate_t);
    return MAP(node)->gate;
}


char *
lib_get_pin_delay(node, pin)
node_t *node;
int pin;
{
    if (MAP(node) == 0) return 0;
    if (MAP(node)->gate == 0) return 0;
    /* ### should check for out of bounds ? */
    return MAP(node)->gate->delay_info[pin];
}


int
lib_network_is_mapped(network)
network_t *network;
{
  network_type_t type;

  type = map_get_network_type(network);
  return (type == MAPPED_NETWORK) ? 1 : 0;
}

/*
 *  backwards compatibility
 */

char *
lib_get_gate_name(node)
node_t *node;
{
    return lib_gate_name(lib_gate_of(node));
}


char *
lib_get_pin_name(node, pin)
node_t *node;
int pin;
{
    return lib_gate_pin_name(lib_gate_of(node), pin, /* inflag */ 1);
}


char *
lib_get_out_pin_name(node, pin)
node_t *node;
int pin;
{
    return lib_gate_pin_name(lib_gate_of(node), pin, /* inflag */ 0);
}


double
lib_get_gate_area(node)
node_t *node;
{
    return lib_gate_area(lib_gate_of(node));
}

/*
 *  set the gate of a node -- assumes single-output gate
 */

int
lib_set_gate(user_node, gate, formals, actuals, nin)
node_t *user_node;
lib_gate_t *gate;
char **formals;
node_t **actuals;
int nin;
{
    register int i, j;
    node_t *node, *fanin, **new_fanin;
    char errmsg[1024];

    if (nin != network_num_pi(gate->network)) {
	error_append("lib_set_gate: number of inputs on gate is incorrect\n");
	return 0;
    }
    /*
    if (network_num_po(gate->network) != 1) {
	error_append("lib_set_gate: number of outputs on gate is incorrect\n");
	return 0;
    }
    */

    new_fanin = ALLOC(node_t *, nin);
    node = node_get_fanin(network_get_po(gate->network, 0), 0);
    foreach_fanin(node, i, fanin) {
	new_fanin[i] = NIL(node_t);
	for(j = 0; j < nin; j++) {
	    if (strcmp(fanin->name, formals[j]) == 0) {
		new_fanin[i] = actuals[j];
		break;
	    }
	}
	if (new_fanin[i] == 0) {
	    (void) sprintf(errmsg, 
		"lib_set_gate: pin '%s' not in formals list\n", fanin->name);
	    error_append(errmsg);
	    return 0;
	}
    }

    node_replace_internal(user_node, new_fanin, nin, sf_save(node->F)); 
    map_alloc(user_node);
    MAP(user_node)->gate = gate;
    return 1;
}

void
lib_free(library)
library_t *library;
{
    lsGen gen;
    lib_gate_t *gate;
    lib_class_t *class;
    prim_t *prim;
    int i;

    lsForeachItem(library->gates, gen, gate) {
	for(i = network_num_pi(gate->network)-1; i >= 0; i--) {
	    FREE(gate->delay_info[i]);
	}
	FREE(gate->delay_info);
	FREE(gate->name);
	FREE(gate);
    }
    lsDestroy(library->gates, (void (*)()) 0);

    lsForeachItem(library->classes, gen, class) {
	network_free(class->network);
	lsDestroy(class->gates, (void (*)()) 0);
	FREE(class->name);
	FREE(class);
    }
    lsDestroy(library->classes, (void (*)()) 0);

    lsForeachItem(library->patterns, gen, prim) {
	prim_free(prim);
    }
    lsDestroy(library->patterns, (void (*)()) 0);
    FREE(library);
}

void
lib_dump(fp, library, detail)
FILE *fp;
library_t *library;
int detail;
{
    lsGen gen;
    prim_t *prim;
    lib_class_t *class;

    lsForeachItem(library->classes, gen, class) {
	lib_dump_class(fp, class);
    }

    (void) fprintf(fp, "\n");
    lsForeachItem(library->patterns, gen, prim) {
	map_prim_dump(fp, prim, detail);
    }
}



void
lib_dump_class(fp, class)
FILE *fp;
lib_class_t *class;
{
    lsGen gen, gen1;
    lib_gate_t *gate;
    node_t *node;
    char *dual_name;

    (void) fprintf(fp, "class: %s\n", lib_class_name(class));

    lsForeachItem(class->gates, gen, gate) {
	dual_name = lib_class_name(class->dual_class);
	if (dual_name == 0) dual_name = "<none>";

	(void) fprintf(fp, "    gate: %-10s  area: %5.2f  dual_class: %s\n", 
	    lib_gate_name(gate), lib_gate_area(gate), dual_name);

	foreach_primary_output(gate->network, gen1, node) {
	    (void) fprintf(fp, "      ");
	    (void) node_print_negative(fp, node_get_fanin(node, 0));
	}

	assert(gate->class == class);
    }
    (void) fprintf(fp, "\n");
}
