 /* file %M% release %I% */
 /* last modified: %G% at %U% */
#include <math.h>
#include "mis.h"
#include "lib_int.h"
#include "map_macros.h"
#include "map_delay.h"

static void free_parameters();
static void precompute_parameters();
static void map_set_current_network();

#define ARRIVAL(time, new_time)	\
    if (new_time > time) time = new_time;

void delay_map_compute_required_times(nin, model, node_required, load, required)
int nin;
char **model;
delay_time_t node_required;
double load;
delay_time_t *required;
{
  register int i;
  register delay_pin_t *pin_delay, **lib_delay_model;
  delay_time_t delay;

  lib_delay_model = (delay_pin_t **) model;
  for(i = nin-1; i >= 0; i--) {
    pin_delay = lib_delay_model[i];
    delay.rise = pin_delay->block.rise + pin_delay->drive.rise * load;
    delay.fall = pin_delay->block.fall + pin_delay->drive.fall * load;

    switch(pin_delay->phase) {
    case PHASE_INVERTING:
      required[i].rise = node_required.fall - delay.fall;
      required[i].fall = node_required.rise - delay.rise; 
      break;
    case PHASE_NONINVERTING:
      required[i].rise = node_required.rise - delay.rise;
      required[i].fall = node_required.fall - delay.fall;
      break;
    case PHASE_NEITHER:
      required[i].rise = MIN(node_required.rise - delay.rise, node_required.fall - delay.fall);
      required[i].fall = required[i].rise;
      break;
    }
  }
}


 /* TREATMENT OF PRIMARY INPUTS / OUTPUTS */

typedef struct {
  lib_gate_t *default_inv;
  int arrival_set;
  delay_time_t arrival;
  int drive_set;
  delay_time_t drive;
  int load_set;
  double load;
  int required_set;
  delay_time_t required;
} pipo_t;

static pipo_t DEFAULT_PIPO_SETTINGS = {0, 0, {0.0, 0.0}, 0, {0.0, 0.0}, 0, 0.0, 0, {0.0, 0.0}};
static pipo_t pipo_settings;

static network_t *map_current_network;

static void map_set_current_network(network)
network_t *network;
{
  assert(network != NIL(network_t));
  map_current_network = network;
}

void map_reset_default_delay_info(network)
network_t *network;
{
  map_set_current_network(network);
  pipo_settings = DEFAULT_PIPO_SETTINGS;
  pipo_settings.default_inv = lib_get_default_inverter();
  pipo_settings.drive = delay_get_drive(pipo_settings.default_inv->delay_info[0]);
  pipo_settings.load = delay_get_load(pipo_settings.default_inv->delay_info[0]);
}

void pipo_set_default_po_required(default_value)
delay_time_t default_value;
{
  pipo_settings.required = default_value;
  pipo_settings.required_set = 0;
}

delay_time_t pipo_get_pi_arrival(node)
node_t *node;
{
  delay_time_t arrival;

  if (node != NIL(node_t) && delay_get_pi_arrival_time(node, &arrival)) return arrival;
  if (pipo_settings.arrival_set) return pipo_settings.arrival;
  fprintf(miserr, "WARNING: uses as primary input arrival time the value (%2.2f,%2.2f)\n",
	  pipo_settings.arrival.rise, pipo_settings.arrival.fall);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_ARRIVAL_RISE, pipo_settings.arrival.rise);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_ARRIVAL_FALL, pipo_settings.arrival.fall);
  pipo_settings.arrival_set = 1;
  return pipo_settings.arrival;
}


delay_time_t pipo_get_pi_drive(node)
node_t *node;
{
  delay_time_t drive;

  if (node != NIL(node_t) && delay_get_pi_drive(node, &drive)) return drive;
  if (pipo_settings.drive_set) return pipo_settings.drive;
  fprintf(miserr, "WARNING: uses as primary input drive the value (%2.2f,%2.2f)\n",
	  pipo_settings.drive.rise, pipo_settings.drive.fall);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_DRIVE_RISE, pipo_settings.drive.rise);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_DRIVE_FALL, pipo_settings.drive.fall);
  pipo_settings.drive_set = 1;
  return pipo_settings.drive;
}

delay_time_t pipo_get_po_required(node)
node_t *node;
{
  delay_time_t required;

  if (node != NIL(node_t) && delay_get_po_required_time(node, &required)) return required;
  if (pipo_settings.required_set) return pipo_settings.required;
  fprintf(miserr, "WARNING: uses as primary output required time the value (%2.2f,%2.2f)\n",
	  pipo_settings.required.rise, pipo_settings.required.fall);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_REQUIRED_RISE, pipo_settings.required.rise);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_REQUIRED_FALL, pipo_settings.required.fall);
  pipo_settings.required_set = 1;
  return pipo_settings.required;
}

double pipo_get_po_load(node)
node_t *node;
{
  double load;

  if (node != NIL(node_t) && delay_get_po_load(node, &load)) return load;
  if (pipo_settings.load_set) return pipo_settings.load;
  fprintf(miserr, "WARNING: uses as primary output load the value %2.2f\n", pipo_settings.load);
  delay_set_default_parameter(map_current_network, DELAY_DEFAULT_OUTPUT_LOAD, pipo_settings.load);
  pipo_settings.load_set = 1;
  return pipo_settings.load;
}

 /* should be in delay/delay.c really */

pin_phase_t delay_get_polarity(pin_delay)
char *pin_delay;
{
  return ((delay_pin_t *) pin_delay)->phase;
}

double map_compute_wire_load(n_fanouts)
int n_fanouts;
{
  return compute_wire_load(map_current_network, n_fanouts);
}
