 /* file map_int.h release 1.3 */
 /* last modified: 9/9/90 at 17:32:10 */
#ifndef MAP_INT_H
#define MAP_INT_H

#include "map_defs.h"
#include "lib_int.h"

typedef struct prim_struct prim_t;
typedef struct prim_node_struct prim_node_t;
typedef struct prim_edge_struct prim_edge_t;

typedef enum edge_dir_enum edge_dir_t;
enum edge_dir_enum {DIR_IN, DIR_OUT};

struct prim_struct {
    int ninputs;
    prim_node_t **inputs;	/* direct pointers to primitive inputs */
    int noutputs;
    prim_node_t **outputs;	/* direct pointers to primitive outputs */
    prim_node_t *nodes; 	/* linked list of nodes for the primitive */
    prim_edge_t *edges; 	/* linked list of edge for the primitive */
    lsList gates;
};


struct prim_node_struct {
    char *name;			/* name of node (inputs and outputs only) */
    unsigned isomorphic_sons:1;
    unsigned nfanin:15;
    unsigned nfanout:15;
    node_type_t type;
    node_t *binding;		/* used during matching: binding for node */
    prim_node_t *next;		/* linked list of nodes */
};


struct prim_edge_struct {
    prim_node_t *this_node;
    prim_node_t *connected_node;
    edge_dir_t dir;
    prim_edge_t *next;		/* linked list of edges */
};


typedef enum {FANOUT_NODE, ROOT_NODE, TREE_NODE} map_type_t;

typedef enum {
  RAW_NETWORK,
  ANNOTATED_NETWORK,
  MAPPED_NETWORK,
  INCONSISTENT_NETWORK,
  EMPTY_NETWORK,
  UNKNOWN_TYPE
} network_type_t;

#ifdef _IBMR2
typedef struct ibmr2_map_struct ibmr2_map_t;
struct ibmr2_map_struct {
#else
typedef struct map_struct map_t;
struct map_struct {
#endif
    lib_gate_t *gate;		/* best gate which matches */
    prim_node_t *binding;	/* used during matching */
    node_t *match_node;		/* used to re-build the graph */
    int ninputs;		/* number of inputs for this gate */
    node_t **save_binding;	/* bindings for the gate inputs */
    delay_time_t **arrival_info; /* pointers to arrival times of inputs. used in fanout opt */
    double map_area;		/* current area cost function */
    delay_time_t map_arrival;	/* current arrival time cost function */
    double load;		/* record load on this pin (from last map) */
    int slack_is_critical;	/* slack at node is 'critical' */
    int inverter_paid_for;
    delay_time_t inverter_arrival;
    int index;
    delay_time_t required;
    st_table *gate_link;	/* used in bin_delay_build_network: bindings gate input -> gate root */
    st_generator *gen; 		/* to be used in conjunction with gate_link */
    node_t *node_y; 		/* used in bin_power.c to cache information */
    map_type_t type;
};

typedef struct match_struct match_t;
struct match_struct {
    int ninputs;
    int noutputs;
    node_t **inputs;
    node_t **outputs;
    prim_t *prim;
    lib_gate_t *gate;
};


#define MAP_SLOT		map
#ifdef _IBMR2
#define MAP(node)		((ibmr2_map_t *) (node)->MAP_SLOT)
#else
#define MAP(node)		((map_t *) (node)->MAP_SLOT)
#endif

extern void map_network_dup();
extern int map_network_to_prim();
extern st_table *map_find_isomorphisms();
extern void prim_free();
extern void prim_clear_binding();
extern void map_prim_dump();
extern int tree_map();
extern void map_print_tree_size();
extern node_t *map_prim_get_root();
extern void map_alloc(), map_free(), map_invalid();
extern void map_dup(), map_setup_network();
extern void gen_all_matches(), gen_all_matches_allowing_fanout();
extern int map_check_form();
extern int do_tree_premap();
extern void map_prep();
extern void patch_constant_cells();
extern network_t *map_build_network ();
extern lib_gate_t *choose_smallest_gate();

 /* map_interface.c */
extern network_t *complete_map_interface();
extern network_t *build_mapped_network_interface();
extern network_t *tree_map_interface();
extern network_t *fanout_opt_interface();
extern network_type_t map_get_network_type(/* network_t *network */);

#define is_tree_leaf(node)		\
    (node->type == PRIMARY_INPUT || node_num_fanout(node) > 1)

#define match_is_inverter(node, prim) 	\
    (prim->ninputs == 1 && prim->inputs[0]->binding == node_get_fanin(node, 0))

#define best_match_is_inverter(node) 	\
    (MAP(node)->ninputs == 1 && 	\
	MAP(node)->save_binding[0] == node_get_fanin(node, 0))

#endif
