 /* file maputil.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:47 */
#include "mis.h"
#include "map_int.h"
#include "gate_link.h"

void
map_alloc(node)
node_t *node;
{
    if (MAP(node) != 0) map_free(node);
#ifdef _IBMR2
    node->MAP_SLOT = (char *) ALLOC(ibmr2_map_t, 1);
#else
    node->MAP_SLOT = (char *) ALLOC(map_t, 1);
#endif
    MAP(node)->type = TREE_NODE;
    MAP(node)->gate = 0;
    MAP(node)->ninputs = 0;
    MAP(node)->save_binding = 0;
    MAP(node)->arrival_info = NIL(delay_time_t *);
    MAP(node)->binding = 0;
    MAP(node)->match_node = 0;
    MAP(node)->inverter_paid_for = 0;
    MAP(node)->slack_is_critical = 0;
    MAP(node)->index = 0;
    MAP(node)->load = 0.0;
    MAP(node)->required.fall = INFINITY;
    MAP(node)->required.rise = INFINITY;
    MAP(node)->node_y = NIL(node_t);
    MAP(node)->gate_link = NIL(st_table);
    gate_link_init(node);
}


void
map_free(node)
node_t *node;
{
  if (MAP(node) != 0) {
    gate_link_free(node);
    if (MAP(node)->ninputs)
      FREE(MAP(node)->save_binding);
    if (MAP(node)->arrival_info) {
      FREE(MAP(node)->arrival_info);
      MAP(node)->arrival_info = NIL(delay_time_t *);
    }
    FREE(node->MAP_SLOT);
    node->MAP_SLOT = NIL(char);
  }
}

void map_dup(old, new)
node_t *old, *new;
{
  int i;

  if (MAP(old) == 0) return;
  map_alloc(new);
  gate_link_free(new);
  *MAP(new) = *MAP(old);
  if (MAP(old)->ninputs) {
    MAP(new)->save_binding = ALLOC(node_t *, MAP(old)->ninputs);
    for (i = 0; i < MAP(old)->ninputs; i++) {
      MAP(new)->save_binding[i] = MAP(old)->save_binding[i];
    }
  }
  MAP(new)->gate_link = NIL(st_table);
  MAP(new)->arrival_info = NIL(delay_time_t *);
  gate_link_init(new);
}


void 
map_invalid(node)
node_t *node;
{
    if (MAP(node) != 0) {
	MAP(node)->gate = 0;
    }
}


void
map_setup_network(network)
network_t *network;
{
    lsGen gen;
    node_t *node;

    foreach_node(network, gen, node) {
	map_alloc(node);
    }
}

void
map_prep(network, library)
network_t *network;
library_t *library;
{
    decomp_quick_network(network);
    if (library->nand_flag) {
	decomp_tech_network(network, /* and_limit */ 0, /* or_limit */ 2);
    } else {
	decomp_tech_network(network, /* and_limit */ 2, /* or_limit */ 0);
    }
    add_inv_network(network);
}


 /* this should be the new network */

void map_network_dup(network)
network_t *network;
{
  int i;
  lsGen gen;
  node_t *node;
  
  foreach_node(network, gen, node) {
    if (MAP(node) == 0) continue;
    if (MAP(node)->ninputs == 0) continue;
    for (i = 0; i < MAP(node)->ninputs; i++) {
      MAP(node)->save_binding[i] = (MAP(node)->save_binding[i])->copy;
    }
  }
}
