/***************************************************
*
*    Function:     cutset()
*
*    Author:        Kanwar Jit Singh
*
*****************************************************/

#include "mis.h"
#include "maxflow_int.h"

#define VLARGE (int)1000000
int	maxflow_debug;


/* ARGSUSED */
enum st_retval
speed_add_node_to_graph(node, cap, store)
char *node, **cap;
char *store;
{
    mfgptr graph;
    char name[MF_MAXSTR], dup_name[MF_MAXSTR];

    graph = (mfgptr)store;

    (void) sprintf(name,"%s",node_long_name((node_t *)node));
    (void) sprintf(dup_name,"%s_dup",node_long_name((node_t *)node));

    mf_read_node(graph, name, 0);
    mf_read_node(graph, dup_name, 0);

    return ST_CONTINUE;
}

array_t *
cutset(network,node_table)
network_t *network;
st_table *node_table;
{
    st_table *name_node_table;
    array_t *array, *df_array;
    mfgptr graph;
    mfcptr mf_cutset;
    node_t *fi, *fo, *node;
    lsGen gen;
    st_generator *gene;
    char *key, *dummy, temp_name[MF_MAXSTR], name[MF_MAXSTR];
    int j, i, capacity, terminal_flag;

    graph = mf_alloc_graph();
    name_node_table = st_init_table(strcmp,st_strhash);
    array = array_alloc(node_t *,0);

    /*
     * For each node in the network create
     * a node and also a duplicate         
     */
    mf_read_node(graph,"maxflow_source",1);
    mf_read_node(graph,"maxflow_sink",2);
    st_foreach(node_table, speed_add_node_to_graph, (char *)graph);

    /*
     * For each node create an edge to each fanin
     * which is in node_table, assign capacity 
     * to the duplicate.            
     */
    df_array = network_dfs_from_input(network);
    for(i = 0; i < array_n(df_array) ; i++){
        node = array_fetch(node_t *, df_array, i);
        if (node->type == INTERNAL){
            if(st_lookup(node_table, (char *)node, &dummy)){
                capacity = (int) dummy;
                /* Add the edge between the node and its dup */
                (void) sprintf(name,"%s",node_long_name(node));
                (void) sprintf(temp_name,"%s_dup",node_long_name(node));
                mf_read_edge(graph,name, temp_name,capacity);


                /* Save the pointer referenced by name */
                (void) st_insert(name_node_table, util_strsav(name), (char *)node);

                /* Add an edge to the source if no fanin is
                in the table or if the fanin is a primary i/p */
                terminal_flag = 1;
                foreach_fanin(node, j, fi){
                    if (st_lookup(node_table,(char *)fi, &dummy)){
                        (void) sprintf(temp_name,"%s_dup", node_long_name(fi));
                        mf_read_edge(graph, temp_name, name, VLARGE);
                        terminal_flag = 0;
                    }
                }
                if (terminal_flag){
                    mf_read_edge(graph, "maxflow_source", name, VLARGE);
                }

                /* Add  an edge to the sink if no fanout is in table */
                terminal_flag = 1;
                foreach_fanout(node, gen, fo){
                    if (st_lookup(node_table,(char *)fo, &dummy)){
                        terminal_flag = 0;
                    }
                }
                if (terminal_flag){
                    (void) sprintf(temp_name,"%s_dup", node_long_name(node));
                    mf_read_edge(graph, temp_name, "maxflow_sink", VLARGE);
                }
            }
        }
    }
    array_free(df_array);

    /* Solve and get the cutset . Translate the 
    cutset into a node array and return it */

    maxflow(graph,0);
    mf_cutset = MF_ALLOC(1, mf_cutset_t );
    mf_cutset->graph = graph;
    get_cutset(mf_cutset);

    if (maxflow_debug){
        (void) fprintf(misout,"Cutset edges:");
        for(i = 0; i < mf_cutset->narcs; i++){
            (void) fprintf(misout,"  %s(%d)", (mf_cutset->from_node)[i], (mf_cutset->capacity)[i]);
        }
        (void) fprintf(misout,"\n");
    }
    for(i = 0; i < mf_cutset->narcs; i++){
        if(st_lookup(name_node_table, (mf_cutset->from_node)[i], &dummy)){
            node = (node_t *)dummy;
            array_insert_last(node_t *, array, node);
        }
        else{
            (void) fprintf(misout,"Unknown node %s encountered \n",(mf_cutset->from_node)[i]);
        }
    }

    st_foreach_item(name_node_table, gene, &key, &dummy){
        FREE(key);
    }
    st_free_table(name_node_table);
    mf_free_cutset(mf_cutset);
    mf_free_graph(graph);
    return array;
}
