#include "mis.h"

int
net_verify_with_dc(network1, network2, method, verbose)
network_t *network1;
network_t *network2;
int method;
int verbose;
{
    network_t *dc1, *dc2;
    network_t *net1, *net2;
    network_t *or_net_dcnet();
    network_t *network_dc_network();
    int   eql;

    error_init();
    dc1= network_dc_network(network1);
    dc2= network_dc_network(network2);
    if ((dc1 == NIL (network_t)) && (dc2 == NIL (network_t))){
        eql = network_verify(network1, network2, method);
    }else if ((dc1 != NIL (network_t)) && (dc2 != NIL (network_t))){
        eql = network_verify(dc1, dc2, method);
        if (!eql){
        (void) fprintf(miserr, "%s", error_string());
        (void)fprintf(miserr,"External don't care networks are not equal.\n");
        return 1;
        }
        net1= or_net_dcnet(network1);
        net2= or_net_dcnet(network2);
        eql = network_verify(net1, net2, method);
        network_free(net1);
        network_free(net2);
    }else{
        (void)fprintf(miserr,"External don't care networks are not equal.\n");
        return 1;
    }
    if (eql) {
    if (verbose) {
        (void) fprintf(misout, "Networks compared equal.\n");
    }
    return 0;
    } else {
    (void) fprintf(miserr, "%s", error_string());
    (void) fprintf(miserr, "Networks are not equal.\n");
    return 1;
    }
}

int
network_verify(network1_t, network2_t, method)
network_t *network1_t;
network_t *network2_t;
int method;    /* 0 - collapse, 1 - bdd, others - illegal */
{
    network_t *network1, *network2;
    node_t *po1, *po2, *node1, *node2, *new_node;
    lsGen gen;
    char errmsg[1024];
    int eql;

    network1 = network_dup(network1_t);
    network2 = network_dup(network2_t);

    /* add a buffer if PO is directly connected to a PI */
    foreach_primary_output(network1, gen, po1) {
    node1 = node_get_fanin(po1, 0);
    if (node1->type == PRIMARY_INPUT) {
        new_node = node_literal(node1, 1);
        (void) network_add_node(network1, new_node);
        assert(node_patch_fanin(po1, node1, new_node));
    }
    }
    foreach_primary_output(network2, gen, po2) {
    node2 = node_get_fanin(po2, 0);
    if (node2->type == PRIMARY_INPUT) {
        new_node = node_literal(node2, 1);
        (void) network_add_node(network2, new_node);
        assert(node_patch_fanin(po2, node2, new_node));
    }
    }

    if (method == 1) {
	return bdd_verify(network1, network2);
    }

    if (method != 0) {
    (void) fprintf(miserr, "Error: unknown verification method\n");
    return FALSE;
    }

    (void) network_collapse(network1);
    (void) network_collapse(network2);

    eql = 1;


    foreach_primary_output(network1, gen, po1) {
	po2 = network_find_node(network2, po1->name);
	if ((po2 == NIL(node_t)) || (po2->type != PRIMARY_OUTPUT)) {
	    (void) sprintf(errmsg, "output '%s' only in network '%s'\n",
	    po1->name, network_name(network1));
	    error_append(errmsg);
	    eql = 0;
	} else {
	    node1 = node_get_fanin(po1, 0);
	    node2 = node_get_fanin(po2, 0);

	    if (! node_equal_by_name(node1, node2)) {
	    (void) sprintf(errmsg, "output '%s' is not the same\n", 
		po1->name);
	    error_append(errmsg);
	    eql = 0;
	    }
	}
    }

    foreach_primary_output(network2, gen, po2) {
	po1 = network_find_node(network1, po2->name);
	if ((po1 == NIL(node_t)) || (po1->type != PRIMARY_OUTPUT)) {
	    (void) sprintf(errmsg, "output '%s' only in network '%s'\n",
	    po2->name, network_name(network2));
	    error_append(errmsg);
	    eql = 0;
	}
    }

    foreach_primary_input(network1, gen, node1) {
	node2 = network_find_node(network2, node1->name);
	if ((node2 == NIL(node_t)) || (node2->type != PRIMARY_INPUT)) {
	    (void) sprintf(errmsg, "input '%s' only in network '%s'\n",
	    node1->name, network_name(network1));
	    error_append(errmsg);
	    eql = 0;
	}
    }

    foreach_primary_input(network2, gen, node2) {
	node1 = network_find_node(network1, node2->name);
	if ((node1 == NIL(node_t)) || (node1->type != PRIMARY_INPUT)) {
	    (void) sprintf(errmsg, "input '%s' only in network '%s'\n",
	    node2->name, network_name(network2));
	    error_append(errmsg);
	    eql = 0;
	}
    }

    network_free(network1);
    network_free(network2);

    return eql;
}

