#include "mis.h"
#include "pld_int.h"

static void print_partition_usage();
FILE *BDNET_FILE; 
int ACT_DEBUG;
int ACT_STATISTICS;
extern int num_or_patterns;

/* 	traverse a DAG, executing the function manipuilate at each vertex
*/
void
traverse(v, manipulate)
ACT_VERTEX_PTR v;
void (*manipulate)();

{
	v->mark =  ( v->mark == 1) ? 0 : 1 ;
	(*manipulate)(v);
	if((v->index) != v->index_size){	
	/* not a terminal vertex   */	
	  if((v->mark)!=v->low->mark) traverse(v->low, manipulate);
	  if((v->mark)!=v->high->mark) traverse(v->high, manipulate);
	}
}

/* command line interpreter for the ULM_merge command 
		-Nishizaki*/
int
com_ULM_merge(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int MAX_FANIN, MAX_COMMON_FANIN, MAX_UNION_FANIN;
    int c;
    char filename[BUFSIZE];
    int fileflag;
    int verbose;

    /* default parameters */
    /*--------------------*/
    MAX_FANIN = 4;
    MAX_COMMON_FANIN = 4;
    MAX_UNION_FANIN  = 5;
    fileflag = 0;
    verbose = 0;

    getopt_reset();
    while ((c = getopt(argc, argv, "f:c:u:o:v")) != EOF) {
	switch(c) {
	case 'f':
	    MAX_FANIN = atoi(optarg);
	    if (MAX_FANIN < 0) goto usage;
	    break;
	case 'c':
	    MAX_COMMON_FANIN = atoi(optarg);
	    if (MAX_COMMON_FANIN < 0) goto usage;
	    break;
	case 'u':
	    MAX_UNION_FANIN = atoi(optarg);
	    if (MAX_UNION_FANIN < 0) goto usage;
	    break;
	case 'o':
            (void) strcpy (filename, optarg);
	    fileflag = 1;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	default:
	    goto usage;
	}
    }
    if (argc - optind != 0) goto usage;
    if (fileflag == 0) goto usage;

    merge_node(*network, MAX_FANIN, MAX_COMMON_FANIN, MAX_UNION_FANIN, filename, verbose);
    return 0;

usage:
    (void) fprintf(miserr,
	"usage: xl_merge [-f MAX_FANIN] [-c MAX_COMMON_FANIN] [-u MAX_UNION_FANIN] [-o filename] [-v]\n");
    (void) fprintf(miserr,
	"    -f \t\tMAX_FANIN is the limit on the fanin of a mergeable node(default = 4)\n");
    (void) fprintf(miserr,
	"    -c \t\tMAX_COMMON_FANIN is the limit on the common fanins of two mergeable nodes(default = 4)\n");
    (void) fprintf(miserr,
	"    -u \t\tMAX_UNION_FANIN is the limit on the union of the fanins of two mergeable nodes(default = 5)\n");
    (void) fprintf(miserr,
	"    -o \t\tfilename is the file in which information about the nodes merged is printed. Must specify.\n");
    (void) fprintf(miserr,
	"    -v \t\tverbosity option.\n");
    return 1;



}

 /*command line interpreter for the ULM_merge command */
int
com_ULM_cover(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int		bincov_heuristics; /* if 0, exact binate covering
				      if 1, Luciano's heuristic
				      if 2, Rajeev's heuristic - fast
					    and use for large networks. 
				      if 3, automatically chooses 0 for a 
					    small cover matrix, else chooses 2 */
    int		c;
    int 	n;  /* added by Rajeev Dec. 27, 1989 - parameter to 
	      	       partition_network routine -> instead of 5 */

    bincov_heuristics = 3;
    n = 5;
    getopt_reset();
    while ((c = getopt(argc, argv, "n:h:d")) != EOF) {
	switch(c) {
	case 'n':
	    n = atoi(optarg);
	    break;
	case 'h':
	    bincov_heuristics = atoi(optarg);  /* changed by Rajeev */
	    break;
	case 'd':
	    debug = 1;
	    break;

	default:
	    goto usage;
	}
    }
    if (argc - optind != 0) goto usage;

    partition_network(*network, n, bincov_heuristics);
    return 0;

usage:
    (void) fprintf(miserr,
	"usage: xl_cover -n [number] -h [heuristic number]\n");
    (void) fprintf(miserr,
	"    -n \t\tlimit on the fanin\n");
    (void) fprintf(miserr,
	"    -h 0,1,2,3 \tuses heuristics for binate covering problem\n");
    (void) fprintf(miserr,
	"   -  0(exact), 1 (Luciano's heuristic), 2 (greedy - for large examples),\n    3 (default)(automatically switches heuristic number)\n");
    return 1;


}

/* This a routine that helps find the number of nodes with fanins greater 
 * than the value 'value'. This is used mainly to keep tabs on the fanin
 * restriction of the nodes in the network*/ 
int
com__node_value(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
   lsGen gen;
   node_t *node;
   int c, count, value;

   getopt_reset();
   count = 0;

   value = -1;
   while((c = getopt(argc, argv,"v:")) != EOF){
	switch(c){
	   case 'v':
		value = atoi(optarg);
		break;
	    default:
		(void)fprintf(misout, "usage: _xl_node_value -v \n");
		(void)fprintf(misout, "-v should be followed by the fanin limit for the node\n");
		return 1;
		break;
	}
   }
   if ( value < 1) return 1;


    foreach_node(*network, gen, node){
	if ( node_num_fanin(node) >= value){
	    count++;
	    (void)fprintf(misout,"%s:%d  ",node_name(node),
		node_num_fanin(node));
	}
   }
   (void)fprintf(misout,"\n%d nodes with fanin >= %d\n", count, value);
   return 0;
}

/* command line interpreter for the partition network command 
 * This is a routine that guarantees that the network will be 
 * partitioned into nodes with the fanin less than the value 
 * specified. This may be over kill at times! However this rou-
 * tine in very parochial so to speak as it looks only for 
 * optimal parent child relationships and DOES not look at the 
 * whole network.*/ 
int
com_partition(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int value=0;
    int c;
    getopt_reset();
    debug=0;
    while((c = getopt(argc, argv, "n:")) != EOF){
	switch(c) {
	  case 'n':
	    value =atoi(optarg);
	    break;
	  default:
	    print_partition_usage();
	    return 0;
	    break;
	}
    }
    if(value == 0){
		(void)fprintf(misout, "Error, the partition parameter must be > 0\n");
		print_partition_usage();
		return 0;
	}
    part_network(*network,value);
    (void) network_sweep(*network);
    return 0;
}

/* This routine is used to count the number of wires in the network. It
 * can be used as an indication of the routing complexity of the present
 * network. Hope to implement this in a good feedback loop for the system.
 * Currently used to keep track of wirability !*/

int
com__count_nets(network)
network_t **network;
{
    int net_number;

    getopt_reset();
    net_number = estimate_net_no(*network);
    (void)fprintf(misout, "\n the number of nets is %d \n", net_number); 
    return 0;
}


/* This routine is used for an upper bound on the clbs for the current network.
 * This does an and or decomposition and then partitions the network. It
 * hoped that this estimate is good enough . DOes not change network*/
  
com__estimate_clb(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int upper_bound, value;
    if(argc != 2) {
	(void)fprintf(misout, "Error in commamnd\n ");
	(void)fprintf(misout, "usage: _xl_clb n , n = maximum # of inputs                                             to a CLB \n");
	return 1;
    }
    value = atoi(argv[argc-1]);
    upper_bound = estimate_clb_no(*network, value);
    (void)fprintf(misout, "estimate is %d\n", upper_bound);
    return 0;
    
}
   
/* karp decomposer*/
com_karp_decomp(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
      int c;
      int support; /* cardinality of the bound set */
      node_t  *node;
      int     just_one_node;
    
      /* initialize */
      /*------------*/
      just_one_node = 0;
      node =  NULL;
      support = 5;

      getopt_reset();
      while(( c = getopt(argc, argv, "n:p:")) != EOF){
	    switch(c){
		case 'n':
		    support = atoi(optarg);
	   	    if (support < 2) {
		       (void)printf("Error: xl_k_decomp: violation*** 2 <=number: exiting \n");
	    	       return 1;
	   	    }
		    break;
		case 'p':
	            node = network_find_node(*network, optarg);
	            if (node == NULL) {
 			(void)printf("Error: xl_k_decomp: node %s not found, try altname of the node\n", 
				      optarg);
		        return 1;
	            }
		    just_one_node = 1;
		    break;

		default:
	 	    (void) fprintf(misout, "usage: xl_k_decomp [-n number] [-p nodename]\n");
	 	    (void)fprintf(misout, "-n\tnumber\tCardinality of bound set.\n");
	 	    (void) fprintf(misout, "-p\tnodename\tNode to decompose.\n");
     	            return 1;
	    }
	}
        (void)karp_decomp_network(*network, support, just_one_node, node);          
	return 0;
}

/* And or map only : takes the network and does the and or decomp and
 * partitions the network. A rather crude mapping of the network*/

com_and_or_map(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int size;

    if(argc != 2) {
	(void)fprintf(misout, "Error in command\n");
	(void)fprintf(misout, "usage: _xl_do_clb n , n = # fanin to a CLB\n");
	return 1;
    }
    size = atoi(argv[argc-1]); 
    (void)and_or_map(*network, size);
    return 0;
}

/* This routine scans the network for nodes larger than size and splits them*/ 
com_divide_network(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int size=0, c;
    getopt_reset();
    debug=0;
    
    while((c = getopt(argc, argv, "n:d")) != EOF){
		switch(c) {
	  		case 'n':
	    		size = atoi(optarg);
	    		break;
	  		case 'd':
	    		debug = 1;
	    		break;
	  		default:
	    		(void)fprintf(misout, "usage: xl_split -n -d \n");
	    		(void)fprintf(misout, "\t-n to be followed by allowed size of anin limit \n");
	    		(void)fprintf(misout, "\t-d turns debug on\n");
	    		return 0;
	    		break;
		}
    }
	if(size == 0){
		(void)fprintf(misout, "Error, split parameter must be greater than zero\n");
                (void)fprintf(misout, "usage: xl_split -n -d \n");
	   	(void)fprintf(misout, "\t-n to be followed by allowed size of anin limit \n");
	   	(void)fprintf(misout, "\t-d turns debug on\n");
		return 0;
	}
    (void)split_network(*network, size);
    return 0;
}

static void
print_partition_usage()
{
    (void) fprintf(miserr, "usage: xl_partition -n \n");
    (void) fprintf(miserr, "n - #of inputs to each partition\n");
}


com_act_main(network, argc, argv)
  network_t **network;
  int argc;
  char **argv;
{
  st_table *cost_table;
  int c;

int NUM_ITER; 	   /* maximum num of iterations before stopping - default 0*/
float GAIN_FACTOR; /* go to next iteration only if the 
		      gain > GAIN_FACTOR * network_cost - default 0.01*/
int FANIN_COLLAPSE;/* only collapse those nodes whose fanin
		       is <= FANIN_COLLAPSE - default 3*/
int DECOMP_FANIN;  /* decomp nodes with fanin >= this - default 4*/
int QUICK_PHASE;   /* used if number of iterations = 0, and you
		       do/do not wish to perform QUICK_PHASE.
			default= 0 - it will not be performed */
int NODE_REMAP;    /* break the network such that each node realized by one basic
		      block; create a bdnet like file */
int DISJOINT_DECOMP; /* do a disjoint decomposition of the network */
int HEURISTIC_NUM; /* chooses the heuristic for building the 
		      ACT. If = 1, constructs optimum ACT for nodes
		      less than 6 inputs; = 2 constructs "ACT"
		      using Narendra's heuristic (no longer ACT - 
		      cofactor tree), = 3 a combination of the two, = 4 contructs both 
		      subject graphs for each node, pick the one with lower cost.*/
int SIMPLIFY;	    /* simplifies all the nodes of the network			       */
		    /* use to have a minimum-base function at each node 	       */

char filename[BUFSIZE]; /* bdnet file name */
int cost_network;

  /* default settings */
  /*------------------*/
   NUM_ITER = 0;
   GAIN_FACTOR = 0.01;
   FANIN_COLLAPSE = 3;
   DECOMP_FANIN   = 4;
   QUICK_PHASE    = 0;
   NODE_REMAP     = 0;
   DISJOINT_DECOMP = 0;
   HEURISTIC_NUM = 2;
   ACT_STATISTICS    = 0;
   ACT_DEBUG = 0;
   MAXOPTIMAL = 6;
   SIMPLIFY = 0;
   /* input handler */
   /*---------------*/
    getopt_reset();  
    while ((c = getopt(argc, argv, "M:h:n:f:d:g:r:qDSsv")) != EOF) {
      switch (c) {
	case 'M':
		MAXOPTIMAL = atoi(optarg);
		break;
	case 'n': 
		NUM_ITER = atoi(optarg);
 	        if (NUM_ITER < 0) {
	     	  (void) printf("num_iteration(%d) should be > 0\n", NUM_ITER);
	          return 1;
		}		
	        break;
        case 'h': 
		HEURISTIC_NUM = atoi(optarg);
	        if ((HEURISTIC_NUM < 1) || (HEURISTIC_NUM > 4)) {
		     (void) printf("heuristic_num(%d) should be from 1 to 4\n", 
			      HEURISTIC_NUM);
		     return 1;
		}	
		break;
        case 'f': 
		FANIN_COLLAPSE = atoi(optarg);
	        if (FANIN_COLLAPSE < 0) {	
		       (void) printf("collapse_fanin_limit(%d) should be > 0\n", FANIN_COLLAPSE);
		       return 1;
	        }
		break;
        case 'd': 
		DECOMP_FANIN = atoi(optarg);
	        if (DECOMP_FANIN < 1) {	
		       (void) printf("decomp_fanin(%d) should be > 0\n", DECOMP_FANIN);
		       return 1;
	        }
		break;
        case 'g': 
		GAIN_FACTOR = atof(optarg);
	        if ((GAIN_FACTOR < 0) || (GAIN_FACTOR > 1)) {	
		       (void) printf("gain_factor(%f) should be from 0.0 to 1.0\n", GAIN_FACTOR);
		       return 1;
	        }
		break;
        case 'r': 
		NODE_REMAP = 1;
		(void) strcpy(filename, optarg);
	        BDNET_FILE = fopen(filename, "w");
		if (BDNET_FILE == NULL) {
			(void) printf(" **error** file %s cannot be opened\n", filename);
			return 1;
		}
		break;
        case 'q': 
		QUICK_PHASE = 1;
		break;
        case 'D': 
		DISJOINT_DECOMP = 1;
		break;
        case 'S': 
		SIMPLIFY = 1;
		break;
        case 's': 
		ACT_STATISTICS = 1;
		break;
        case 'v': 
		ACT_DEBUG = 1;
		break;
  	default : 
		(void) printf(" usage: act_map [-h heuristic_num] [-n num_iteration] [-f collapse_fanin]\n");
		(void) printf("	 \t\t[-g gain_factor] [-d decomp_fanin] [-r filename] [-M maxoptimal] [-qDsv]\n");
		(void) printf("     -h\theuristic_num\tFor making subject-graphs (default = 2).\n");
		(void) printf("     -n\tnum_iteration\tNumber of iterations (default = 0).\n");
		(void) printf("     -f\tcollapse_fanin\tUpper bound on fanin for collapse (default = 3).\n");
		(void) printf("     -d\tdecomp_fanin\tLower bound on fanin for decomposition (default = 4).\n");
		(void) printf("     -g\tgain_factor\tCriterion for going to next iteration (default = 0.01).\n");
		(void) printf("     -r\tfilename\tFinal restructuring to be done and bdnet-like file.\n");
		(void) printf("     \t\t\t       filename to be created. \n");	
		(void) printf("     -M\tmaxoptimal\tWhenever OBDD is created for a function f,\n");
		(void) printf("     \t\t\t       an optimum one is created if f has no more  than\n");	
		(void) printf("     \t\t\t       maxoptimal inputs (default = 6). \n");	
		(void) printf("     -q\t\t\tQuick phase run before exiting.\n");
		(void) printf("     -D\t\t\tDisjoint decomposition done in the beginning.\n");
		(void) printf("     -S\t\t\tSimplification done in the beginning.\n");
		(void) printf("     \t\t\t	 Amust if not done in the optimization phase earlier.\n");
		(void) printf("     -s\t\t\tStatistics option.\n");
		(void) printf("     -v\t\t\tVerbosity option.\n");
		return 1;
     } /* switch c */
  } /* while */

  cost_table = st_init_table(strcmp, st_strhash);

  *network = 
  act_map_network(*network, HEURISTIC_NUM, NUM_ITER, FANIN_COLLAPSE, GAIN_FACTOR, DECOMP_FANIN, 
		          DISJOINT_DECOMP, QUICK_PHASE, SIMPLIFY, cost_table);


   /* break the network into nodes, that can be implemented
    each by one basic block */
   /*-------------------------------------------------------*/
   cost_network = print_network(*network, cost_table);
   if (ACT_STATISTICS) 
	(void) printf(" block count: %d\n", cost_network);
   if (NODE_REMAP) {
	*network = act_break_network(*network, cost_table);
        if (ACT_STATISTICS) 
	      (void) printf(" number of blocks using OR gate = %d\n", num_or_patterns);
   }	

   (void) free_cost_table(cost_table);
    st_free_table(cost_table);
    if (!network_check(*network)) {
       (void) printf("com_act_main():%s\n", error_string());
       exit(1);
    } 
    if (NODE_REMAP) 
	(void) fclose(BDNET_FILE);
    return 0;

}



/* init routines*/
init_pld()
{
    com_add_command("act_map", com_act_main, 1); 

    node_register_daemon(DAEMON_ALLOC, p_actAlloc);
    node_register_daemon(DAEMON_FREE, p_actFree);
    node_register_daemon(DAEMON_DUP, p_actDup);

    com_add_command("xl_partition", com_partition, 1);
    com_add_command("xl_merge", com_ULM_merge, 0);
    com_add_command("xl_cover", com_ULM_cover, 1);
    com_add_command("xl_k_decomp", com_karp_decomp, 1);
    com_add_command("xl_split", com_divide_network, 1);
    com_add_command("_xl_nodevalue", com__node_value, 0);
    com_add_command("_xl_cnets", com__count_nets, 0); 
    com_add_command("_xl_clb", com__estimate_clb, 0);
    com_add_command("_xl_do_clb", com_and_or_map, 1);
}

end_pld()
{
}

