#include "mis.h"
#include "pld_int.h"

static int combine();
static int cost_fanin();
static void search_part();
  
part_network(network, size) 
network_t *network;
int size;
{
    array_t *order;
    node_t *node, *fout ;
    int i, inp=0, composite_fanin();
    
/*Collapse simple nodes with single fanout into larger nodes*/
  
    order = network_dfs(network);
    for( i=0; i < array_n(order); i++) {
	node = array_fetch(node_t *, order, i);
	if(node_num_fanout(node) == 1) {
	    fout = node_get_fanout(node, 0);
	    inp = composite_fanin(fout, node);
	    if (inp < size) {
		if (node_collapse(fout, node)) {
		    network_delete_node(network, node);
		}
	    }
	}

    }
    array_free(order);
    search_part(network, size);
}



/*select good partitioning of network*/ 
static void
search_part(network, size)
network_t *network;
int size;
{
    node_t *node;
    int loop, i;
   
    loop = 1 ;
   /* (void)split_network(network, size);*/
    while (loop) {
 /*Try and decompose nodes that have a fanout greater than size */
	loop = combine(network, size);
    }
    for(i = network_num_po(network)-1; i>= 0; i--) {
	node = node_get_fanin(network_get_po(network, i), 0);
	if(node_num_fanin(node) > size) {
	    decomp_good_node(network, node);
	}
    }
}

static int
combine(network, size)
network_t *network;
int size;  
{
    static int compare();
    array_t *order_f, *order;
    int j, value=0, num;
    node_t *po, *fanin, *fanout;
    int i, composite_fanin();
    a_node  *node;
    char *node1, *node2;
    lsGen gen;

    order = network_dfs(network);
    order_f = array_alloc(a_node *, 0);
    for(j = 0; j< array_n(order);j++) {
	po = array_fetch(node_t *, order, j);
	if(node_function(po) != NODE_PI) { 
	    foreach_fanin(po, i, fanin) {
		node = ALLOC(a_node, 1);
		num = 0;
		foreach_fanout(fanin, gen, fanout) {
		    num +=  cost_fanin(fanout, fanin);
		    /*(void)fprintf(misout, "the node is ");
		    (void)node_print(misout, fanin);
		    (void)fprintf(misout, "num is %d\n", num);*/
		}
		node->value = (float)num;
		node->name = node_long_name(fanin);
		node->fanout = node_long_name(po);
		(void)array_insert_last(a_node *, order_f, node);
	    }
	}
    }
/* sort in increasing order*/
    array_sort(order_f, compare);
	i=0;
    while((i< array_n(order_f))&&(!value)) { 
	node1 = array_fetch(a_node *, order_f, i)->name;
	fanin  = network_find_node(network, node1);
	node2 = array_fetch(a_node *, order_f, i)->fanout;
	po = network_find_node(network, node2);
	if(composite_fanin(po, fanin) <= size) {
	    if(node_collapse(po, fanin)) {
		value = 1;
		break;
	    }
	}
	i++; 
    }
    array_free(order);
    for(i = 0; i < array_n(order_f);i++) {
	node = array_fetch(a_node *, order_f, i);
	FREE(node);
    }
    array_free(order_f);
    return(value);
   }


/* Finds number of distinct fanins of fanin- fanin is a fanin to po
 * ie should fanin be merged into po then the number of wires that 
 * would have to be routed to po in addition to the fanins of po*/
static int
cost_fanin(po, fanin)
node_t *po, *fanin;
{
    node_t *fi;
    int j, cost;

    cost = node_num_fanin(fanin);
    foreach_fanin(po, j, fi) {
	if(node_get_fanin_index(fanin, fi) >= 0) cost--;
    }
    return(cost);
}


int 
composite_fanin(node1, node2)
node_t *node1, *node2 ;
{
node_t *fi;
int inp, j;

inp = node_num_fanin(node1)-1;
   foreach_fanin(node2, j, fi) {
       if(node_get_fanin_index(node1, fi) < 0) inp++;
   }
return(inp);
}
 	


static int 
compare(obj1, obj2)
char **obj1, **obj2;
{
    a_node *node1 = *(a_node **)obj1;
    a_node *node2 = *(a_node **)obj2;
    int value;

    if((node1->value) == (node2->value)) value = 0;
    if((node1->value) >  (node2->value)) value = 1;
    if((node1->value) <  (node2->value)) value = -1;
    return(value);
}


