#include "mis.h"
#include "pld_int.h"

static int split_node();
static void find_best_div();

split_network(network, size)
network_t *network;
int size;
{
    array_t *order;
    node_t *node;
    int i ; 
    char *err;

    order = network_dfs(network);
    for(i=0 ; i < array_n(order); i++) {
	node = array_fetch(node_t *, order, i);
	if(node_num_fanin(node) > size) {
	    (void) split_node(network, node, size);
	    error_init();
	    if(!network_check(network)) {
		err = error_string();
		(void)fprintf(misout, "%s", err);
	    }
	}
    }
    array_free(order);
    (void) network_sweep(network);
}



static int
split_node(network, node, size) 
network_t *network;
node_t *node;
int size;
{
    static int kernel_value();
    int i, value = 1;
    kern_node *sorted;
    divisor_t *div, *best_div;
    node_t  *best, *co_best, *rem, *t_node;
   
    /* First try finding kernels */
    if((t_node = ex_find_divisor_quick(node)) != NIL(node_t))  {
	sorted = ALLOC(kern_node, 1);
	sorted->node = node;
	sorted->cost_array = array_alloc(divisor_t *, 0);
	ex_kernel_gen(node, kernel_value, (char *)sorted);
	ex_subkernel_gen(node, kernel_value, 0, (char *)sorted);
	best_div = ALLOC( divisor_t, 1);
	best_div->divisor = NIL(node_t);
	best_div->kernelsize = 0;

	for( i=0; i< array_n(sorted->cost_array); i++){
	    div = array_fetch(divisor_t *, sorted->cost_array, i);
	    (void)find_best_div(div, best_div, size);
	}
    
	array_free(sorted->cost_array);
	FREE(sorted);

	if(best_div->divisor != NIL(node_t)) {
	    best = best_div->divisor;
	    FREE(best_div);
	    if(debug){
		(void)fprintf(misout,"Node %s has been added to the network\n",
	      node_name(best));
		(void)node_print(misout, best);
	    }
	    co_best = node_div(node, best, &rem);
	    if(debug){
		(void)fprintf(misout, "cokernel");
		node_print(misout, co_best);
		(void)fprintf(misout, "remainder");
		node_print(misout, rem);
	    }
	    network_add_node(network, best);
	    network_add_node(network, co_best);
	    if (node_function(rem) != NODE_0) {
		network_add_node(network, rem);
		(void)node_substitute(node, rem, 1);
		if(node_num_fanin(rem) > size) {
		    (void) split_node(network, rem, size);
		}
	    } else {
		node_free(rem);
	    }
	    (void)node_substitute(node, best, 1);
	    (void)node_substitute(node, co_best, 1);
	    if(node_num_fanin(co_best) > size) {
		(void) split_node(network, co_best, size);
	    }
	    if(debug) {
		node_print(misout, co_best);
	    }
	} else {
	    node_free(best_div->divisor);
	    FREE(best_div);
	    value = pld_decomp_and_or(network, node, size);
	}
    } else {
	if( !pld_decomp_and_or(network, node, size)){
	    (void)fprintf(misout,"No and_or decomposition possible\n");
	    value = 0;
	}
	    
    }
    node_free(t_node);
    return (value);
}


    
static int
kernel_value(kernel, cokernel, state)
node_t *kernel;
node_t *cokernel;
char *state;
{
    divisor_t *k_data, *c_data;
    kern_node *store; 
    node_t *rem , *new_cokernel;
 
    store = (kern_node *)state;
    new_cokernel = node_div(store->node, kernel, &rem);
    c_data = ALLOC(divisor_t, 1);
    c_data->divisor = new_cokernel;
    c_data->kernelsize = composite_fanin(kernel, new_cokernel) + 
                                                 node_num_fanin(rem); 
    k_data = ALLOC(divisor_t, 1);
    k_data->divisor = kernel;
    k_data->kernelsize = composite_fanin(kernel,new_cokernel) +
                                                 node_num_fanin(rem);
    array_insert_last(divisor_t *, store->cost_array, k_data);
    array_insert_last(divisor_t *, store->cost_array, c_data);
    node_free(rem);
    node_free(cokernel);
}


static void
find_best_div(div, best_div, size)
divisor_t *div, *best_div;
int size;
{
    if(( div->kernelsize > best_div->kernelsize)&&(node_num_fanin(div->divisor)
					   <= size)) {
	if(best_div->divisor != NIL(node_t )){
	    node_free(best_div->divisor);
	}
	best_div->divisor = node_dup(div->divisor);
	best_div->kernelsize = div->kernelsize;
    } 
    node_free(div->divisor);
    FREE(div);
}



