#include "mis.h"
#include "sim_int.h"


static int
dump_vector(fp, vec)
FILE *fp;
array_t *vec;
{
    int i;

    for(i = 0; i < array_n(vec); i++) {
       (void) fprintf(fp, " %d", array_fetch(int, vec, i));
    }
    (void) fprintf(fp, "\n");
}



static int
com_simulate(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i;
    array_t *in_value, *out_value, *node_vec;

    if (argc - 1 != network_num_pi(*network)) {
	(void) fprintf(miserr, 
	    "simulate: network has %d inputs; %d values were supplied.\n",
	    network_num_pi(*network), argc - 1);
	goto usage;
    }

    in_value = array_alloc(int, 10);
    for(i = 1; i < argc; i++) {
	if (strcmp(argv[i], "0") == 0) {
	    array_insert_last(int, in_value, 0);
	} else if (strcmp(argv[i], "1") == 0) {
	    array_insert_last(int, in_value, 1);
	} else {
	    (void) fprintf(miserr, 
		"simulate: bad value '%s' -- should be 0 or 1\n", argv[i]);
	    array_free(in_value);
	    goto usage;
	}
    }

#if 1
    node_vec = network_dfs(*network);
    out_value = simulate_network(*network, node_vec, in_value);
    array_free(node_vec);
#else
    {
	int i, j;
	network_t *net1;
	node_t *node;
	lsGen gen;

	net1 = network_dup(*network);

	i = 0;
	foreach_primary_input(net1, gen, node) {
	    j = array_fetch(int, in_value, i++);
	    node_replace(node, j ? node_constant(1) : node_constant(0));
	}
	(void) network_collapse(net1);
	out_value = array_alloc(int, 0);
	foreach_primary_output(net1, gen, node) {
	    j = node_function(node_get_fanin(node, 0)) == NODE_1;
	    array_insert_last(int, out_value, j);
	}
	network_free(net1);
    }
#endif

    dump_vector(misout, out_value);

    array_free(in_value);
    array_free(out_value);
    return 0;

usage:
    (void) fprintf(miserr, "usage: simulate in1 in2 in3 ...\n");
    return 1;
}

static int
com_sim_verify(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int c;
  int n_patterns = 1024;
  network_t *network1;
  char buffer[1024];
  int status;

  getopt_reset();
  while ((c = getopt(argc, argv, "n:")) != EOF) {
    switch (c) {
    case 'n':
      n_patterns = atoi(optarg);
      if (n_patterns < 1) goto usage;
      if (n_patterns % 32 != 0)
	n_patterns += (32 - (n_patterns % 32));
      break;
    default:
      goto usage;
    }
  }
  if (argc - optind != 1) goto usage;
  (void) sprintf(buffer, "read_blif %s", argv[optind]);
  network1 = network_alloc();
  if (com_execute(&network1, buffer) != 0) {
    network_free(network1);
    return 1;
  }
  if (network_num_pi(network1) != network_num_pi(*network)) {
    (void) fprintf(miserr, "Number of inputs do not agree\n");
    network_free(network1);
    return 1;
  }
  if (network_num_po(network1) != network_num_po(*network)) {
    (void) fprintf(miserr, "Number of outputs do not agree\n");
    network_free(network1);
    return 1;
  }

  status = sim_verify(*network, network1, n_patterns);
  network_free(network1);
  return status;

 usage:
  (void) fprintf(miserr, "usage: sim_verify [-n n_patterns] network2.blif\n");
  return 1;
}

init_sim()
{
    com_add_command("simulate", com_simulate, 0);
    com_add_command("sim_verify", com_sim_verify, 0);
}


end_sim()
{
}
