#include "mis.h"
#include "simp_int.h"

int com_simplify();
int com_simp();
int com_simall();
int com_myps();
int com_cspf();
static bool get_cone_levels();

init_simplify()
{
    simp_trace = FALSE;
    simp_debug = FALSE;

    node_register_daemon(DAEMON_FREE, simp_free);
    node_register_daemon(DAEMON_ALLOC, simp_alloc);
    node_register_daemon(DAEMON_DUP, simp_dup);
    node_register_daemon(DAEMON_INVALID, simp_invalid);

    com_add_command("simplify", com_simplify, 1);
    com_add_command("_simp", com_simp, 1);
    com_add_command("_simall", com_simall, 1);
    com_add_command("_myps", com_myps, 0);
    com_add_command("cspf_simplify", com_cspf, 0);
}

end_simplify()
{
}

static void
simp_usage(command)
char *command;
{
    (void) fprintf(miserr, "usage: %s [-d]", command);
    (void) fprintf(miserr, "[-i <num>[:<num>]] [-m method] [-f filter] ");
    (void) fprintf(miserr, "[node-list]\n");
    (void) fprintf(miserr, "    -d\tNo don't care set ");
    (void) fprintf(miserr, "(default is a subset of the fanin don't cares)\n");
    (void) fprintf(miserr, "    -i <num>[:<num>]\tGenerate fanin don't-care");
    (void) fprintf(miserr, " for nodes specified in cone\n");
    (void) fprintf(miserr, "    -m snocomp\tSingle pass minimizer (default)\n");
    (void) fprintf(miserr, "    -m nocomp\tMultiple pass minimizer\n");
    (void) fprintf(miserr, "    -m dcsimp\tAnother single pass minimizer\n");
    (void) fprintf(miserr, "    -f exact\tExact filter (default)\n");
    (void) fprintf(miserr, "    -f disj_sup\tDisjoint support filter\n");
}

int
com_simplify(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, i; 
    array_t *nodevec;
    node_t *np;
    sim_accept_t accept;
    sim_method_t method;
    sim_dctype_t dctype;
    sim_filter_t filter;

    dctype = SIM_DCTYPE_SUB_FANIN;
    accept = SIM_ACCEPT_FCT_LITS;
    method = SIM_METHOD_SNOCOMP;
    filter = SIM_FILTER_EXACT;
    getopt_reset();
    simp_debug = FALSE;
    simp_trace = FALSE;
    simp_fanin_level = 1;
    simp_fanin_fanout_level = 0;
    while ((c = getopt(argc, argv, "i:f:m:td")) != EOF) {
	switch (c) {
	case 'd':
	    dctype = SIM_DCTYPE_NONE;
	    break;
        case 'm':
	    if (strcmp(optarg, "nocomp") == 0) {
		    method = SIM_METHOD_NOCOMP;
	    } else if (strcmp(optarg, "snocomp") == 0) {
		    method = SIM_METHOD_SNOCOMP;
	    } else if (strcmp(optarg, "dcsimp") == 0) {
		    method = SIM_METHOD_DCSIMP;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'f':
	    if (strcmp(optarg, "exact") == 0) {
		filter = SIM_FILTER_EXACT;
	    } else if (strcmp(optarg, "disj_sup") == 0) {
		filter = SIM_FILTER_DISJ_SUPPORT;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'i':
	    dctype = SIM_DCTYPE_FANIN;
	    if (!get_cone_levels(optarg, &simp_fanin_level,
				 &simp_fanin_fanout_level)) {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 't':
	    simp_trace = TRUE;
	    break;
	default:
	    simp_usage(argv[0]);
	    return 1;
	}
    }

    nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
    nodevec = simp_order(nodevec);
    for(i = 0; i < array_n(nodevec); i++) {
	np = array_fetch(node_t *, nodevec, i); 
	if (np->type == INTERNAL) {
     	simplify_node(np, method, dctype, accept, filter); 
	}
    }
    array_free(nodevec);
    return 0;
}

int
com_simp(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, i; 
    array_t *nodevec;
    node_t *np;
    sim_method_t method;
    sim_accept_t accept;
    sim_dctype_t dctype;
    sim_filter_t filter;

    dctype = SIM_DCTYPE_FANIN;
    accept = SIM_ACCEPT_FCT_LITS;
    method = SIM_METHOD_SNOCOMP;
    filter = SIM_FILTER_EXACT;
    getopt_reset();
    simp_debug = FALSE;
    simp_trace = FALSE;
    simp_fanin_level = 1;
    simp_fanin_fanout_level = 0;
    while ((c = getopt(argc, argv, "a:d:m:i:o:f:Dt")) != EOF) {
	switch (c) {
	case 'a':
	    if (strcmp(optarg, "fct_lits") == 0) {
		accept = SIM_ACCEPT_FCT_LITS;
	    } else if (strcmp(optarg, "sop_lits") == 0) {
		accept = SIM_ACCEPT_SOP_LITS;
	    } else if (strcmp(optarg, "cubes") == 0) {
		accept = SIM_ACCEPT_CUBES;
	    } else if (strcmp(optarg, "always") == 0) {
		accept = SIM_ACCEPT_ALWAYS;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'd':
	    if (strcmp(optarg, "none") == 0) {
		dctype = SIM_DCTYPE_NONE;
	    } else if (strcmp(optarg, "fanin") == 0) {
		dctype = SIM_DCTYPE_FANIN;
	    } else if (strcmp(optarg, "fanout") == 0) {
		dctype = SIM_DCTYPE_FANOUT;
	    } else if (strcmp(optarg, "inout") == 0) {
		dctype = SIM_DCTYPE_INOUT;
	    } else if (strcmp(optarg, "support") == 0) {
		dctype = SIM_DCTYPE_SUB_FANIN;
	    } else if (strcmp(optarg, "all") == 0) {
		dctype = SIM_DCTYPE_ALL;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'm':
	    if (strcmp(optarg, "simpcomp") == 0) {
		method = SIM_METHOD_SIMPCOMP;
	    } else if (strcmp(optarg, "espresso") == 0) {
		method = SIM_METHOD_ESPRESSO;
	    } else if (strcmp(optarg, "exact") == 0) {
		method = SIM_METHOD_EXACT;
	    } else if (strcmp(optarg, "min_lit") == 0) {
		method = SIM_METHOD_EXACT_LITS;
	    } else if (strcmp(optarg, "dcsimp") == 0) {
		method = SIM_METHOD_DCSIMP;
	    } else if (strcmp(optarg, "nocomp") == 0) {
		method = SIM_METHOD_NOCOMP;
	    } else if (strcmp(optarg, "snocomp") == 0) {
		method = SIM_METHOD_SNOCOMP;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'f':
	    if (strcmp(optarg, "exact") == 0) {
		filter = SIM_FILTER_EXACT;
	    } else if (strcmp(optarg, "disj_sup") == 0) {
		filter = SIM_FILTER_DISJ_SUPPORT;
	    } else if (strcmp(optarg, "size") == 0) {
		filter = SIM_FILTER_SIZE;
            } else if (strcmp(optarg, "fdist") == 0) {
		filter = SIM_FILTER_FDIST;
	    } else if (strcmp(optarg, "sdist") == 0) {
	        filter = SIM_FILTER_SDIST;
	    } else if (strcmp(optarg, "none") == 0) {
		filter = SIM_FILTER_NONE;
	    } else if (strcmp(optarg, "all") == 0) {
		filter = SIM_FILTER_ALL;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'i':
	    simp_fanin_level = atoi(optarg);
	    switch (dctype) {
	    case SIM_DCTYPE_FANOUT:
		dctype = SIM_DCTYPE_INOUT;
		break;
	    case SIM_DCTYPE_INOUT:
	    case SIM_DCTYPE_ALL:
		break;
	    default:
		dctype = SIM_DCTYPE_FANIN;
	    }
	    break;
	case 'o':
	    simp_fanin_fanout_level = atoi(optarg);
	    switch (dctype) {
	    case SIM_DCTYPE_FANOUT:
		dctype = SIM_DCTYPE_INOUT;
		break;
	    case SIM_DCTYPE_INOUT:
	    case SIM_DCTYPE_ALL:
		break;
	    default:
		dctype = SIM_DCTYPE_FANIN;
	    }
	    break;
	case 't':
	    simp_trace = TRUE;
	    break;
	case 'D':
	    simp_debug = TRUE;
	    break;
	default:
	    simp_usage(argv[0]);
	    return 1;
	}
    }

    nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
    for(i = 0; i < array_n(nodevec); i++) {
	np = array_fetch(node_t *, nodevec, i); 
	if (np->type == INTERNAL) {
     	simplify_node(np, method, dctype, accept, filter); 
	}
    }
    array_free(nodevec);
    return 0;
}

int
com_simall(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c;
    sim_method_t method;
    sim_accept_t accept;
    sim_dctype_t dctype;
    sim_filter_t filter;

    dctype = SIM_DCTYPE_ALL;
    accept = SIM_ACCEPT_FCT_LITS;
    method = SIM_METHOD_ESPRESSO;
    filter = SIM_FILTER_EXACT;
    simp_debug = FALSE;
    simp_trace = FALSE;
    simp_fanin_level = 1;
    simp_fanin_fanout_level = 0;
    getopt_reset();
    while ((c = getopt(argc, argv, "a:d:m:i:o:f:Dt")) != EOF) {
	switch (c) {
	case 'a':
	    if (strcmp(optarg, "fct_lits") == 0) {
		accept = SIM_ACCEPT_FCT_LITS;
	    } else if (strcmp(optarg, "sop_lits") == 0) {
		accept = SIM_ACCEPT_SOP_LITS;
	    } else if (strcmp(optarg, "cubes") == 0) {
		accept = SIM_ACCEPT_CUBES;
	    } else if (strcmp(optarg, "always") == 0) {
		accept = SIM_ACCEPT_ALWAYS;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'd':
	    if (strcmp(optarg, "none") == 0) {
		dctype = SIM_DCTYPE_NONE;
	    } else if (strcmp(optarg, "fanin") == 0) {
		dctype = SIM_DCTYPE_FANIN;
	    } else if (strcmp(optarg, "fanout") == 0) {
		dctype = SIM_DCTYPE_FANOUT;
	    } else if (strcmp(optarg, "inout") == 0) {
		dctype = SIM_DCTYPE_INOUT;
	    } else if (strcmp(optarg, "support") == 0) {
		dctype = SIM_DCTYPE_SUB_FANIN;
	    } else if (strcmp(optarg, "all") == 0) {
		dctype = SIM_DCTYPE_ALL;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'm':
	    if (strcmp(optarg, "simpcomp") == 0) {
		method = SIM_METHOD_SIMPCOMP;
	    } else if (strcmp(optarg, "espresso") == 0) {
		method = SIM_METHOD_ESPRESSO;
	    } else if (strcmp(optarg, "exact") == 0) {
		method = SIM_METHOD_EXACT;
	    } else if (strcmp(optarg, "min_lit") == 0) {
		method = SIM_METHOD_EXACT_LITS;
	    } else if (strcmp(optarg, "dcsimp") == 0) {
		method = SIM_METHOD_DCSIMP;
	    } else if (strcmp(optarg, "nocomp") == 0) {
		method = SIM_METHOD_NOCOMP;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'f':
	    if (strcmp(optarg, "exact") == 0) {
		filter = SIM_FILTER_EXACT;
	    } else if (strcmp(optarg, "disj_sup") == 0) {
		filter = SIM_FILTER_DISJ_SUPPORT;
	    } else if (strcmp(optarg, "size") == 0) {
		filter = SIM_FILTER_SIZE;
	    } else if (strcmp(optarg, "fdist") == 0) {
	        filter = SIM_FILTER_FDIST;
	    } else if (strcmp(optarg, "sdist") == 0) {
		filter = SIM_FILTER_SDIST;
	    } else if (strcmp(optarg, "none") == 0) {
		filter = SIM_FILTER_NONE;
	    } else if (strcmp(optarg, "all") == 0) {
		filter = SIM_FILTER_ALL;
	    } else {
		simp_usage(argv[0]);
		return 1;
	    }
	    break;
	case 'i':
	    simp_fanin_level = atoi(optarg);
	    switch (dctype) {
	    case SIM_DCTYPE_FANOUT:
		dctype = SIM_DCTYPE_INOUT;
		break;
	    case SIM_DCTYPE_INOUT:
	    case SIM_DCTYPE_ALL:
		break;
	    default:
		dctype = SIM_DCTYPE_FANIN;
	    }
	    break;
	case 'o':
	    simp_fanin_fanout_level = atoi(optarg);
	    switch (dctype) {
	    case SIM_DCTYPE_FANOUT:
		dctype = SIM_DCTYPE_INOUT;
		break;
	    case SIM_DCTYPE_INOUT:
	    case SIM_DCTYPE_ALL:
		break;
	    default:
		dctype = SIM_DCTYPE_FANIN;
	    }
	    break;
	case 'D':
	    simp_debug = TRUE;
	    break;
	case 't':
	    simp_trace = TRUE;
	    break;
	default:
	    simp_usage(argv[0]);
	    return 1;
	}
    }

    if (argc - optind == 0) {
	simplify_all(*network, method, dctype, accept, filter);
	return 0;
    } else {
	simp_usage(argv[0]);
	return 1;
    }
}

/* ARGSUSED */
com_myps(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int pi, po, sl, nd, fl;
    node_t *np;
    lsGen gen;

    pi = network_num_pi(*network);
    po = network_num_po(*network);
    nd = network_num_internal(*network);

    sl = fl = 0;
    foreach_node(*network, gen, np) {
	if (np->type == INTERNAL) {
	    sl += node_num_literal(np);
	    fl += factor_num_literal(np);
	}
    }

    (void) fprintf(misout, "%d\t%d\t%d\t%d\t%d\n", pi, po, nd, sl, fl);
}

static bool
get_cone_levels(name, par1, par2)
char *name;
int *par1, *par2;
{
    char *s, *str, *str1 = NIL(char), *str2 = NIL(char);

    /* Check for silly input */
    if (name == NIL(char) || strcmp(name, "") == 0) 
	return FALSE;
    if (strchr(name, ' ') != NIL(char)) 
	return FALSE;
	
    /* Copy the input string (because we will tear it apart) */ 
    str = ALLOC(char, strlen(name)+1);
    s = strcpy(str, name);

    if (s[0] != ':') str1 = s;
    if ((s = strchr(s, ':')) != NIL(char)) {
	*s++ = '\0';
	str2 = s;
    }

    if (str1 != NIL(char))
        *par1 = atoi(str1);
    if (str2 != NIL(char))
        *par2 = atoi(str2);
    FREE(str);
    return TRUE;
}


int
com_cspf(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c; 
    sim_accept_t accept;
    sim_method_t method;
    sim_dctype_t dctype;
    sim_filter_t filter;

    dctype = SIM_DCTYPE_INOUT;
    accept = SIM_ACCEPT_FCT_LITS;
    method = SIM_METHOD_SNOCOMP;
    filter = SIM_FILTER_EXACT;
    getopt_reset();
    simp_debug = FALSE;
    simp_trace = FALSE;
    simp_fanin_level = 1;
    simp_fanin_fanout_level = 0;
    while ((c = getopt(argc, argv, "i:f:m:td")) != EOF) {
    switch (c) {
    case 'd':
        dctype = SIM_DCTYPE_NONE;
        break;
        case 'm':
        if (strcmp(optarg, "nocomp") == 0) {
    	    method = SIM_METHOD_NOCOMP;
        } else if (strcmp(optarg, "snocomp") == 0) {
    	    method = SIM_METHOD_SNOCOMP;
        } else if (strcmp(optarg, "dcsimp") == 0) {
    	    method = SIM_METHOD_DCSIMP;
        } else {
    	return 1;
        }
        break;
    case 'f':
        if (strcmp(optarg, "exact") == 0) {
    	filter = SIM_FILTER_EXACT;
        } else if (strcmp(optarg, "disj_sup") == 0) {
    	filter = SIM_FILTER_DISJ_SUPPORT;
        } else if (strcmp(optarg, "none") == 0) {
    	filter = SIM_FILTER_NONE;
        } else {
    	return 1;
        }
        break;
    case 'i':
    	simp_fanin_level= atoi(optarg);
        break;
    case 't':
        simp_trace = TRUE;
        break;
    default:
        return 1;
    }
    }

    comp_perm_fn(network, accept, method, dctype, filter);
    return 0;
}

