#include "mis.h"
#include "simp_int.h"

static node_t *compute_mspf(); 
static node_t *compute_cspf();
static st_table *cspf_inout_table();
static void cspf_alloc();
static void cspf_free();

/* Compute a set of permissible functions for each of the nodes of the
 * network. Use the CSPF and some subset of the satisfiability don't
 * cares to simplify the node using a two-level minimizer.
 */
void
comp_perm_fn(networkptr, accept, method, dctype, filter)
network_t **networkptr;
sim_accept_t accept;
sim_method_t method;
sim_dctype_t dctype;
sim_filter_t filter;
{
    network_t *network;
    array_t *network_node_list;/*an array of the nodes in the network*/
    array_t *fanin_list, *fanout_list;
    array_t *node_list;
    int        numin,    array_size;
    int        i,j,k,m,l;
    node_t  *p, *q, *node, *fanout;
    node_t    *pfanin, *nd;
    node_t  *dcfanin, *dcnode;
    node_t    *pi, *FDC;
    node_t    *ANDpi; /* ANDing of all the primary inputs*/
    st_table *node_table1, *node_table2;
    st_table *node_table3;
    lsGen    gen;
    char    *dummy;
    st_table *node_exdc_table;

    network= *networkptr;

/*create a new node by ANDing all the PIs. This is used for filtering*/
    ANDpi= node_constant(1);
    foreach_primary_input(network, gen, pi){
        p= node_literal(pi, 1);
        q= node_and(p, ANDpi);
        node_free(p);
        node_free(ANDpi);
        ANDpi= q;
    }

    node_exdc_table = attach_dcnetwork_to_network(network);

/*allocate space for CSPFs; Get the external don't cares for Primary Outputs*/
    foreach_node(network, gen,node){
        if (node_function(node) != NODE_PI)
            cspf_alloc(node);
        if (node_function(node) == NODE_PO){
            CSPF(node)->node= find_ex_dc(node, node_exdc_table);
        }
    }


    network_node_list= network_dfs_from_input(network);
    array_size= array_n(network_node_list);

/*compute CSPF and simplify each node*/
    for(i= 0; i< array_size; i++){
        node= array_fetch(node_t *, network_node_list, i);
        if (node_function(node) == NODE_PI) 
            continue;
        if (node_function(node) == NODE_PO) 
            continue;
        if (node_num_fanin(node) == 0){
            CSPF(node)->node= node_constant(0);
            continue;
        }
        fanout_list= network_tfo(node, 1);

        node_table1= cspf_inout_table(node, INFINITY, 1);
        node_table3= cspf_inout_table(node, 2, 1);
        node_list= network_tfi(node, simp_fanin_level);
        node_table2 = st_init_table(st_ptrcmp, st_ptrhash);
        for(m= 0; m < array_n(node_list) ; m++){
            nd= array_fetch(node_t *, node_list, m);
            (void) st_insert(node_table2, (char *)nd, NIL(char));
        }

        FDC= NULL;    

/* compute CSPF for each of the fanout edges; intersect them to
 * find the CSPF for the node.
 */
        for(j= 0; j < array_n(fanout_list) ; j++){
            fanout= array_fetch(node_t *, fanout_list, j);

            dcnode= compute_mspf(fanout, node, node_table1, ANDpi);

            fanin_list= network_tfi(fanout, 1);

/* make the mspf compatible with the cspf computed for previous edges. */
            for(k= 0 ; k < array_n(fanin_list); k++){
                pfanin= array_fetch(node_t *, fanin_list, k);
                if (node_function(pfanin) != NODE_PI) 
                dcnode= compute_cspf(dcnode, pfanin, node_table2, 
                    ANDpi, node_table3);
            }
            array_free(fanin_list);

/* OR the computed CSPF with the CSPF of the fanout node.*/
            if (CSPF(fanout)->node != NULL){
                q= dcnode;
                dcnode= node_or(CSPF(fanout)->node, q);
                node_free(q);
            }

/* Filter the CSPF and collapse all the nodes that are not transitive
 * fanins of the node being simplified.
 */
            if (node_num_fanin(dcnode) > 0){
                dcnode= simp_obsdc_filter(dcnode, node_table2, 3,ANDpi);
                numin =node_num_fanin(dcnode);
                for(l=0; l< numin; l++){
                    dcfanin= node_get_fanin(dcnode, l);
                    if (!st_lookup(node_table2,(char *)dcfanin,&dummy)){
                    if (node_function(dcfanin) != NODE_PI){
                    if (st_lookup(node_table3,(char *)dcfanin,&dummy)){
                        (void)node_collapse(dcnode, dcfanin);
                        l--;
                        numin= node_num_fanin(dcnode);
                    }
                    }
                    }
                }
                dcnode= simp_obsdc_filter(dcnode, node_table2, 3,ANDpi);
            }

/* Do the intersection to find the CSPF of the node */
            if (FDC == NULL) {
                FDC= dcnode;
            }else{
                p = node_and (FDC, dcnode);
                node_free(dcnode);
                node_free(FDC);
                FDC= p;
            }
        }
        st_free_table(node_table1);
        st_free_table(node_table3);
        st_free_table(node_table2);
        array_free(node_list);
        array_free(fanout_list);

/* store the CSPF */
        if (FDC != NULL) {
            CSPF(node)->node= FDC;
        }else{
            CSPF(node)->node= node_constant(0);
        }

/* simplify the node */
        if (node->type== INTERNAL){ 
            simplify_cspf_node(node, method,dctype,accept,filter);
            dctype = SIM_DCTYPE_INOUT;
        }
    }

/* free the space allocated for CSPFs*/
    foreach_node(network, gen,node){
        if (node_function(node)!= NODE_PI){
            node_free(CSPF(node)->node);
            cspf_free(node);
        }
    }
    node_free(ANDpi);
    array_free(network_node_list);
    if (node_exdc_table != NIL(st_table)) {
	st_free_table(node_exdc_table);
    }
}

/* allocate space for CSPFs.
 *
 */
static void 
cspf_alloc(node)
node_t    *node;
{
    node->OBS = (char *)  ALLOC(cspf_type_t, 1);
    CSPF(node)->level= 0;
    CSPF(node)->node= NULL;
}


/* free space allocated for CSPFs.
 *
 */
static void 
cspf_free(node)
node_t    *node;
{
    FREE(node->OBS);
}


/* compute maximum set of permissible functions for an edge.
 *
 */
static node_t *
compute_mspf(node, fanin, node_table,ANDpi)
node_t *node;
node_t *fanin;
st_table *node_table; /*used for filtering*/
node_t *ANDpi;/*ANDing of the PIs used for filtering*/

{
    node_t *p, *q, *fneg, *fpos, *DC;

    if (node_function(node) == NODE_PO) 
        return(node_constant(0));    
    if (node_function(fanin) == NODE_PI) 
        return(node_constant(0));    

    fpos= node_literal(fanin, 1);
    fneg= node_literal(fanin, 0);
    p= node_cofactor(node, fpos);
    q= node_cofactor(node, fneg);

    DC= node_xnor(p, q);
    DC= simp_obsdc_filter(DC, node_table, 0, ANDpi);
    node_free(fpos);
    node_free(fneg);
    node_free(p);
    node_free(q);
    return(DC);
}

/* Makes the don't care computed for an edge compatible with a previous
 * edge. If the obs. don't care for current edge is D and for the previous
 * edge DP, and the Boolean variable at the node that previous edge
 * originates from is x, then we find   Dx . Dx' + D . DP'.
 */
static node_t *
compute_cspf(DC, pfanin, node_table2, ANDpi, node_table3)
node_t *DC;
node_t *pfanin;
st_table *node_table2; /* used for filtering. */
node_t *ANDpi; /* ANDing of the PIs used for filtering. */
st_table *node_table3; /* used for collapsing. */
{
    node_t *p, *q, *n0, *n1, *f0, *f1, *ndc;
    int l,numin;
    node_t *dcfanin;
    char *dummy;



    if ((p= CSPF(pfanin)->node) == NULL){
        return(DC);
    }
        
    if (node_function(p) == NODE_0){
        return(DC);
    }

/*finds Dx . Dx', given x is the previous fanin*/
    f0= node_literal(pfanin, 0);
    f1= node_literal(pfanin, 1);
    n0= node_cofactor(DC, f0);
    n1= node_cofactor(DC, f1);
    q= node_and(n0, n1);


    node_free(n0);
    node_free(n1);
    node_free(f0);
    node_free(f1);

    n0= node_not(p);

/* filter DP before computing DP' . D. */
    if (node_num_fanin(n0) > 0){
        n0= simp_obsdc_filter(n0, node_table2, 3,ANDpi);
        numin =node_num_fanin(n0);
        for(l=0; l< numin; l++){
            dcfanin= node_get_fanin(n0, l);
            if (!st_lookup(node_table2,(char *)dcfanin,&dummy)){
            if (node_function(dcfanin) != NODE_PI){
            if (st_lookup(node_table3,(char *)dcfanin,&dummy)){
                (void)node_collapse(n0, dcfanin);
                l--;
                numin= node_num_fanin(n0);
            }
            }
            }
        }
        n0= simp_obsdc_filter(n0, node_table2, 0, ANDpi);
    }

    n1= node_and(DC, n0);
    node_free(n0);

    ndc= node_or(q, n1);
    node_free(q);
    node_free(n1);
    node_free(DC);
    return(ndc);
}

/* Returns the CSPF for a node if it exists.
 *
 */
node_t *
cspf_dc(f)
node_t *f;
{
    node_t *p;

    if ((p= CSPF(f)->node) != NULL){
        return(node_dup(p));
    }else{
        (void) fprintf(miserr, "cspf does not exist\n");
        return(node_constant(0));    
    }
}



/* A table of transitive fanouts of transitive fanins of a node.
 *
 */
static st_table *
cspf_inout_table(f, il, ol)
node_t *f;
int il, ol;
{
    st_table *finout_table;
    array_t *fanin_list, *fanout_list;
    node_t *fanin, *np;
    int i, j;
    
    finout_table= st_init_table(st_ptrcmp, st_ptrhash);
    fanin_list = network_tfi(f, il);
    for (i = 0; i < array_n(fanin_list); i++) {
    fanin = array_fetch(node_t *, fanin_list, i); 
    (void) st_insert(finout_table, (char *) fanin, NIL(char));
    fanout_list = network_tfo(fanin, ol);
    for (j = 0; j < array_n(fanout_list); j++) {
        np = array_fetch(node_t *, fanout_list, j);
        (void) st_insert(finout_table, (char *) np, NIL(char));
    }
    array_free(fanout_list);
    }
    array_free(fanin_list);
    return(finout_table);
}

