#include "mis.h"
#include "simp_int.h"

static void fobsdc_sm_bp_1();
static void node_d2merge();

/* removes the cubes in the observability don't care set that are unlikely
 * to be used for simplification.
 */
node_t *
simp_obsdc_filter(dc, node_table, var, nodepi)
node_t *dc; 
st_table *node_table; /* a list of variables allowed in each cube. */
int	var;          /* indicates how many variables that are not in the 
		       * table are allowed in each cube. */
node_t *nodepi;
{

    sm_matrix *Mfdc;
    node_t *new_dc;
    node_t *n1, *n2, *n3;
    st_generator *gen;
    char *dummy;
    node_t *np;

    /* if nothing to filter. */
    if (node_num_fanin(dc) == 0) {
        return(dc);
    }

    /* AND all the nodes in the table and all Primary Inputs. */
    n1= node_constant(1);
    st_foreach_item(node_table, gen, &dummy, NIL(char *)) {
	np = (node_t *) dummy;
    	n3= node_literal(np,1);
    	n2= node_and(n3, n1);
    	node_free(n1);
    	node_free(n3);
    	n1= n2;
    }
    n3= node_literal(nodepi,1);
    n2= node_and(n3, n1);
    node_free(n1);
    node_free(n3);
    n1= n2;


    node_d2merge(dc);
    if (node_num_fanin(dc) == 0) {
        return(dc);
    }


    /* create sparse matrix for the AND function and its dont care */
    Mfdc = simp_node_to_sm(n1, dc);

    fobsdc_sm_bp_1(Mfdc, var);

    new_dc = simp_sm_to_node(Mfdc);

    node_free(dc);
    node_free(n1);
    sm_free(Mfdc);
    return(new_dc);
}


/* remove any cube from observability don't care set that has more than
 * "var" literals that are not in any cubes of the onset function.
 */
static void
fobsdc_sm_bp_1(M, var)
sm_matrix *M;
int    var;
{
    sm_row *prow;
    sm_col *pcol;
    sm_element *p;
    int i, init_rows;

    if (M->nrows == 0)
        return;

    /* mark every row and column as unvisited */
    for (prow = M->first_row; prow != NULL; prow = prow->next_row)
        prow->flag = 0;
    for (pcol = M->first_col; pcol != NULL; pcol = pcol->next_col)
        pcol->flag = 0;
    
    /* for each row in the on set mark the columns in them as visited */
    for (prow = M->first_row; prow != NULL; prow = prow->next_row) {
        if (((int) prow->user_word) == F_SET) {
            prow->flag = 1;
            for (p = prow->first_col; p != NULL; p = p->next_col) {
                pcol = sm_get_col(M, p->col_num);
                pcol->flag = 1;
            }
        }
    }
    
    for (prow = M->first_row; prow != NULL; prow = prow->next_row) {
        if (!(((int) prow->user_word) == F_SET)) {
            for (p = prow->first_col; p != NULL; p = p->next_col) {
                pcol = sm_get_col(M, p->col_num);
    			if (!pcol->flag)
    				prow->flag++;
            }
        }
    }


    /* any unmarked rows can now be deleted => inessential dont cares */
    init_rows = M->rows_size;
    if (simp_debug)
        (void) fprintf(misout, "rows deleted : ");
    for (i = 0; i < init_rows; i ++) {
        prow = sm_get_row(M, i);
    	if (prow != NULL){
        if (!(((int) prow->user_word) == F_SET)) {
            if ( prow->flag > var) {
               	 sm_delrow(M, i);
            }
    	}
    	}
    }
}

node_t *
simp_obssatdc_filter(dc, f, var)
node_t *dc; /*obs. don't care to be filtered. */
node_t *f;
int    var;
{
    sm_matrix *Mfdc;
    node_t *new_dc;

    /* if nothing to filter. */
    if (node_num_fanin(dc) == 0) {
        return(dc);
    }


    /* create sparse matrix for the AND function and its dont care */
    Mfdc = simp_node_to_sm(f, dc);

    fobsdc_sm_bp_1(Mfdc, var);

    new_dc = simp_sm_to_node(Mfdc);

    node_free(dc);
    sm_free(Mfdc);
    return(new_dc);
}


/* repeated distance 1 merge 
 *
 */
static void
node_d2merge(f)
node_t *f;
{

    int nin, i, lit_count;
    int old_litcount;

    nin = node_num_fanin(f);
    define_cube_size(nin);

    old_litcount = node_num_literal(f);
    do {
    lit_count = node_num_literal(f);
    for (i=0; i< nin; i++){
    	f->F = d1merge(f->F, i);
    	}
    } while (lit_count > node_num_literal(f));

    if (old_litcount > node_num_literal(f))
    	node_scc(f);

    f->is_dup_free = 1;
    f->is_scc_minimal = 1;
    node_minimum_base(f);
}
