#include "mis.h"
#include "simp_int.h"

static void filter_exact();
static void filter_disj_support();
static void filter_size();
static void filter_fdist();
static void filter_sdist();

/* filter the dont care set according to the filter type indicated 
 */
node_t *
simp_dc_filter(f, dc, filter)
node_t *f,*dc;
sim_filter_t filter;
{

    sm_matrix *Mfdc;
    node_t *new_dc;
    
    /* constant functions ignored for filtering */
    if (node_num_fanin(dc) == 0) {
        return(dc);
    }

    (void) node_d1merge(dc);

    /* create sparse matrix for function and its dont care */
    Mfdc = simp_node_to_sm(f, dc);

    if (simp_debug) {
        simp_sm_print(Mfdc);
    }

    switch (filter) {
    case SIM_FILTER_EXACT :
        (void) filter_exact(Mfdc);
        break;
    case SIM_FILTER_DISJ_SUPPORT :
        (void) filter_exact(Mfdc);
        (void) filter_disj_support(Mfdc);
        break;
    case SIM_FILTER_SIZE :
        (void) filter_exact(Mfdc);
        (void) filter_size(Mfdc);
        break;
    case SIM_FILTER_FDIST :
	(void) filter_exact(Mfdc);
	(void) filter_fdist(Mfdc, DIST_BOUND);
        break;
    case SIM_FILTER_SDIST :
	(void) filter_exact(Mfdc);
	(void) filter_sdist(Mfdc, DIST_BOUND);
        break;
    case SIM_FILTER_ALL :
        (void) filter_exact(Mfdc);
        (void) filter_disj_support(Mfdc);
        (void) filter_size(Mfdc);
        break;
    case SIM_FILTER_NONE:
        break;
    default:
        fail("Wrong filter type");
        exit(-1);
    }

    if (simp_debug) {
        simp_sm_print(Mfdc);
    }

    new_dc = simp_sm_to_node(Mfdc);
    node_free(dc);
    sm_free(Mfdc);

    return(new_dc);
}

/* attempt to find a partition (iteratively) of the form :
 *
 *     xy-
 *     -yz
 *
 * where y is a single variable.
 */
static void
filter_exact(Mfdc)
sm_matrix *Mfdc;
{
    array_t *chosen;
    sm_col *col;

    chosen = sm_col_count_init(sm_ncols(Mfdc));
    while ((col = sm_get_long_col(Mfdc, chosen)) != NULL) {
        (void) fdc_sm_bp_1(Mfdc, col);
    }
}

static void
filter_disj_support(Mfdc)
sm_matrix *Mfdc;
{
    (void) fdc_sm_bp_2(Mfdc);
}

static void
filter_size(Mfdc)
sm_matrix *Mfdc;
{
    (void) fdc_sm_bp_4(Mfdc);
}

static void
filter_fdist(Mfdc, distance)
sm_matrix *Mfdc;
int distance;
{
    (void) fdc_sm_bp_6(Mfdc, distance);
}

static void
filter_sdist(Mfdc, distance)
sm_matrix *Mfdc;
int distance;
{
    (void) fdc_sm_bp_7(Mfdc, distance);
}
