#include "mis.h"
#include "simp_int.h"

static void	absorb_fanin();
static array_t  *build_fanin_array();
static node_t   *select_node();
static bool     need_simplify();

void
simplify_all(network, method, dctype, accept, filter)
network_t *network;
sim_method_t method;
sim_dctype_t dctype;
sim_accept_t accept;
sim_filter_t filter;
{
    st_table *table;
    node_t *np;

    table = st_init_table(st_ptrcmp, st_ptrhash);
    while ((np = select_node(network, table)) != NIL(node_t)) { 
        absorb_fanin(network, np);
        simplify_node(np, method, dctype, accept, filter);
        decomp_quick_node(network, np);
    }
    st_free_table(table);

    eliminate(network, -1, 1000);
}

static node_t *
select_node(network, table)
network_t *network;
st_table *table;
{
   array_t *nodevec;
   node_t *np;
   int i;

   /*  need sweep? */
   nodevec = network_dfs_from_input(network);
   for (i = 0; i < array_n(nodevec); i++) {
       np = array_fetch(node_t *, nodevec, i);
       if (np->type == INTERNAL && node_num_fanout(np) == 0) {
	   (void) network_cleanup(network);
	   array_free(nodevec);
	   nodevec = network_dfs_from_input(network);
	   break;
       }
   }

   for (i = 0; i < array_n(nodevec); i++) {
       np = array_fetch(node_t *, nodevec, i);
       if (need_simplify(np)) {
	   if (st_insert(table, (char *) np, NIL(char)) == 0) {
	       array_free(nodevec);
	       return np;
	   }
       }
   }
   array_free(nodevec);

   return NIL(node_t);
}

static bool
need_simplify(np)
node_t *np;
{
    switch (node_function(np)) {
    case NODE_PI:
    case NODE_PO:
	return FALSE;
    default: 
	if (node_num_fanout(np) > 1) {
	    return TRUE;
	} else if (node_function(node_get_fanout(np, 0)) == NODE_PO) {
	    return TRUE;
	} else {
	    return FALSE;
	}
    }
    /* NOTREACHED */
}

static void
absorb_fanin(network, f)
network_t *network;
node_t *f;
{
    array_t *fanins;
    node_t *np;
    int i;

    fanins = build_fanin_array(f);
    for (i = 0; i < array_n(fanins); i++) {
	np = array_fetch(node_t *, fanins, i);
	if (!node_collapse(f, np)) {
	    fail("absorb fanin: can't collapse np");
	    exit(-1);
	}
	network_delete_node(network, np);
    }
    array_free(fanins);
}

static array_t *
build_fanin_array(f)
node_t *f;
{
    node_t *fanin;
    array_t *tfanin, *temp;
    int i;

    tfanin = array_alloc(node_t *, 0);
    foreach_fanin(f, i, fanin) {
	if (fanin->type == INTERNAL && node_num_fanout(fanin) == 1) {
	    array_insert_last(node_t *, tfanin, fanin);
	    temp = build_fanin_array(fanin);
	    array_append(tfanin, temp);
	    array_free(temp);
	}
    }
    return tfanin;
}
