#include "mis.h"
#include "simp_int.h"

static void simp_accept();
static node_t *simp_cspf_dc();

void
simplify_node(f, method, dctype, accept, filter)
node_t *f;
sim_method_t method;
sim_dctype_t dctype;
sim_accept_t accept;
sim_filter_t filter;
{
    node_t *DC, *newf;
    node_sim_type_t node_sim_type;
    char buffer[80];

    if (simp_trace) {
	(void) fprintf(misout, "%-6s ", node_name(f));
	(void) fprintf(misout, "(%3d): ", factor_num_literal(f));
	(void) fflush(misout);
    }

    /* generate the don't care set */ 
    switch (dctype) {
    case SIM_DCTYPE_NONE:
	DC = node_constant(0); 
	break;
    case SIM_DCTYPE_FANIN:
	DC = simp_tfanin_dc(f, simp_fanin_level, simp_fanin_fanout_level);
	break;
    case SIM_DCTYPE_FANOUT:
	DC = simp_fanout_dc(f);
	break;
    case SIM_DCTYPE_INOUT:
	DC = simp_inout_dc(f, simp_fanin_level, simp_fanin_fanout_level);
	break;
    case SIM_DCTYPE_SUB_FANIN:
	DC = simp_sub_fanin_dc(f);
	break;
    case SIM_DCTYPE_ALL:
	DC = simp_all_dc(f);
	break;
    case SIM_DCTYPE_UNKNOWN:
    default:
	fail("Error: unknown dctype");
	/* NOTREACHED */
    }

    if (simp_trace) {
	(void) sprintf(buffer, "DCb=(%d,%d,%d)", 
		node_num_fanin(DC), node_num_cube(DC), node_num_literal(DC));
	(void) fprintf(misout, "%-24s", buffer);
	(void) fflush(misout);
    }
    if (simp_debug) {
	(void) fprintf(misout, "DC(%s) = ", node_name(f));
        node_print_rhs(misout, DC);
	(void) fprintf(misout, "\n");
	(void) fflush(misout);
    }

    /* trim the don't-care set */
    DC = simp_dc_filter(f, DC, filter);

    if (simp_trace) {
	(void) sprintf(buffer, "DCa=(%d,%d,%d)", 
		node_num_fanin(DC), node_num_cube(DC), node_num_literal(DC));
	(void) fprintf(misout, "%-24s", buffer);
	(void) fflush(misout);
    }
    if (simp_debug) {
	(void) fprintf(misout, "simp_dc_filtered(%s) = ", node_name(f));
        node_print_rhs(misout, DC);
	(void) fprintf(misout, "\n");
	(void) fflush(misout);
    }


    /* minimize f */
    switch (method) {
    case SIM_METHOD_SIMPCOMP:
	node_sim_type = NODE_SIM_SIMPCOMP;
	break;
    case SIM_METHOD_ESPRESSO:
	node_sim_type = NODE_SIM_ESPRESSO;
	break;
    case SIM_METHOD_EXACT:
	node_sim_type = NODE_SIM_EXACT;
	break;
    case SIM_METHOD_EXACT_LITS:
	node_sim_type = NODE_SIM_EXACT_LITS;
	break;
    case SIM_METHOD_DCSIMP:
	node_sim_type = NODE_SIM_DCSIMP;
	break;
    case SIM_METHOD_NOCOMP:
	node_sim_type = NODE_SIM_NOCOMP;
	break;
    case SIM_METHOD_SNOCOMP:
	node_sim_type = NODE_SIM_SNOCOMP;
	break;
    case SIM_METHOD_UNKNOWN:
    default:
	fail("Error: unknown simplication method");
	/* NOTREACHED */
    }

    newf = node_simplify(f, DC, node_sim_type);
    node_free(DC);

    /* accept ? */
    simp_accept(f, newf, accept);

    /* save update the sim_flag */ 
    SIM_FLAG(f)->method = method;
    SIM_FLAG(f)->accept = accept;
    SIM_FLAG(f)->dctype = dctype;
}

static void
simp_accept(f, newf, accept)
node_t *f, *newf;
sim_accept_t accept;
{
    bool do_accept;
    int old, new;

    do_accept = FALSE;

    switch (accept) {
    case SIM_ACCEPT_FCT_LITS:
	old = factor_num_literal(f);
	new = factor_num_literal(newf);
	break;
    case SIM_ACCEPT_SOP_LITS:
	old = node_num_literal(f);
	new = node_num_literal(newf);
	break;
    case SIM_ACCEPT_CUBES:
	old = node_num_cube(f);
	new = node_num_cube(newf);
	break;
    case SIM_ACCEPT_ALWAYS:
	old = factor_num_literal(f);
	new = factor_num_literal(newf);
	do_accept = TRUE;
	break;
    case SIM_ACCEPT_UNKNOWN:
    default:
	fail("Error: unknown acceptance criteria.");
	exit(-1);
	/* NOTREACHED */
    }

    if (do_accept || new < old) {
	if (simp_trace) {
	    (void) fprintf(misout, "simplified ", node_name(f));
	    (void) fprintf(misout, "(%d)\n", new);
	    (void) fflush(misout);
	}
	(void) node_replace(f, newf);
    } else {
	if (simp_trace) {
	    (void) fprintf(misout, "not simplified\n", node_name(f));
	    (void) fflush(misout);
	}
	node_free(newf);
    }
}

void
simplify_cspf_node(f, method, dctype, accept, filter)
node_t *f;
sim_method_t method;
sim_dctype_t dctype;
sim_accept_t accept;
sim_filter_t filter;
{
    node_t *DC, *newf;
    node_sim_type_t node_sim_type;
    char buffer[80];

    if (simp_trace) {
	(void) fprintf(misout, "%-6s ", node_name(f));
	(void) fprintf(misout, "(%3d): ", factor_num_literal(f));
	(void) fflush(misout);
    }

    DC = simp_cspf_dc(f, simp_fanin_level, simp_fanin_fanout_level);

    if (simp_trace) {
	(void) sprintf(buffer, "DCb=(%d,%d,%d)", 
		node_num_fanin(DC), node_num_cube(DC), node_num_literal(DC));
	(void) fprintf(misout, "%-24s", buffer);
	(void) fflush(misout);
    }

    if (filter != SIM_FILTER_NONE){
    DC = simp_obssatdc_filter(DC, f, 2);
    if ((node_num_cube(DC) > 100) && (node_num_fanin(DC)> 20))
	    DC = simp_obssatdc_filter(DC, f, 1);
    if ((node_num_cube(DC) > 100) && (node_num_fanin(DC)> 20))
	    DC = simp_obssatdc_filter(DC, f, 0);
    }

    if (simp_trace) {
	(void) sprintf(buffer, "DCa=(%d,%d,%d)", 
		node_num_fanin(DC), node_num_cube(DC), node_num_literal(DC));
	(void) fprintf(misout, "%-24s", buffer);
	(void) fflush(misout);
    }

    /* minimize f */
    switch (method) {
    case SIM_METHOD_SIMPCOMP:
	node_sim_type = NODE_SIM_SIMPCOMP;
	break;
    case SIM_METHOD_ESPRESSO:
	node_sim_type = NODE_SIM_ESPRESSO;
	break;
    case SIM_METHOD_EXACT:
	node_sim_type = NODE_SIM_EXACT;
	break;
    case SIM_METHOD_EXACT_LITS:
	node_sim_type = NODE_SIM_EXACT_LITS;
	break;
    case SIM_METHOD_DCSIMP:
	node_sim_type = NODE_SIM_DCSIMP;
	break;
    case SIM_METHOD_NOCOMP:
	node_sim_type = NODE_SIM_NOCOMP;
	break;
    case SIM_METHOD_SNOCOMP:
	node_sim_type = NODE_SIM_SNOCOMP;
	break;
    case SIM_METHOD_UNKNOWN:
    default:
	fail("Error: unknown simplication method");
	/* NOTREACHED */
    }
	
    newf = node_simplify(f, DC, node_sim_type);

    node_free(DC);

    /* accept ? */
    simp_accept(f, newf, accept);

    /* save update the sim_flag */ 
    SIM_FLAG(f)->method = method;
    SIM_FLAG(f)->accept = accept;
    SIM_FLAG(f)->dctype = dctype;
}

static node_t *
simp_cspf_dc(f, il, iol)
node_t *f;
int il, iol;
{
    node_t *DC, *faninDC, *fanoutDC;
    node_t *tmp;

    faninDC = simp_tfanin_dc(f, il, iol); 
    tmp	= simp_sub_fanin_dc(f);
    DC= node_or(faninDC, tmp);
    node_free(faninDC);
    node_free(tmp);
    faninDC= DC;
    fanoutDC = cspf_dc(f);
    DC = node_or(faninDC, fanoutDC);
    node_free(faninDC);
    node_free(fanoutDC);
    return DC;
}
