#include "mis.h"
#include "speed_int.h"

array_t *
speed_decomp( innode, coeff, model, delay_flag)
node_t *innode;
double coeff;
delay_model_t model;
int delay_flag; 	/* 1=> use user field */
{
    node_t *node;
    lsGen gen;
    array_t *array;
    network_t *network;

    network = speed_network_create_from_node(innode, delay_flag);

    /* Do a delay trace on the network and call the
    decomposition routine on the network */

    if ( !speed_delay_trace(network, model) ){
        fail(error_string());
    }

    foreach_node(network, gen, node){
	if ( node->type == INTERNAL ){
	    (void) lsFinish(gen);
	    break;
	}
    }

    /* If the MAPPED model is used get quick delays from global data_str */
    speed_set_library_accl(1);
    if ( !speed_decomp_network( network, node, coeff, model)){
	error_append("Failed trying to speed_decomp network");
	fail(error_string());
    }
    /* reset the library accelerator so as to get realistic delays */
    speed_set_library_accl(0);

    /*
     * Cleanup all the inverters and buffers if required 
     */

    if ( sp_add_inv) {
	add_inv_network(network);
	(void) speed_delay_trace( network, model);
    }


    /* Convert the network into an array */
    if ( speed_debug) {
	(void) fprintf(misout,"After decomposition ----- \n");
	(void) com_execute( &network, "p");
    }
    array = network_and_node_to_array(network, innode, NIL(st_table));

    network_free(network);
    return array;
}


void
speed_adjust_phase(network)
network_t *network;
{
    node_t *node, *fo;
    lsGen gen1;
    int i;
    array_t *nodevec;

    /* Collapse all the inverters and buffers into their
    fanouts -- except if fanout is a primary output  */

    nodevec = network_dfs_from_input(network);
    for ( i = 0; i < array_n(nodevec); i++){
        node = array_fetch(node_t *, nodevec, i);
        if ( node->type == INTERNAL){
            if ( (node_function(node) == NODE_BUF) ||
                    (node_function(node) == NODE_INV)) {
                foreach_fanout(node, gen1, fo){
                    if ( node_function(fo) != NODE_PO){
                        (void) node_collapse( fo, node);
                    }
                }

                if ( node_num_fanout( node) == 0){
		    /*
                    if ( speed_debug){
                        (void) fprintf(misout,"Deleting node %s\n",
                                node_name(node));
                    }
		    */
                    network_delete_node(network, node);
                }
            }
        }
    }
    array_free(nodevec);
}
