#include "mis.h"
#include "speed_int.h"
#include <math.h>

typedef enum plot_action_enum plot_action_t;
enum plot_action_enum {
    PLOT_PLOT, PLOT_KILL
};
void speed_plot_use();
void speed_plot_crit_network();

int
com__speed_plot(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    plot_action_t action;
    delay_model_t model;
    int c, hilite_flag, print_gate_name, dist;
    int plot_all_flag;
    double thresh, coeff;
    char name[40];

    weight_debug = FALSE;
    speed_debug = FALSE;

    model = DELAY_MODEL_UNIT_FANOUT;
    action = PLOT_PLOT;
    print_gate_name = FALSE;
    coeff = DEFAULT_SPEED_COEFF;
    dist = DEFAULT_SPEED_DIST;
    thresh = DEFAULT_SPEED_THRESH;
    hilite_flag = FALSE;
    plot_all_flag = FALSE;
    (void) strcpy(name, network_name(*network));
    getopt_reset();
    while ((c = getopt(argc, argv, "t:n:agkhm:d:w:")) != EOF) {
	switch (c) { 
	case 'g':
	    print_gate_name = TRUE;
	    break;
	case 'a':
	    plot_all_flag = 1;
	    break;
	case 'h':
	    hilite_flag = 1;
	    break;
	case 'w':
	    coeff = atof(optarg);
	    if (coeff < 0) coeff = 0;
	    if (coeff > 1) coeff = 1;
	    break;
	case 'd':
	    dist = atoi(optarg);
	    break;
	case 't':
	    thresh = atof(optarg);
	    break;
	case 'n':
	    (void) strcpy(name, optarg);
	    break;
	case 'k':
	    action = PLOT_KILL;
	    break;
	default:
	    speed_plot_use();
	    return 1;
	}
    }

    if (argc - optind > 0) {
	speed_plot_use();
    	return 1;
    }

    speed_plot_crit_network(*network, name, action, thresh, hilite_flag,
	plot_all_flag, print_gate_name, model, dist, coeff);
    return 0;
}


void
speed_plot_crit_network(network, name, action, thresh, flag, all_flag, print_gate_name, model, dist, coeff)
network_t *network;
char *name;
plot_action_t action;
delay_model_t model;
int dist, flag, all_flag; /* flag == 0 -> only crit nodes */
int print_gate_name;	  /* Print name of the gate in place of node names */
double coeff, thresh;
{
    plot_t *ploter;
    array_t *mincut;
    char *dummy, pname[40], *sname;
    int num, i;
    double time;
    lsGen gen;
    node_t *np;
    lib_gate_t *gate;
    st_table *weight_table, *name_table, *print_table, *shape_table;

    if (action == PLOT_KILL) {
    if (st_delete(speed_plot_table, &name, &dummy)) {
        ploter = (plot_t *) dummy; 
        FREE(name);
        graphics_kill(ploter);
    } else {
        (void) fprintf(misout, "%s: No window with this name!\n", name);
    }
    return;
    }

    /* Build the name to be printed */
    num = 0;
    foreach_node(network, gen, np) {
	if (np->type == INTERNAL) {
	    num += factor_num_literal(np);
	}
    }
    assert(delay_trace(network, model));
    set_speed_thresh(network, thresh);
    np = delay_latest_output(network, &time);
    (void) sprintf(pname, "%s: %d nodes, %d lit, delay = %-5.2f", name, network_num_internal(network), num, time);

    /* Build up the critical network */
    print_table = st_init_table(st_ptrcmp, st_ptrhash);
    shape_table = st_init_table(st_ptrcmp, st_ptrhash);
    name_table = st_init_table(st_ptrcmp, st_ptrhash);

    if (print_gate_name){
	foreach_node(network, gen, np){
	    if (np->type == INTERNAL){
		if ((gate = lib_gate_of(np)) != NIL(lib_gate_t)){
		    (void)st_insert(name_table, (char *)np, lib_gate_name(gate));
		}
	    }
	}
    }

    if (!all_flag) {
	(void) fprintf(misout, "Distance = %d ; Threshold = %5.2f\n",dist, thresh);
        /* Get the weights for the cutset analysis */
	weight_table = speed_compute_weight(network, model, coeff, 
		dist, NIL(array_t));

        /* compute the minimum weighted cutset of the critical paths */
	(void) com_execute(&network, "_maxflow");
        mincut =  cutset(network, weight_table);
	(void) com_execute(&network, "_maxflow");

	/* Mark the critical path in yellow */
        foreach_node(network, gen, np){
            if (speed_critical(np)){
                (void) st_insert(print_table, (char *)np, util_strsav("yellow"));
            }
        }
        for (i = 0; i < array_n(mincut); i++){
            np = array_fetch(node_t *, mincut, i);
            (void) st_insert(print_table, (char *)np, util_strsav("coral"));
	    (void) st_insert(shape_table, (char *)np, (char *)6);
        }
        array_free(mincut);
	st_free_table(weight_table);
    }

    if (st_lookup(speed_plot_table, name, &dummy)) {
    	ploter = (plot_t *) dummy; 
    	sname = name;
    } else {
    	ploter = NIL(plot_t); 
    	sname = util_strsav(name);
    }

    if (flag){
        ploter = graphics_plot(ploter, network, pname, print_table,
	NIL(st_table), shape_table, name_table, NIL(st_table),
	NIL(st_table), NIL(st_table), NIL(st_table));
    }
    else if (all_flag) {
        ploter = graphics_plot(ploter, network, pname,  NIL(st_table),
	NIL(st_table), NIL(st_table), name_table, NIL(st_table),
	NIL(st_table), NIL(st_table), NIL(st_table));
    } else {
        ploter = graphics_plot(ploter, network, pname,  NIL(st_table),
	NIL(st_table), shape_table, name_table, NIL(st_table),
	NIL(st_table), NIL(st_table), print_table);
    }

    st_free_table(print_table);
    st_free_table(shape_table);
    st_free_table(name_table);
    if (ploter){
	(void) st_insert(speed_plot_table, sname, (char *) ploter);
    }
}


void
speed_plot_use()
{
    (void) fprintf(misout, "usage: _speed_plot [-n name] [-t thresh] [-a] [-g] [-h] [-k] [-t n.n] [-w n.n] [-d n]\n");
    (void) fprintf(misout, "    -n\tname\tDisplay name.\n");
    (void) fprintf(misout, "    -k\t\tKill the display\n");
    (void) fprintf(misout, "    -h\t\tHighlight critical nodes. \n");
    (void) fprintf(misout, "    -g\t\tUse gate names instead of node names\n");
    (void) fprintf(misout, "    -a\t\tPrint network without considering delay. \n");
    (void) fprintf(misout, "    -t\tn.n\tCritical threshold (ignored for -k & -a).\n");
    (void) fprintf(misout, "    -w\tn.n\tRelative weight of area (ignored for -k & -a).\n");
    (void) fprintf(misout, "    -d\tn\tDistance for collapsing (ignored for -k & -a).\n");
}
