#include "mis.h"
#include "speed_int.h"
#include <math.h>

static void weight_compute();
static int speed_weight_bfs();
static void speed_mark_input_cone();
static void linfit();
static void speed_sort2();
static void speed_rank();
enum st_retval speed_weight_count_lit();
/*
 * If the return value is 1 only then the components should
 * be combined. If the return is 0 then a large weight needs
 * put on the node
 */

int
speed_weight(node, dist, model, t_cost, a_cost)
node_t *node;
int dist;
delay_model_t model;
double *t_cost;
int *a_cost;
{
    double max_a, min_a, min_d, mean_a , mean_d; /* For scaling */
    double m, o_time, a_time, d_time; /* For correlation */
    double s_dev, tf, crit_frac; 	
    double rad, sum_a, sum_d, co, angle, p_angle, slope;
    int i, n, crit;
    delay_time_t t;
    st_table *table;
    array_t *a_array, *d_array;
    /* Some day do special casing of a Boolean nature */
    /*
    int to_cofactor;
    */

    a_array = array_alloc(double,0);
    d_array = array_alloc(double,0);
    min_a = min_d = POS_LARGE;
    max_a = NEG_LARGE;
    sum_a = sum_d = rad = 0.0;

    table = st_init_table(st_ptrcmp, st_ptrhash);

    t = delay_arrival_time(node);
    o_time = D_MAX(t.rise, t.fall);

    *a_cost = 0;
    crit = 0;
    n = 0;

    /*
    to_cofactor = speed_weight_bfs(node, dist, table, model, a_array, d_array,
    	    o_time, &crit, &n, &min_a, &max_a, &min_d, &sum_a, &sum_d, a_cost);
    */
    (void) speed_weight_bfs(node, dist, table, model, a_array, d_array,
    	    o_time, &crit, &n, &min_a, &max_a, &min_d, &sum_a, &sum_d, a_cost);

    /*
     * Compute the t_frac as based on the Distribution
     * of the delays and a_times in a 2-D plane 
     */

    if (n > 2 ) {
        m = (double) n;
	/*
        mean_a = (sum_a / m ) - min_a;
        mean_d = (sum_d / m ) - min_d;
	*/
        mean_a = (sum_a / m );
        mean_d = (sum_d / m );

        for (i = 0; i < n ; i++){
                a_time = array_fetch(double, a_array, i) - mean_a ;
                d_time = array_fetch(double, d_array, i) - mean_d ;

                rad += (pow(d_time, 2.0) + pow(a_time, 2.0));
        }
        /*
	 * Get the standard deviation (preferred over the absolute
    	 * deviation) since we want a large separation and the
    	 * angle from the linear regression.
	 */
        s_dev = sqrt(rad) / m;

        if (min_a == max_a ){
            angle = SP_PI_2;
        }
        else {
            (void) linfit(a_array, d_array, min_a, min_d, &slope);
            angle = atan (slope);
        }

        if (angle > SP_PI_4){
            co = 4 - (4 * angle * SP_1_PI);
        }
        else{
            co = 2 + (4 * angle * SP_1_PI);
        }
	co = MAX(co, 0.3);

        tf = co  *  pow((s_dev / (double)dist), 2.0);
	/*
        tf = pow(co, 2.0) * (s_dev / (double)dist);
	*/
	/*
        tf = co  *  (s_dev / (double)dist);
	*/
    }
    else {
        tf = -1.0;
    	s_dev = 0.0; /* Default */
    	angle = - SP_PI;
    }

    if (weight_debug) {
	p_angle = angle * 180 * SP_1_PI;
        (void) fprintf(misout,"%-10s T_comp =(%4.0f:%4.2f) Area = %d   %3d d_fanin inputs \n",
        node_name(node),p_angle , s_dev, *a_cost, array_n(a_array));
    }

    array_free(d_array);
    array_free(a_array);
    st_free_table(table);

    crit_frac = (n > 0 ? ((double)(crit) / (double)(n)) : 1);

    /*
     * tf < 0 => this node should never be considered 
     */
    if (tf <= 0 ) {
	*t_cost = POS_LARGE;
    	return 0;
    } else {
    	*t_cost = (MAXWEIGHT * crit_frac) / tf; 
	return 1;
    }
}

static int
speed_weight_bfs(node, dist, table, model, a_array, d_array, o_time,
crit, n, min_a, max_a, min_d, sum_a, sum_d, area)
node_t *node;
int dist, *crit, *n;
delay_model_t model;
st_table *table;
array_t *a_array, *d_array;
double o_time;
double *sum_a, *sum_d, *min_a, *min_d, *max_a;
int *area;
{
    int i, j, k, first, last;
    int to_cofactor = FALSE;
    int more_to_come, fanout_flag;
    array_t *temp_array;
    st_table *area_table;
    lsGen gen;
    node_t *temp, *fi, *fo, *no;

    temp_array = array_alloc(node_t *, 0);
    area_table = st_init_table(st_ptrcmp, st_ptrhash);
    array_insert_last(node_t *, temp_array, node); 
    (void) st_insert(table, (char *)node, (char *)(dist+1));
    first = 0;
    more_to_come = TRUE;
    for (i = dist; (i > 0) && more_to_come; i--){
	more_to_come = FALSE;
	last = array_n(temp_array);
	for (j = first; j < last; j++){
	    no = array_fetch(node_t *, temp_array, j);
	    foreach_fanin(no, k, fi) {
		if (node_function(fi) == NODE_PI){
		    if (! st_insert(table, (char *)fi, (char *)0)){
			if (speed_critical(fi) ) (*crit)++;
			(*n)++;
			weight_compute(fi, model, a_array, d_array, o_time,
			min_a, max_a, min_d, sum_a, sum_d);
		    }
		} else if (!st_is_member(table, (char *)fi)){
		/* Not yet visited */
		    if (speed_critical(fi) ){
		    /*
		     * Add to elements that will be processed 
		     */
			(void) st_insert(table, (char *)fi, (char *)i);
			array_insert_last(node_t *, temp_array, fi);
			more_to_come = TRUE;
		    } else {
		    /*
		     * Will be a fanin to collapsed node 
		     */
			(void) st_insert(table, (char *)fi, (char *)0);
			(*n)++;
			weight_compute(fi, model, a_array, d_array,
			    o_time, min_a, max_a, min_d, sum_a, sum_d);
		    }
		}
	    }
	}
	first = last;
    }

    /*
     * For the nodes in the array  see if they
     * fanout elsewhere. If so then mark and count the
     * the nodes in the fanin cone
     */
    for (i = 1; i < array_n(temp_array); i++){
	temp = array_fetch(node_t *, temp_array, i);
	if (!st_is_member(area_table, (char *)temp)){
	    if (node_num_fanout(temp) > 1) {
		fanout_flag = FALSE;
		foreach_fanout(temp, gen, fo){
		    if (!st_is_member(table, (char *)fo) ){
			fanout_flag = TRUE;
			(void) lsFinish(gen);
			break;
		    }
		}
		if (fanout_flag){
		    speed_mark_input_cone(temp, area_table, table);
		}
	    }
	}
    }
    *area = 0;
    st_foreach(area_table, speed_weight_count_lit, (char *)area);

    /*
     * Now the elements added in the last loop are 
     * the inputs to the collapsed node.
     */

    for (i = first; i < array_n(temp_array); i++){
	no = array_fetch(node_t *, temp_array, i);
	foreach_fanin(no, k, fi){
	    if (!st_insert(table, (char *)fi, (char *)0)){
		if (speed_critical(fi)) (*crit)++;
		(*n)++;
		weight_compute(fi, model, a_array, d_array,
			o_time, min_a, max_a, min_d, sum_a, sum_d);
	    }
	}
    }

    array_free(temp_array);
    st_free_table(area_table);
    return to_cofactor;
}


static void
speed_mark_input_cone(node, area_table, table)
node_t *node;
st_table *area_table, *table;
{
     array_t *array;
     int i, j, first = 0, last;
     int more_to_come = TRUE;
     node_t *temp, *fi;

     array = array_alloc(node_t *, 0);
     array_insert_last(node_t *, array, node);
     (void) st_insert(area_table, (char *)node, (char *)1);
     while (more_to_come){
	 more_to_come = FALSE;
	 last = array_n(array);
	 for (i = first; i < last; i++){
	    temp = array_fetch(node_t *, array, i);
	    foreach_fanin(temp, j, fi){
		if (st_is_member(table, (char *)fi)){
		    more_to_come = TRUE;
		    array_insert_last(node_t *, array, fi);
		    (void) st_insert(area_table, (char *)fi, (char *)1);
		}
	    }
	    first = last;
	 }
     }
     array_free(array);
}

static void
linfit(xin, yin, xmin, ymin, slope)
array_t *yin, *xin;
double *slope, xmin, ymin;
{

    double a=0, d=0, aa=0, ad=0, a_dev;
    double *x, *y, m;
    int num, j;
    
    num = array_n(xin);

    if (num != array_n(yin)){
        (void) fprintf(miserr, " Arrays not of same size \n");
        exit (-1);
    }
    x = ALLOC(double, num+1);
    y = ALLOC(double, num+1);

    for (j = 1; j <= num; j++){
        x[j] = (array_fetch(double, xin, j-1) - xmin);
        y[j] = (array_fetch(double, yin, j-1) - ymin);
    }

    /* Work with ranks rather than the delay values */
    speed_sort2(num, y, x);
    speed_rank(num, y);
    speed_sort2(num, x, y);
    speed_rank(num, x);

    /* Compute the least square fit */
    for (j = 1; j <= num; j++){
    	a += x[j];
    	d += y[j];
    	aa += x[j] * x[j];
    	ad += x[j] * y[j];
    }
    m = (double)num;

    a_dev = (aa / m) - ((a * a)/(m * m ));
    *slope = ((ad/m) - ((a * d)/(m * m)))/(sqrt(a_dev));

    FREE(x);
    FREE(y);
}


static void
speed_sort2(n, ra, rb)
int n;
double ra[], rb[];
{
    int i, j, ir, l;
    double rra, rrb;

    l = (n >> 1)+1;
    ir = n;
    for(;;){
        if (l > 1){
            rra = ra[--l];
            rrb = rb[l];
        } else{
            rra = ra[ir];
            rrb = rb[ir];
            ra[ir] = ra[1];
            rb[ir] = rb[1];
            if (--ir == 1){
                ra[1] = rra;
                rb[1] = rrb;
                return;
            }
        }
        i = l;
        j = l << 1;
        while(j <= ir){
            if (j < ir && ra[j] < ra[j+1]) ++j;
            if (rra < ra[j]){
                ra[i] = ra[j];
                rb[i] = rb[j];
                j += (i = j);
            }
            else j = ir + 1;
        }
        ra[i] = rra;
        rb[i] = rrb;
    }
}

static void
speed_rank(n, w)
double w[];
int n;
{
    int j=1, ji, jt;
    double rank;

    while(j < n){
        if (w[j+1] != w[j]){
            w[j] = (double)j;
            ++j;
        } else{
            for(jt = j+1; jt < n; jt++){
                if(w[jt] != w[j]) break;
            }
            rank = 0.5 * (j + jt - 1);
            for(ji = j; ji <=(jt-1); ji++){
                w[ji] = rank;
            }
            j = jt;
        }
    }
    if (j == n) w[n] = n;
}


/*
 * For the node, gets the slacks and arrival times.
 * Updates the min, max and sums and inserts the
 * delay and arrival times in the corresponding arrays
 */
static void
weight_compute(node, model, a_array, d_array, o_time, min_a, 
	max_a, min_d, sum_a, sum_d)
node_t *node;
delay_model_t model;
array_t *a_array, *d_array;
double o_time;
double *min_a, *max_a, *min_d, *sum_a, *sum_d;
{
    delay_time_t at, st;
    double d_time, a_time, s_time;

    st = delay_slack_time(node);
    at = delay_arrival_time(node);
    if ((model != DELAY_MODEL_UNIT_FANOUT) &&
	    (model != DELAY_MODEL_UNIT)) {
	a_time = D_MIN(at.rise, at.fall);
	s_time = D_MIN(st.rise, st.fall);
    } else {
	a_time = at.rise;
	s_time = st.rise;
    }

    d_time = o_time - s_time - a_time;
    *min_a = D_MIN(*min_a, a_time);
    *max_a = D_MAX(*max_a, a_time);
    *min_d = D_MIN(*min_d, d_time);

    (void) array_insert_last(double, a_array, a_time);
    (void) array_insert_last(double, d_array, d_time);

    *sum_a += a_time;
    *sum_d += d_time;
}

/* ARGSUSED */
enum st_retval
speed_weight_count_lit(key, value, arg)
char *key, *value, *arg;
{
    node_t *node;
    int *in;

    node = (node_t *)key;
    in = (int *)arg;
    if (node->type == INTERNAL) {
	*in += node_num_literal(node);
    }

    return ST_CONTINUE;
}
