#include "copyright.h"
#include "mustang.h"


/*
 *    ***********************************
 *    *** SEQUENTIAL STATE ASSIGNMENT ***
 *    ***********************************
 *
 *  Srinivas Devadas
*/

#define DISPLACEMENT 1
#define INTERCHANGE  2

double STATES_BEGIN, STATES_END;
double ALPHA, START_TEMP;
int init_cost;
NODE shift_node1, shift_node2;
int new_encode;

simulated_anneal(anneal_type)
int anneal_type;
{
/*
* Given a graph with weighted edges, find the encoding of the nodes
* such that the multiplication of the weights and the distance
* betweent the nodes is minimum.
*/

 int c1, c2, c3, start_cost;
 double temp, update_temp(), find_starting_temp();
 int cur_gen, max_gen;
 int try, yes, type;


   assign_initial_encoding();
   START_TEMP = find_starting_temp(anneal_type);
   ALPHA = 0.90;

   if (debug)
     (void)printf("# Initial cost is %d\n",init_cost);

   STATES_BEGIN = STATES_END = 1.0;
   if (specify is TRUE) {
     (void)printf("# Number of states at the beginning:");
     (void)fscanf(stdin,"%lf",&STATES_BEGIN);
     (void)printf("# Number of states at the end:");
     (void)fscanf(stdin,"%lf",&STATES_END);
     (void)printf("# Temperature is %lf, want to change it (1/0)?:",START_TEMP);
     (void)fscanf(stdin,"%d",&yes);
     if (yes is TRUE) {
       (void)printf("# New temperature:");
       (void)fscanf(stdin,"%lf",&START_TEMP);
     }
   }
   c1 = init_cost - 30;
   c2 = init_cost - 20;
   c3 = init_cost - 10;
   max_gen = STATES_BEGIN * node_num;
   temp = START_TEMP;
   while ( c1 isnt init_cost || c2 isnt init_cost || c3 isnt init_cost) {
     cur_gen  = 0; start_cost = init_cost;
     while (cur_gen < max_gen) {
       cur_gen++;
       try = generate_state(&type);
       if (try)
         accept_or_reject_state(temp, type, anneal_type);
     }
     temp = update_temp(temp, &max_gen);
     if (debug) {
       (void)printf("# Changing temperature to %lf, cost is %d\n",
                                                          temp,init_cost);
     }
     c1 = c2; c2 = c3; c3 = start_cost;
  }
  if (debug)
    (void)printf("# Final cost is %d\n",init_cost);
      
  return;

} /* end of simulated_anneal */


assign_initial_encoding()
{
/*
* Assign nodes initial encodings prior to annealing
*/
 int i;

  for(i=0;i<encode_num;i++) {
    poss_encodings[i]->node = NIL(struct node);
    poss_encodings[i]->assigned = FALSE;
  }

  for(i=1;i<=node_num;i++) {
    nodes[i]->assigned = TRUE;
    nodes[i]->enc = i-1;
    poss_encodings[i-1]->assigned = TRUE;
    poss_encodings[i-1]->node = nodes[i];
  }

 return;

} /* end of assign_initial_encoding */



generate_state(type)
int *type;
{
/*
* randomly generate an interchange or a displacement
*/

 int num1, num2;
 double RANDOM();


  num1 = RANDOM(0,encode_num);
  num2 = RANDOM(0,encode_num);
  if (num1 is num2 || num1 is encode_num || num2 is encode_num) return(FALSE);
  if (poss_encodings[num1]->assigned is FALSE &&
                           poss_encodings[num2]->assigned is FALSE)
    return(FALSE);
  if (poss_encodings[num1]->assigned is FALSE) {
    *type = DISPLACEMENT;
    shift_node1 = poss_encodings[num2]->node;
    shift_node2 = NIL(struct node);
    new_encode = num1;
    return(TRUE);
  }
  else if (poss_encodings[num2]->assigned is FALSE) {
    *type = DISPLACEMENT;
    shift_node1 = poss_encodings[num1]->node;
    shift_node2 = NIL(struct node);
    new_encode = num2;
    return(TRUE);
  }
  else {
    *type = INTERCHANGE;
    shift_node1 = poss_encodings[num1]->node;
    shift_node2 = poss_encodings[num2]->node;
    return(TRUE);
  }

} /* end of generate_state */


double
RANDOM(left,right)
int left,right;
{

 long divide,random(),random_num;
 double number;

   random_num = random();
   divide = 2147483647;
   number = ((double)random_num)/divide;
   number = number*abs(right - left) + left ;

 return(number);

} /* end of RANDOM */


double
find_starting_temp(anneal_type)
int anneal_type;
{

  double delta_c, temp;
  int find_cost();
/*
*  Let's not accept a move which increases the initial cost by more than
*   X percent.
*/

   init_cost = find_cost(TRUE, NULL, anneal_type);
   delta_c = 0.04 * (double)init_cost;
   temp = delta_c/log(1.25) ;
  return(temp);

} /* end of find_starting_temp */


int
find_cost(first, type, anneal_type)
int first, type, anneal_type;
{
/*
* find the cost of the given encoding
*/

 EDGE_PTR eptr;
 int num1, num2;
 int cost, ncost;
 int i;
 int check_constraint();

  cost = 0;
  if (first is TRUE) {
    if (anneal_type is GRAPH) {
      for(i=0;i<num_node_pairs;i++) {
        num1 = nodes[node_pairs[i].num1]->enc;
        num2 = nodes[node_pairs[i].num2]->enc;
        cost += node_pairs[i].cost * (distance(poss_encodings[num1]->cube,
                                             poss_encodings[num2]->cube) - 1);
      }
    }
    else if (anneal_type is TOTAL) {
      for(i=1;i<=node_num;i++) {
        for(eptr=nodes[i]->fanout;eptr!=NIL(struct edge_ptr);eptr=eptr->nextPtr)
          cost += distance(poss_encodings[nodes[i]->enc]->cube,
                           poss_encodings[eptr->edge->to_node->enc]->cube);
      }
    }
  }
  else {
    cost = init_cost;
    if (anneal_type is GRAPH) {
      if (type is INTERCHANGE) {
        for(i=0;i<num_node_pairs;i++) {
          num1 = node_pairs[i].num1;
          num2 = node_pairs[i].num2;
          ncost = node_pairs[i].cost;
          if (num1 is shift_node1->num && num2 is shift_node2->num) continue;
          if (num2 is shift_node1->num && num1 is shift_node2->num) continue;
          if (num1 is shift_node1->num) {
            cost += ncost * (distance(poss_encodings[shift_node2->enc]->cube,
                                  poss_encodings[nodes[num2]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                  poss_encodings[nodes[num2]->enc]->cube) - 1);
          }
          else if (num2 is shift_node1->num) {
            cost += ncost * (distance(poss_encodings[shift_node2->enc]->cube,
                                  poss_encodings[nodes[num1]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                  poss_encodings[nodes[num1]->enc]->cube) - 1);
          }
          else if (num1 is shift_node2->num) {
            cost += ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                   poss_encodings[nodes[num2]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node2->enc]->cube,
                                   poss_encodings[nodes[num2]->enc]->cube) - 1);
          }
          else if (num2 is shift_node2->num) {
            cost += ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                   poss_encodings[nodes[num1]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node2->enc]->cube,
                                   poss_encodings[nodes[num1]->enc]->cube) - 1);
          }
        }
      }
      else if (type is DISPLACEMENT) {
        for(i=0;i<num_node_pairs;i++) {
          num1 = node_pairs[i].num1;
          num2 = node_pairs[i].num2;
          ncost = node_pairs[i].cost;
          if (num1 is shift_node1->num) {
            cost += ncost * (distance(poss_encodings[new_encode]->cube,
                                   poss_encodings[nodes[num2]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                   poss_encodings[nodes[num2]->enc]->cube) - 1);
          }
          else if (num2 is shift_node1->num) {
            cost += ncost * (distance(poss_encodings[new_encode]->cube,
                                   poss_encodings[nodes[num1]->enc]->cube) - 1);
            cost -= ncost * (distance(poss_encodings[shift_node1->enc]->cube,
                                   poss_encodings[nodes[num1]->enc]->cube) - 1);
          }
        }
      }
    }
    else if (anneal_type is TOTAL) {
      if (type is INTERCHANGE) {
        for(eptr=shift_node1->fanout;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          if (eptr->edge->to_node isnt shift_node2) {
            cost += distance(poss_encodings[shift_node2->enc]->cube,
                          poss_encodings[eptr->edge->to_node->enc]->cube);
            cost -= distance(poss_encodings[shift_node1->enc]->cube,
                          poss_encodings[eptr->edge->to_node->enc]->cube);
          }
        }
        for(eptr=shift_node2->fanout;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          if (eptr->edge->to_node isnt shift_node1) {
            cost += distance(poss_encodings[shift_node1->enc]->cube,
                          poss_encodings[eptr->edge->to_node->enc]->cube);
            cost -= distance(poss_encodings[shift_node2->enc]->cube,
                          poss_encodings[eptr->edge->to_node->enc]->cube);
          }
        }
        for(eptr=shift_node1->fanin;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          if (eptr->edge->fr_node isnt shift_node2) {
            cost += distance(poss_encodings[shift_node2->enc]->cube,
                          poss_encodings[eptr->edge->fr_node->enc]->cube);
            cost -= distance(poss_encodings[shift_node1->enc]->cube,
                          poss_encodings[eptr->edge->fr_node->enc]->cube);
          }
        }
        for(eptr=shift_node2->fanin;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          if (eptr->edge->fr_node isnt shift_node1) {
            cost += distance(poss_encodings[shift_node1->enc]->cube,
                          poss_encodings[eptr->edge->fr_node->enc]->cube);
            cost -= distance(poss_encodings[shift_node2->enc]->cube,
                          poss_encodings[eptr->edge->fr_node->enc]->cube);
          }
        }
      }
      else if (type is DISPLACEMENT) {
        for(eptr=shift_node1->fanout;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          cost += distance(poss_encodings[new_encode]->cube,
                        poss_encodings[eptr->edge->to_node->enc]->cube);
          cost -= distance(poss_encodings[shift_node1->enc]->cube,
                        poss_encodings[eptr->edge->to_node->enc]->cube);
        }
        for(eptr=shift_node1->fanin;eptr!=NIL(struct edge_ptr);
                                                          eptr=eptr->nextPtr) {
          cost += distance(poss_encodings[new_encode]->cube,
                        poss_encodings[eptr->edge->fr_node->enc]->cube);
          cost -= distance(poss_encodings[shift_node1->enc]->cube,
                        poss_encodings[eptr->edge->fr_node->enc]->cube);
        }
      }
    }
  }

 return(cost);

} /* end of find_cost */


should_accept(change_in_cost, temp)
int change_in_cost;
double temp;
{

 double R,Y,RANDOM();


  if (change_in_cost < 0)
    return(TRUE);
  else {
    Y = exp(- (double)change_in_cost/temp);
    R = RANDOM(0,1);
    if (R < Y) 
      return(TRUE);
    else
      return(FALSE);
  }

} /* end of should_accept */


accept_or_reject_state(temp, type, anneal_type)
double temp;
int type;
int anneal_type;
{

 int new_cost, find_cost();
 int temp_enc;

   new_cost = find_cost(FALSE, type, anneal_type);
   if (should_accept(new_cost-init_cost, temp) is TRUE) {
     if (type is INTERCHANGE) {
       poss_encodings[shift_node1->enc]->node = shift_node2;
       poss_encodings[shift_node2->enc]->node = shift_node1;

       temp_enc = shift_node1->enc;
       shift_node1->enc = shift_node2->enc;
       shift_node2->enc = temp_enc;
     }
     else if (type is DISPLACEMENT) {
       poss_encodings[shift_node1->enc]->node = NIL(struct node);
       poss_encodings[shift_node1->enc]->assigned = FALSE;

       poss_encodings[new_encode]->node = shift_node1;
       poss_encodings[new_encode]->assigned = TRUE;

       shift_node1->enc = new_encode;
     }
     init_cost = new_cost;
   }

  return;

} /* end of accept_or_reject_state */

  

double
update_temp(temp, new_states)
double temp;
int *new_states;
{

  double new_temp,factor;

  if (sqrt(START_TEMP) < temp )
     new_temp = temp*ALPHA*ALPHA ;
  else
     new_temp = ALPHA * temp;
  if (new_temp >= 1.0 ) 
     factor =  log(new_temp)/log(START_TEMP);
  else
     factor = 0;
  *new_states = (int) ((STATES_END - 
                             (STATES_END - STATES_BEGIN)*factor) * node_num);

  return(new_temp);

} /* end of update_temp */
  
