#include "copyright.h"
#include "mustang.h"


/*
 *    ***********************************
 *    *** SEQUENTIAL STATE ASSIGNMENT ***
 *    ***********************************
 *
 *  Srinivas Devadas
*/


assign_closest_possible_codes(node1, node2, num)
NODE node1, node2;
int num;
{
/*
* Assign two minimally distant codes from unassigned codes 
* the two next state nodes.
*/

 CUBE_LIST BEST_cptr1, BEST_cptr2;
 int dist, BEST_dist;
 int distance();
 CUBE_LIST cptr1, cptr2;
 int i, j;


  BEST_dist = INFINITY;
  for(i=0;i<encode_num;i++) {
    if (poss_encodings[i]->assigned is TRUE) continue;
    cptr1 = poss_encodings[i];
    for(j=i+1;j<encode_num;j++) {
      if (poss_encodings[j]->assigned is TRUE) continue;
      cptr2 = poss_encodings[j];
      if ( (dist  = distance(cptr1->cube, cptr2->cube)) is 1) {

        /* assign uni-distance codes */
        node1->assigned = node2->assigned = TRUE;
        cptr1->assigned = cptr2->assigned = TRUE;
        node1->enc = cptr1->num;
        node2->enc = cptr2->num;
        (void) strcpy(node1->code[num], cptr1->cube);
        (void) strcpy(node2->code[num], cptr2->cube);
        if (debug)
          (void) printf("#assigned %s with %s  &  %s with %s\n",
            node1->nodeName,node1->code[num],node2->nodeName,node2->code[num]);
        return;
      }
      else if (dist > 1) {
        if (dist < BEST_dist) {
          
          /* store result */
          BEST_cptr1 = cptr1;
          BEST_cptr2 = cptr2;
          BEST_dist = dist;
        }
      }
    }
  }
  if (BEST_dist is INFINITY)
    error("couldn't find two unassigned codes");
  else {
    node1->assigned = node2->assigned = TRUE;
    BEST_cptr1->assigned = BEST_cptr2->assigned = TRUE;
    node1->enc = BEST_cptr1->num;
    node2->enc = BEST_cptr2->num;
    (void) strcpy(node1->code[num], BEST_cptr1->cube);
    (void) strcpy(node2->code[num], BEST_cptr2->cube);
    if (debug)
      (void) printf("assigned %s with %s  &  %s with %s\n",
         node1->nodeName,node1->code[num],node2->nodeName,node2->code[num]);
  } 

 return;

} /* end of assign_closest_possible_codes */



assign_code_closest_to_node(node1, node2, num)
NODE node1, node2;
int num;
{
/*
* node2 has already been assigned, find a minimally distant
* code to node2's assignment in poss_encodings.
*/

 CUBE_LIST cptr, BEST_cptr;
 int dist, BEST_dist;
 int distance();
 int i;

  if (node2->assigned is FALSE)
    error("node2 has not been assigned");

  BEST_dist = INFINITY;
  for(i=0;i<encode_num;i++) {
    cptr = poss_encodings[i];
    if (cptr->assigned is TRUE) continue;
    if ( (dist = distance(cptr->cube, node2->code[num]) ) is 1) {
      node1->assigned = TRUE;
      cptr->assigned = TRUE;
      node1->enc = cptr->num;
      (void) strcpy(node1->code[num], cptr->cube);
      if (debug)
        (void) printf("#Given %s with %s assigned %s with %s\n",
           node2->nodeName,node2->code[num],node1->nodeName,node1->code[num]);
      return;
    }
    else if ( dist > 1 ) {
      if (dist < BEST_dist) {
        BEST_dist = dist;
        BEST_cptr = cptr;
        BEST_dist = dist;
      }
    }
  }

  if (BEST_dist is INFINITY) 
    error("Couldn't find a assignable code");
  else {
    node1->assigned = TRUE;
    BEST_cptr->assigned = TRUE;
    node1->enc = BEST_cptr->num;
    (void) strcpy(node1->code[num], BEST_cptr->cube);
    if (debug)
      (void) printf("#Given %s with %s assigned %s with %s\n",
         node2->nodeName,node2->code[num],node1->nodeName,node1->code[num]);
  }

 return;

} /* end of assign_code_closest_to_node */
   

assign_codes_to_cluster(nlist, num_nodes, dist, num)
int *nlist;
int num_nodes;
int dist;
int num;
{
/*
* Assign the nodes in nlist to a unit distance cluster from the first
* node.
*/
 CUBE_LIST cptr1, cptr2;
 int i, j, k, count;

  if (num_nodes > num_states + 1)
    error("too large a number of nodes in cluster");
  for(i=1;i<=num_nodes;i++) {
    if (nodes[nlist[i]]->assigned is TRUE)
      error("node has already been assigned");

  }

  for(i=0;i<encode_num;i++) {
    count = 1;
    cptr1 = poss_encodings[i];
    if (cptr1->assigned is FALSE) {
      for(j=0;j<encode_num;j++) {
        cptr2 = poss_encodings[j];
        if (cptr1 is cptr2 || cptr2->assigned is TRUE) continue;
        if (distance(cptr1->cube, cptr2->cube) <= dist)
          count++;
        if (count >= num_nodes) break;
      }
    }
    if (count >= num_nodes ) {
      count = 0;
      (void) strcpy(nodes[nlist[++count]]->code[num],cptr1->cube);
      cptr1->assigned = TRUE;
      nodes[nlist[count]]->assigned = TRUE;
      nodes[nlist[count]]->num = cptr1->num;
      for(j=0;j<encode_num;j++) {
        cptr2 = poss_encodings[j];
        if (cptr1 is cptr2 || cptr2->assigned is TRUE) continue;
        if (distance(cptr1->cube, cptr2->cube) <= dist) {
          (void) strcpy(nodes[nlist[++count]]->code[num],cptr2->cube);
          cptr2->assigned = TRUE;
          nodes[nlist[count]]->assigned = TRUE;
          nodes[nlist[count]]->num = cptr2->num;
        }
        if (count >= num_nodes) {
          if (! debug) return;
          (void) printf("Cluster");
          for(k=1;k<=num_nodes;k++) 
            (void) printf(" %s with %s",nodes[nlist[k]]->nodeName,
                                                nodes[nlist[k]]->code[num]);
          (void) printf("\n");
          return;
        }
      }
    }
  }
  (void) printf("#Couldnt make assignment of distance %d to cluster\n",dist);

 return;

} /* end of assign_codes_to_clusters */
        


distance(cube1, cube2)
STRING cube1, cube2;
{

 int i;
 int dist, pdist;

  dist = pdist = 0;
  for(i=0;i<strlen(cube1);i++) {
    if (cube1[i] isnt cube2[i] && cube1[i] isnt '-' && cube2[i] isnt '-')
      dist++;
    if ( (cube1[i] is '-' && cube2[i] isnt '-') ||
         (cube1[i] isnt '-' && cube2[i] is '-'))
      pdist++;
  }
 
 return(dist + pdist/2);

} /* end of distance */


assign_present_state_codes(two_level, approximate)
int two_level, approximate;
{
/*
* Assign present state codes, looking at the sets of edges
* corresponding to o/p = 1 and o/p = 0 for all outputs
*/
 int i, j, k, count;
 int **op_node, **st_node;
 int cost;


  op_node = ALLOC(int *, num_outputs + num_states + 1);
  for(i=0; i< num_outputs + num_states + 1; i++ )
    op_node[i] = ALLOC(int, node_num + 1);

  st_node = ALLOC(int *, node_num + 1);
  for(i=0; i< node_num + 1; i++ )
    st_node[i] = ALLOC(int, node_num + 1);

  for(i=0;i<num_outputs + num_states + 1;i++) {
    for(j=0;j<node_num + 1;j++) {
      op_node[i][j] = 0;
    }
  }
  for(i=0;i<node_num + 1;i++) {
    for(j=0;j<node_num + 1;j++) {
      st_node[i][j] = 0;
    }
  }
 
  for(j=0;j<num_outputs;j++) {
    for(i=1;i<=edge_num;i++) {
      if (edges[i]->output[j] is '1') {
        op_node[j][edges[i]->fr_node->num] +=  1;
      }
    }
  }
  for(i=1;i<=edge_num;i++) 
    st_node[edges[i]->to_node->num][edges[i]->fr_node->num] += 1;

  count = 0;
  for(i=1;i<=node_num;i++) {
    for(j=i+1;j<=node_num;j++) {
      node_pairs[count].num1 = i;
      node_pairs[count].num2 = j;
      cost = 0;
      for(k=0; k<num_outputs;k++)
        cost += op_node[k][i] * op_node[k][j];

      /* weigh the next states by the number of encoding bits */
      if (two_level is TRUE) {
        for(k=1;k<=node_num;k++)
          cost += st_node[k][i] * st_node[k][j] * num_states;
      }
      else {
        for(k=1;k<=node_num;k++)
          cost += st_node[k][i] * st_node[k][j];
      }

      node_pairs[count].cost = cost;
      if (!annealing || cost > 0)
        count++;
    }
  }

  if (approximate) {
    (void) printf("Number of edges is %d\n", count);
    (void) printf("Give number to be used:");
    (void)fscanf(stdin, "%d", &num_node_pairs);
    sort_node_pairs(node_pairs, num_node_pairs);
  }
  else {
    sort_node_pairs(node_pairs, count);
    num_node_pairs = count;
  }

  if (annealing) {
    simulated_anneal(GRAPH);
    for(i=1;i<=node_num;i++) {
      nodes[i]->assigned = TRUE;
      (void) strcpy(nodes[i]->code[0], poss_encodings[nodes[i]->enc]->cube);
    }
  }
  else {
    wedge_clustering(0);
    if (debug) {
      for(i=1;i<=node_num;i++)
        poss_encodings[nodes[i]->enc]->node = nodes[i];
      (void) printf("# Cost of assignment is %d\n",
                                        find_cost(TRUE,NULL,GRAPH));
    }
  }

 return;

} /* end of assign_present_state_codes */



exploit_dont_cares(num)
int num;
{
/*
* Try to introduce don't care literals (-) in present and next state
* assignments
*/
 CUBE_LIST cptr;
 int i, j, can_merge();


  for(i=0;i<encode_num;i++) {
    cptr = poss_encodings[i];
    cptr->assigned = FALSE;
    for(j=1;j<=node_num;j++) {
      if (strcmp(cptr->cube,nodes[j]->code[num]) is FALSE) {
        cptr->assigned = TRUE;
        break;
      }
    }
  }
  for(i=1;i<=node_num;i++) {
    for(j=0;j<encode_num;j++) {
      cptr = poss_encodings[j];
      if (cptr->assigned is TRUE) continue;
      if (can_merge(cptr->cube,nodes[i]->code[num]) is TRUE) {
        merge(cptr->cube,nodes[i]->code[num]);
        cptr->assigned = TRUE;
        break;
      }
    }
  }

 return;

} /* end of exploit_dont_cares */


can_merge(cube1,cube2)
STRING cube1, cube2;
{
/*
* check if two cubes can be merged
*/

 int i, diff;

  diff = 0;
  for(i=0;i<strlen(cube1);i++)
    if (cube1[i] isnt cube2[i]) diff++;
  if (diff is 1) return(TRUE);
  else return(FALSE);

} /* end of can_merge */



merge(cube1, cube2)
STRING cube1, cube2;
{
/*
* merge the two cubes into cube2
*/
 int i;

  for(i=0;i<strlen(cube1);i++) {
    if (cube1[i] isnt cube2[i]) {
      cube2[i] = '-';
      break;
    }
  }

 return;

} /* end of merge */


add_dont_cares(num)
int num;
{

  FILE *dc_file;
  char in_cube[NUM_INPUTS], out_cube[NUM_OUTPUTS];
  char line[MAXLINE];
  int i, num_bits;

 
   num_bits = strlen(nodes[1]->code[num]);
   for(i=0;i<num_bits+num_outputs+1;i++) {
     if (i is num_bits) out_cube[i] = ' ';
     else out_cube[i] = '-';
   }
   out_cube[i] = '\0';
 
   for(i=0;i<num_inputs;i++) 
     in_cube[i] = '-';
   in_cube[i] = '\0';

   if ( (dc_file = fopen("temp3", "w")) is NULL)
     error("could not open dc PLA file");
 
   (void) fprintf(dc_file, ".i %d\n", num_bits);
   (void) fprintf(dc_file, ".o 1\n");
 
   for(i=1;i<=node_num;i++) {
     (void) fprintf(dc_file, "%s 1\n", nodes[i]->code[num]);
   }
 
   (void) fclose(dc_file);
 
   /* run espresso to find the complement of the encodings used */
   (void) printf("# Complementation of encoding using ESPRESSO begun\n");
   if (system("espresso -out r < temp3 > temp4") isnt 0) 
      error("count not run espresso");

   (void) printf("# Complementation of encoding using ESPRESSO completed\n");
 
   if ( (dc_file = fopen("temp4", "r")) is NULL)
     error("could not open dc PLA complement file");
 
   NextLine = 0;
   while(fgets(line, MAXLINE, dc_file) isnt NULL) {
     NextLine++;
     parse_line(line);
     if (targv[0] is NULL) continue;
     if (targv[0][0] is '.' || targv[0][0] is '#') continue;
     (void) printf("%s %s %s\n", in_cube, targv[0], out_cube);
   }
   (void) fflush(stdout);

   if (system("rm -f temp3") isnt 0) error("could not remove file temp3");
   if (system("rm -f temp4") isnt 0) error("could not remove file temp4");

  return;

} /* end of add_dont_cares */



wedge_clustering(num)
int num;
{
/*
* First try assigning node clusters to codes which are minimally
* distant.
*/

 int pair_no, next;
 int num1, num2;
 int *node_list, num_nodes;
 int i, all_assigned;
 NODE node1, node2;


  node_list = ALLOC(int, node_num + 1);

  for(i=1;i<=node_num;i++)
    nodes[i]->assigned = FALSE;
  for(i=0;i<encode_num;i++)
    poss_encodings[i]->assigned = FALSE;

  if (cluster) {
    for(pair_no=1;pair_no<num_node_pairs;pair_no++) {
      num1 = node_pairs[pair_no].num1;
      num2 = node_pairs[pair_no].num2;
      if (nodes[num1]->assigned is TRUE || nodes[num2]->assigned is TRUE)
        continue;
      num_nodes = 0;
      node_list[++num_nodes] = num1;
      node_list[++num_nodes] = num2;
      for(next=pair_no+1;next< num_node_pairs;next++) {
        if (node_pairs[next].num1 is num1 && node_pairs[next].num2 isnt num2) {
          if (nodes[node_pairs[next].num2]->assigned is FALSE)
            node_list[++num_nodes] = node_pairs[next].num2;
        }
        if (node_pairs[next].num2 is num1 && node_pairs[next].num1 isnt num2) {
          if (nodes[node_pairs[next].num1]->assigned is FALSE)
            node_list[++num_nodes] = node_pairs[next].num1;
        }
        if (num_nodes is num_states) break;
      }
/*
* the nodes are assigned to unit distance clusters, if that's not
* possible no assignment is done.
*/
      assign_codes_to_cluster(node_list, num_nodes, 1, num );
      all_assigned = TRUE;
      for(i=1;i<=node_num;i++) {
        if (nodes[i]->assigned is FALSE) {
          all_assigned = FALSE;
          break;
        }
      }
      if (all_assigned) break;
    }
  }

  pair_no = 0;
  all_assigned = FALSE ;
  while( (!all_assigned) && pair_no < num_node_pairs) {
    node1 = nodes[node_pairs[pair_no].num1];
    node2 = nodes[node_pairs[pair_no].num2];
    if (node1->assigned is FALSE && node2->assigned is FALSE) {
      assign_closest_possible_codes(node1, node2, num);
    }
    else if (node1->assigned is TRUE && node2->assigned is FALSE) {
      assign_code_closest_to_node(node2, node1, num);
    }
    else if (node1->assigned is FALSE && node2->assigned is TRUE) {
      assign_code_closest_to_node(node1, node2, num);
    }
    all_assigned = TRUE;
    for(i=1;i<=node_num;i++) {
      if (nodes[i]->assigned is FALSE) {
        all_assigned = FALSE;
        break;
      }
    }
    pair_no++;
  }
  if (all_assigned is FALSE) {
    for(i=1;i<=node_num;i++) {
      if (nodes[i]->assigned is FALSE) {
        assign_code_closest_to_node(nodes[i], nodes[1], num);
      }
    }
  }
  FREE(node_list);

 return;

} /* end of wedge_clustering */
