#include "copyright.h"
#include "mustang.h"
#include <sys/times.h>


/*
 *    ***********************************
 *    *** SEQUENTIAL STATE ASSIGNMENT ***
 *    ***********************************
 *
 *  Srinivas Devadas
 */

main(argc, argv)
int argc;
char **argv;
{
  struct plist **HAcreate();
  int i,j;
  int next_state, present_state, arandom, sequential, two_level, one_hot_encode;
  int approximate;
  int translate;
  FILE *input, *blif_input;
  char inpFile[ID_LENGTH], blifFile[ID_LENGTH];
  void timer();

  if (argc < 2) {
    (void) printf("usage: mustang [options] input_file\n");
err:
    (void) printf("General Options:\n");
    (void) printf("         -d: debug\n");
    (void) printf("\nEncoding Options:\n");
    (void) printf("         -ran: dynamic random encoding\n");
    (void) printf("         -r: deterministic random encoding\n");
    (void) printf("         -s: sequential encoding\n");
    (void) printf("         -l: one hot encode FSM\n");
    (void) printf("\nMulti-level State Assignment Options:\n");
    (void) printf("         -n: fanin oriented algorithm\n");
    (void) printf("         -p: fanout oriented algorithm\n");
    (void) printf("         -t: favor two-level implementations\n");
    (void) printf("         -c: wedge clustering algorithm for embedding\n");
    (void) printf("         -a: annealing algorithm for embedding\n");
    (void) printf("         -e: expand state codes\n");
    (void) printf("         -i: specify annealing parameters\n");
    (void) printf("         -x: approximate by discarding edges in graph\n");
    (void) printf("         -o: generate one hot like codes\n");
    (void) printf("         -onehot: onehot relationships from blif file\n");
    exit(-1);
  }
  i = 1;

  debug = FALSE; next_state = present_state = FALSE;
  cluster = FALSE; two_level = onehotlike = FALSE;
  annealing = FALSE; specify = FALSE; approximate = FALSE;
  one_hot_encode = translate = FALSE;
  arandom = sequential = real_ran = FALSE;

  edge_num = 0;  
  (void) strcpy(inpFile, "xxxx");
  while(i< argc) {
    if (strcmp(argv[i],"-onehot") is 0) {
      one_hot_encode = TRUE;
      (void) strcpy(blifFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-ran") is 0) {
      real_ran = TRUE;
      arandom = TRUE;
      i+=1;
    }
    else if (argv[i][0] is '-') {
      j = 1;
      while (argv[i][j] isnt '\0') {
        if (argv[i][j] is 'd') {
          debug = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'e') {
          expand = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'x') {
          approximate = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'c') {
          cluster = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 't') {
          two_level = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'o') {
          onehotlike = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'n') {
          next_state = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'r') {
          arandom = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 's') {
          sequential = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'i') {
          specify = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'a') {
          annealing = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'l') {
          translate = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'p') {
          present_state = TRUE;
          j++ ;
        }
        else {
          (void) printf("mustang: unknown option\n");
          goto err;
        }
      }
      i++ ;
    }
    else  {
      (void) strcpy(inpFile,argv[i]);
      i++ ;
    }
  }
  if (strcmp(inpFile, "xxxx") is FALSE) {
    input = stdin;
  }
  else {
    if ( (input = fopen(inpFile,"r")) is NULL) {
       (void) printf("Couldn't open input file\n");
       goto err;
    }
  }
  if (one_hot_encode is TRUE) {
    if ( (blif_input = fopen(blifFile,"r")) is NULL) {
       (void) printf("Couldn't open blif file\n");
       exit(1) ;
    }
  }
  timer(stderr,"START");
  Nodehash = HAcreate(NODEHASH);
  if (one_hot_encode is TRUE) {
/*
* Encode the input space depending on the number of bits given
* using the optimized multi-level logic network as a guide.
*/
      NextLine = 0; node_num = 0;
      read_input(input);
      one_hot_based_encoding(blif_input);
  }
  else if (translate is TRUE) {
/*
* assign each input to be encoded, a single literal i.e. 1--, -1- etc.
*/
    NextLine = 0; node_num = 0;
    read_input(input);
    single_line_encode();
  }
  else {
/*
* read in the state transition table.
*/
      NextLine = 0; node_num = 0;
      read_input(input);
/*
*  Assign encodings to each state in the state transition graph for
*  optimal multi-level logic implementations.
*/
      state_assignment(next_state, present_state, arandom, sequential,
                                                   two_level, approximate);
  }
/*
  timer(stderr,"total");
*/
  exit(0);

} /* end of main */



read_input(input)
FILE *input;
{
/*
*  read in the state transition diagram
*/

  NODE node1,node2,findNode();
  EDGE edge;
  char line[MAXLINE], tline[MAXLINE];
  char nodeName1[ID_LENGTH], nodeName2[ID_LENGTH];
  int i;


   while(fgets(line,MAXLINE,input) isnt NULL) {
     NextLine++;
     (void) strcpy(tline,line);
     parse_line(tline);
     if (targv[0] is NULL || targv[0][0] is '#') continue;
     if (strcmp(targv[0],"START") is FALSE) {
       (void)sprintf(nodeName1,"%s",targv[1]);
       start_node = findNode(nodeName1);
       continue;
     }
     else if (strcmp(targv[0],".i") is FALSE) {
       num_inputs = atoi(targv[1]);
       continue;
     }
     else if (strcmp(targv[0],".o") is FALSE) {
       num_outputs = atoi(targv[1]);
       continue;
     }
/*
* for state assignment generation.
*/
     else if (strcmp(targv[0],".s") is FALSE) {
       num_states = atoi(targv[1]);
       continue;
     }
     else if (targv[0][0] is '.') {
       for(i=0; i < targc; i++)
         (void) printf("%s ", targv[i]);
       (void) printf("\n");
     }
     else if (targc is 4 ) {
     
       if (num_inputs is 0) {
         (void) printf("Specify number of inputs using .i num_inp\n");
         exit(-1);
       }
       if (num_outputs is 0) {
         (void) printf("Specify number of outputs using .o num_out\n");
         exit(-1);
       }
       if (num_states is 0) {
         (void) printf("Specify number of encoding bits using .s num_bits\n");
         exit(-1);
       }

       /* read in an edge in the graph */
       (void)sprintf(nodeName1,"%s",targv[1]);
       (void)sprintf(nodeName2,"%s",targv[2]);
       node1 = findNode(nodeName1);
       node2 = findNode(nodeName2);

       /* allocate and fill in edge */
       edge = ALLOC(struct edge, 1);
       edge_num += 1;
       edges[edge_num] = edge;

       edge->input = ALLOC(char, num_inputs + 1);

       (void) strcpy(edge->input,targv[0]);

       edge->output = ALLOC(char, num_outputs + 1);
       (void) strcpy(edge->output,targv[3]);

       /* build the graph */
       if (node1 is node2)
        add_edge(node1,edge,LOOP);
       else {
         add_edge(node1,edge,FANOUT);
         add_edge(node2,edge,FANIN);
       }

     }
     else
       error("wrong input specification");
   }

 return;

} /* end of read_input */


add_edge(node,edge,type)
NODE node;
EDGE edge;
int type;
{
/*
* add the edge to the node's fanin or fanout list, and the node to
* the edge's node pointers.
*/

 EDGE_PTR e_ptr;

  if (type is FANIN) {
    e_ptr = ALLOC(struct edge_ptr, 1);
    e_ptr->edge = edge;
    e_ptr->nextPtr = node->fanin;
    node->fanin = e_ptr;
    edge->to_node = node;
  }
  else if (type is FANOUT) {
    e_ptr = ALLOC(struct edge_ptr, 1);
    e_ptr->edge = edge;
    e_ptr->nextPtr = node->fanout;
    node->fanout = e_ptr;
    edge->fr_node = node;
  }
  else if (type is LOOP) {
    e_ptr = ALLOC(struct edge_ptr, 1);
    e_ptr->edge = edge;
    e_ptr->nextPtr = node->loop;
    node->loop = e_ptr;
    edge->fr_node = node;
    edge->to_node = node;
  }
  else {
    error("in add_edge");
  }

 return;

} /* end of add_edge */


/*
*  parse input line into tokens, filling up targv and setting targc 
*/
parse_line(line)
  register char *line;
  {	register char **carg = targv;
	register char ch;

	targc = 0;
	while (ch = *line++) {
	  if (ch <= ' ') continue;
	  targc++;
	  *carg++ = line-1;
	  while ((ch = *line) && ch > ' ' ) line++;
	  if (ch) *line++ = '\0';
	}
	*carg = 0;

} /*  end of parse_line  */


/*
 * print error message and die
 */
error(message)
STRING message;
{
    (void) printf("mustang: %s around line %d\n",message,NextLine);
    exit(-1);
}



NODE
findNode(nodeName)
/*
 * check for presence of state, else allocate one.
 *
 */
STRING nodeName;
{
    NODE np;
    STRING loc, hname;
 
    if(HAfind(nodeName, &loc, Nodehash, NODEHASH) != FALSE) {
        np = (struct node *)loc;
        return (np);
    }
    np = ALLOC(struct node, 1);
    if(HAinstall(nodeName,(char *)np,&hname,ELMEN,Nodehash,NODEHASH) != OK) {
        error("node installation failed");
    }

    node_num += 1 ; 
    nodes[node_num] = np;
    np->num = node_num;
    np->code[0] = np->code[1] = np->code[2] = NIL(char);
    np->nodeName = hname;

    return(np);
  

} /*  end of findNode  */



write_output()
{
/*
* write out the state assigned automaton
*/

 EDGE edge;
 int i;


  (void) printf("#STATE ASSIGNED FINITE AUTOMATON\n");
  for(i=1;i<=node_num;i++)
    (void) printf("# %s %s %s\n",
               nodes[i]->nodeName, nodes[i]->code[0], nodes[i]->code[1]);
  (void) printf(".i %d\n", num_inputs + num_states);
  (void) printf(".o %d\n", num_outputs + num_states);
  
  /* add the dont cares right here */
  (void) printf(".type fr\n");

  for(i=1;i<=edge_num;i++) {
    edge = edges[i];
    (void) printf("%s %s %s %s\n", edge->input,
       edge->fr_node->code[0], edge->to_node->code[1], edge->output);
  }

 return;

} /* end of write_output */



/*
 *    get the elapsed cputime
 *        (call it the first time with message "START")
 */ 
struct tms Buffer;
double StartTime, LastTime;

void
timer( file, message )
FILE *file;
char *message;
{
    if(strcmp(message,"START") == 0) {
        (void) times(&Buffer);
        StartTime = (double)(Buffer.tms_utime);
        LastTime = StartTime;
        return;
    }
    (void) times(&Buffer);
    (void) fprintf(file,"#mustang: cputime %s: %.1fs %.1fs\n", message,
        (Buffer.tms_utime-LastTime)/60.0, (Buffer.tms_utime-StartTime)/60.0);
    LastTime = Buffer.tms_utime;
    return;
}

