#include "copyright.h"
#include "port.h"
#include "utility.h"

/*
***************************************************************
*                         MUSTANG                             *
*         Header file for State Assignment Program            *
*   This program finds the optimal state assignment for       *
*  a Finite State Machine for a multi-level implementation.   *
*                                                             *
*               AUTHOR :  Srinivas Devadas                    *
***************************************************************
*/

#define TRUE	1
#define FALSE	0

#define MAXLINE 1000
#define NUM_NODES 10000
#define NUM_INPUTS 100
#define NUM_OUTPUTS 100
#define NUM_EDGES 10000
#define NUM_CLUSTERS 100
#define ID_LENGTH 100
#define NODEHASH 2000

#define OK 1
#define ELMEN 1
#define NOROOM -1
#define DUP 2

#define is ==
#define isnt !=
#define INFINITY 100000

#define FANIN 1
#define FANOUT 2
#define LOOP 3

#define GRAPH 1
#define TOTAL 2

typedef char *STRING;

struct	plist {		/* basic hash table entry */
	char *name;	/* string to be hashed */
	char *loc; 	/* pointer to element `name' (cast to char) */
	int  type;	/* either ELEMENT or TABLE  (what loc points to) */
	struct plist *next;	/* next entry in chain */
};
 

typedef struct edge {
   
   STRING input, output;
   STRING enc, nenc;
   struct node *fr_node, *to_node;

} *EDGE;

typedef struct edge_ptr {
  
   EDGE edge;
   struct edge_ptr *nextPtr;

} *EDGE_PTR;


typedef struct node {

   STRING nodeName;
   STRING code[3];
   int num;
   int assigned;
   int enc;
   int level_from_top, level_from_bot;
   int remove;
   EDGE_PTR fanin, fanout, loop;

} *NODE;
   

typedef struct relative {
     struct node *nptr;          /* pointer to the corresponding node */
     struct relative *nextRel;   /* linked list of relatives          */

} *RELATIVE;


typedef struct pair {
   int num1;
   int num2;
   int cost;
} PAIR;

typedef struct cluster {
    int num;
    int weight;
    struct cluster *nextPtr;
} *CLUSTER;

typedef struct cube_list {

    int num;
    char *cube;                /* encoding */
    short assigned;            /* whether it has been used or not */
    NODE node;                 /* ptr to node */
    struct cube_list *nextPtr; /* ptr to next cube */

} *CUBE_LIST;

char *targv[100];	/* pointer to tokens on current code line */
int targc;		/* number of args on current code line */
int32 NextLine;
/*
* Options
*/
int debug, cluster, annealing, specify, real_ran, onehotlike, expand;

int num_inputs, num_outputs, num_states;
int node_num, edge_num;

PAIR *node_pairs;                    /* sorted node pairs */
struct plist **Nodehash;
NODE start_node, nodes[NUM_NODES];
EDGE edges[NUM_EDGES];
CUBE_LIST poss_encodings[NUM_NODES];
int encode_num;         /* total number of possible encodings */
int num_node_pairs;     /* total number of edges of non-zero weight */
