#include "copyright.h"
#include "mustang.h"


/*
 *    ***********************************
 *    *** SEQUENTIAL STATE ASSIGNMENT ***
 *    ***********************************
 *
 *  Srinivas Devadas
*/


one_hot_based_encoding(input)
FILE *input;
{
/*
* "input" is the blif file which has been optimized by the multi-level logic
* optimizer, giving one hot codes to the different states.  We now
* try to encode the states within a certain number of bits, so as to
* minimize the increase in the number of literals.
*/

 int i;

/*
*  Given the number of bits, decisions are made as to how many
*  literals will exist in each state encoding.
*/
  encode_num = 0;
  if (onehotlike)
    one_hot_like_encodings();
  else
    minterm_encodings(num_states);

/*
* allocate memory for optimal assignment of next state and present
* state nodes and in node data structure.
*/
  for(i=1;i<=node_num;i++) {
    nodes[i]->code[0] = ALLOC(char, num_states + 1);
    nodes[i]->code[1] = ALLOC(char, num_states + 1);
  }

/*
* derive the one hot constraints from the optimized BLIF file
*/
  derive_one_hot_constraints(input);

  if (annealing) {
    simulated_anneal(GRAPH);
    for(i=1;i<=node_num;i++) {
      nodes[i]->assigned = TRUE;
      (void) strcpy(nodes[i]->code[0], poss_encodings[nodes[i]->enc]->cube);
      (void) strcpy(nodes[i]->code[1], poss_encodings[nodes[i]->enc]->cube);
    }
  }
  else {
    wedge_clustering(0);
    for(i=1;i<=node_num;i++)
      (void) strcpy(nodes[i]->code[1], nodes[i]->code[0]);
  }

  /* put 0's instead of -'s in the next state codes */
  correct_next_state_codes();

  write_output();
  add_dont_cares(0);

 return;
      
} /* end of one_hot_based_encoding */


same_node(num, name)
int num;
STRING name;
{

 int i;
 char new_name[ID_LENGTH];

  if (name[0] isnt 'v') return(FALSE);
  for(i=1;i<strlen(name);i++) {
    new_name[i-1] = name[i];
  }
  new_name[strlen(name)-1] = '\0';

  if (num + num_inputs - 1 is atoi(new_name)) return(TRUE);

 return(FALSE);

} /* end of same_node */



single_line_encode()
{

 STRING mult_value1, mult_value2;
 int i, j;
 NODE node1, node2;
 EDGE edge;

  mult_value1 = ALLOC(char, node_num + 1);
  mult_value2 = ALLOC(char, node_num + 1);
  (void) printf(".i %d\n.o %d\n", num_inputs + node_num,
                                              num_outputs + node_num);
  for(i=1;i<=edge_num;i++) {
    edge = edges[i];
    node1 = edge->fr_node;
    node2 = edge->to_node;
    for(j=1;j<=node_num; j++) {
     if (j is node1->num)
        mult_value1[j-1] = '1';
      else
        mult_value1[j-1] = '-';
      if (j is node2->num)
        mult_value2[j-1] = '1';
      else
        mult_value2[j-1] = '0';
    }

    mult_value1[node_num] = mult_value2[node_num] = '\0';
    (void) printf("%s %s %s %s\n", edge->input, mult_value1,
                                                mult_value2, edge->output);
   }

  FREE(mult_value1);
  FREE(mult_value2);

 return;

} /* end of single_line_encode */


generate_codes(S, n, limit, ch)
STRING S;
int n, limit;
char ch;
{
    int fzero, fone;

    /* find the first "ch" */
    fzero = n-1;
    while (fzero >=0 && S[fzero] != ch) fzero--;

    /* done? */
    if (fzero < n-limit) {
	return(TRUE);
    }

    /* find the first '1' after the "ch" */
    fone = fzero-1;
    while (fone >= 0 && S[fone] != '1') fone--;

    /* move the first 1 */
    if (fone < 0) {
	S[0] = '1';
	fone = 0;
    } else {
	S[fone] = ch;
	S[fone+1] = '1';
	fone++;
    }

    /* move the leading 1's */
    for (fzero = fzero+1; fzero<n; fzero++) {
	S[fzero] = ch;
	S[++fone] = '1';
    }

  return(FALSE);

} /* end of generate_codes */


int
comb(n, k)
int n, k;
{

 int i, num;

  num = 1;
  for(i=0;i<k;i++) num *= n - i;
  for(i=1;i<=k;i++) num /= i;

 return(num);

} /* end of comb */


derive_one_hot_constraints(input)
FILE *input;
{

 int i, j, count;
 int **enc_node;
 int exist[NUM_NODES];
 char line[MAXLINE];

  enc_node = ALLOC(int *, node_num + 1);
  for(i=0;i<node_num+1;i++) 
    enc_node[i] = ALLOC(int, node_num + 1);

  for(i=0;i<=node_num;i++) 
    for(j=0;j<=node_num;j++) 
      enc_node[i][j] = 0;

  NextLine = 0;
  while(fgets(line, MAXLINE, input) isnt NULL) {
    NextLine++;
    parse_line(line);
    if (targv[0][0] is '#' || targv[0] is NULL) continue;
    if (strcmp(targv[0],".names") is FALSE) {
      
      /* find the relationship between input lines to be encoded */
      for(i=1;i<=node_num;i++) {
        exist[i] = FALSE;
        for(j=1;j<targc-1;j++) {
          if (same_node(i, targv[j]) is TRUE) {
            exist[i] = TRUE;
            break;
          }
        }
      }
          
      /* build the constraint matrix */
      for(i=1;i<=node_num;i++) {
        for(j=i+1;j<=node_num;j++) {
          if (exist[i] && exist[j]) {
            enc_node[i][j] += 1;
          }
        }
      }
    }
  }

  /* make the node pair list */
  node_pairs = ALLOC(PAIR, node_num * node_num + 1);

  count = 0;
  for(i=1;i<=node_num;i++) {
    for(j=i+1;j<=node_num;j++) {
      node_pairs[count].num1 = i;
      node_pairs[count].num2 = j;
      node_pairs[count].cost = enc_node[i][j];
      if (enc_node[i][j] > 0)
        count++;
    }
  }

  sort_node_pairs(node_pairs, count);
  num_node_pairs = count;

  if (debug) {
    for(i=1;i< num_node_pairs;i++) {
      (void) printf("#Node %s and Node %s weight %d\n",
       nodes[node_pairs[i].num1]->nodeName, nodes[node_pairs[i].num2]->nodeName,
                                                        node_pairs[i].cost);
    }
  }

  for(i=0;i<node_num + 1; i++)
    FREE(enc_node[i]);
  FREE(enc_node);
 return;

} /* end of derive_one_hot_constraints */


one_hot_like_encodings()
{

 char encode[NUM_INPUTS] ;
 int generate_codes(), comb();
 int i, num_lits;
 CUBE_LIST cptr;
 double ceil();
 int extra_bits;
 int etype = 0;


  if (etype is 0) {
    extra_bits = num_states - (int) ceil(log( (double) node_num));
    if (debug)
      (void) printf("# Extra bits for encoding is %d\n", extra_bits);

    for(i=0;i<num_states;i++) encode[i] = '-';
    encode[num_states] = '\0';

    for(i=0;i<extra_bits;i++) {
      encode[num_states-i-1] = '1';
      cptr = ALLOC(struct cube_list, 1);
      cptr->cube = ALLOC(char, num_states + 1);
      cptr->assigned = FALSE;
      cptr->node = NIL(struct node);
      cptr->num = i;
      (void) strcpy(cptr->cube, encode);

      poss_encodings[i] = cptr;
      encode[num_states-i-1] = '-';
    }

    encode_num = extra_bits;
    minterm_encodings(num_states - extra_bits);

    for(i=0;i<extra_bits;i++) encode[i] = '0';
    encode[extra_bits] = '\0';
    for(i=extra_bits;i<encode_num;i++) {
      (void) strcat(poss_encodings[i]->cube, encode);
    }
  }
  else if (etype is 1) {
    num_lits = 0;
    for(i=1;i<=num_states;i++) {
      if ( comb(num_states, i) >= node_num) {
        num_lits = i;
        break;
      }
    }
    if (num_lits is 0)
      error("not enough encoding bits for one-hot-like encoding");
    if (debug) {
      (void) printf("# Number of literals in each state is %d\n", num_lits);
    }
    encode_num = comb(num_states, num_lits);
  
    for(i=0;i<num_lits;i++) encode[i] = '1';
    for(i=num_lits;i<num_states;i++) encode[i] = '-';
    encode[num_states] = '\0';
  
    i = 0;
    do {
      cptr = ALLOC(struct cube_list, 1);
      cptr->cube = ALLOC(char, num_states + 1);
      cptr->assigned = FALSE;
      cptr->node = NIL(struct node);
      cptr->num = i;
      (void) strcpy(cptr->cube, encode);
  
      poss_encodings[i] = cptr;
      i++;
    } while (!generate_codes(encode, num_states, num_lits, '-'));
    
    if (i isnt encode_num) error("haven't generated all possible codes");
  }

 return;

} /* end of one_hot_like_encodings */


correct_next_state_codes()
{

 int i, j;

  for(i=1;i<=node_num;i++) {
    for(j=0;j<num_states;j++) {
      if (nodes[i]->code[0][j] is '-')
        nodes[i]->code[1][j] = '0';
    }
  }
 return;

} /* end of correct_next_state_codes */
