#include "copyright.h"
#include "mustang.h"


/*
 *    ***********************************
 *    *** SEQUENTIAL STATE ASSIGNMENT ***
 *    ***********************************
 *
 *  Srinivas Devadas
*/


state_assignment(next_state, present_state, arandom, sequential, two_level, approximate)
int next_state, present_state;
int arandom, sequential;
int two_level, approximate;
{
/*
* Assign binary codes to the states in the state transition
* graph such that the number of common sub-expressions is
* maximized.
*/

 int i;

  encode_num = 0;
  if (onehotlike)
    one_hot_like_encodings();
  else
    minterm_encodings(num_states);
/*
* allocate memory for optimal assignment of next state and present
* state nodes and in node data structure.
*/
  for(i=1;i<=node_num;i++) {
    nodes[i]->code[0] = ALLOC(char, num_states+1);
    nodes[i]->code[1] = ALLOC(char, num_states+1);
  }

  if (arandom) {
    assign_random_codes();
    if (onehotlike)
      correct_next_state_codes();
    write_output();
    add_dont_cares(0);
    return;
  }

  if (sequential) {
    assign_sequential_codes();
    if (onehotlike)
      correct_next_state_codes();
    write_output();
    add_dont_cares(0);
    return;
  }
  node_pairs = ALLOC(PAIR, node_num*node_num + 1);

  if (present_state) {
    assign_present_state_codes(two_level, approximate);
    if (expand)
      exploit_dont_cares(0);
    for(i=1;i<=node_num;i++)
      (void) strcpy(nodes[i]->code[1], nodes[i]->code[0]);
    if (onehotlike)
      correct_next_state_codes();
  }
  else if (next_state) {
    assign_next_state_codes(two_level, approximate);
    if (expand)
      exploit_dont_cares(1);
    for(i=1;i<=node_num;i++)
      (void) strcpy(nodes[i]->code[0], nodes[i]->code[1]);
    if (onehotlike)
      correct_next_state_codes();
  }
  else {
/*
* Basically start with an initial assignment and do an annealing
* based on a cost function
*/
    simulated_anneal(TOTAL);
    for(i=1;i<=node_num;i++) {
      nodes[i]->assigned = TRUE;
      (void) strcpy(nodes[i]->code[1], poss_encodings[nodes[i]->enc]->cube);
    }
    if (expand)
      exploit_dont_cares(1);
    for(i=1;i<=node_num;i++)
      (void) strcpy(nodes[i]->code[0], nodes[i]->code[1]);
    if (onehotlike)
      correct_next_state_codes();
  }
  write_output();
  if (!expand)
    add_dont_cares(0);

  
 return;

} /* end of state_assignment */



assign_next_state_codes(two_level, approximate)
int two_level, approximate;
{
/*
* Assign the next state codes maximizing the number of common
* sub-expressions in the edge inputs.
*/

 int i, j, k, count;
 int **st_node, **in_node /*, **op_node*/;
 int cost;

/*
* look at the fanout edges of every state and decide which states
* are to be kept as close as possible.
* Also look at each input literal and group next states for 0 or 1
*/
  st_node = ALLOC(int *, node_num + 1);
  for(i=0; i< node_num + 1; i++ )
    st_node[i] = ALLOC(int, node_num + 1);
 
  in_node = ALLOC(int *, 2 * num_inputs + 1);
  for(i=0; i< 2 * num_inputs + 1; i++ )
    in_node[i] = ALLOC(int, node_num + 1);
 
/*
  op_node = ALLOC(int *, num_outputs + 1);
  for(i=0; i< num_outputs + 1; i++ )
    op_node[i] = ALLOC(int, node_num + 1);
*/

  for(i=1;i<=node_num;i++)
    for(j=1;j<=node_num;j++)
      st_node[i][j] = 0;

  for(i=0;i<2*num_inputs;i++)
    for(j=1;j<=node_num;j++)
      in_node[i][j] = 0;

/*
  for(i=0;i<num_outputs;i++)
    for(j=1;j<=node_num;j++)
      op_node[i][j] = 0;
*/
/*
* st_node has the information about what nodes each node fans out of
*/
  for(i=1;i<=edge_num;i++) {
    st_node[edges[i]->fr_node->num][edges[i]->to_node->num] += 1;
  }
/*
* in_node has the information about what inputs are high or low
* when the state is the next state
*/
  for(i=0;i<num_inputs;i++) {
    for(j=1;j<=edge_num;j++) {
      if (edges[j]->input[i] is '0') 
        in_node[2*i][edges[j]->to_node->num] += 1;
      else if (edges[j]->input[i] is '1')
        in_node[2*i+1][edges[j]->to_node->num] += 1;
    }
  }
/*
* finally, we have the information as to what present states produce
* the same outputs
*/
/*
  for(i=0;i<num_outputs;i++) {
    for(j=1;j<=edge_num;j++) {
      if (edges[j]->output[i] is '1')
        op_node[i][edges[j]->fr_node->num] += 1;
    }
  }
*/

  count = 0;
  for(i=1;i<=node_num;i++) {
    for(j=i+1;j<=node_num;j++) {
      node_pairs[count].num1 = i;
      node_pairs[count].num2 = j;
      cost = 0;

      /* nodes fanning out of same node */
      for(k=1;k<=node_num;k++)
        cost += st_node[k][i] * st_node[k][j];
      if (two_level is FALSE) {
        cost *= num_states;

        /* fanout nodes with same inputs asserted or low */
        for(k=0;k<2*num_inputs;k++)
          cost += in_node[k][i] * in_node[k][j];
      }

      /* present states which assert the same output */
/*
      for(k=0;k<num_outputs;k++)
        cost += op_node[k][i] * op_node[k][j];
*/

      node_pairs[count].cost = cost;
      if (! annealing || cost > 0)
        count++;
    }
  }

  if (approximate) {
    (void)printf("Number of edges is %d\n", count);
    (void)printf("Give number to be used:");
    (void)fscanf(stdin, "%d", &num_node_pairs);
    sort_node_pairs(node_pairs, num_node_pairs);
  }
  else {
    sort_node_pairs(node_pairs, count);
    num_node_pairs = count;
  }

  if (annealing) {
    simulated_anneal(GRAPH);
    for(i=1;i<=node_num;i++) {
      nodes[i]->assigned = TRUE;
      (void) strcpy(nodes[i]->code[1], poss_encodings[nodes[i]->enc]->cube);
    }
  }
  else {
    wedge_clustering(1);
    if (debug) {
      (void)printf("# Cost of assignment is %d\n",
                                              find_cost(TRUE,NULL,GRAPH));
    }
  }

 return;

} /* end of assign_next_state_codes */
 


sort_node_pairs(v, n)
PAIR v[];
int n;
{
  int gap,i,j;
  int num1, num2, cost;

  for(gap=n/2;gap > 0;gap /= 2) {
    for(i=gap;i < n ; i++) {
      for(j = i-gap;j >= 0 ;j -= gap) {
        if (v[j].cost < v[j+gap].cost) {
          cost = v[j].cost;
          num1 = v[j].num1;
          num2 = v[j].num2;

          v[j].cost = v[j + gap].cost;
          v[j].num1 = v[j+gap].num1;
          v[j].num2 = v[j+gap].num2;

          v[j + gap].cost = cost;
          v[j + gap].num1 = num1;
          v[j + gap].num2 = num2;
        }
      }
    }
  }
   
 return;

} /* end of sort_node_pairs */


minterm_encodings(num_bits)
int num_bits;
{

 int i, j;
 int number, num_codes;
 int encode[NUM_NODES];
 char ch_encode[NUM_NODES];
 int etype = 0;
 CUBE_LIST cptr;

  num_codes = 1;
  for(i=0;i<num_bits;i++)
    num_codes = num_codes * 2;

  if (etype is 0) {
    for(i=0;i<num_codes;i++) {
      number = i;
      for(j=num_bits-1;j>=0;j--) {
        encode[j] = number % 2;
        number /= 2;
      }
      cptr = ALLOC(struct cube_list, 1);
      cptr->cube = ALLOC(char, num_states + 1);
      cptr->assigned = FALSE;
      cptr->node = NIL(struct node);
      cptr->num = i + encode_num;
  
      poss_encodings[i + encode_num] = cptr;
  
      for(j=0;j<num_bits;j++) {
        if (encode[j] is 0) cptr->cube[j] = '0';
        else cptr->cube[j] = '1';
      }
      cptr->cube[num_bits] = '\0';
    }
    encode_num += num_codes;
  }
  else if (etype is 1) {

    for(i=0;i<num_states;i++) ch_encode[i] = '0';
    ch_encode[num_states] = '\0';
  
    i = 0;
    do {
      cptr = ALLOC(struct cube_list, 1);
      cptr->cube = ALLOC(char, num_states + 1);
      cptr->assigned = FALSE;
      cptr->node = NIL(struct node);
      cptr->num = i + encode_num;
      (void) strcpy(cptr->cube, ch_encode);
  
      poss_encodings[i + encode_num] = cptr;
      i++;
    } while (!generate_codes(ch_encode, num_states, num_states, '0'));

    if (i isnt num_codes) error("havent generated all possible codes");
    encode_num += num_codes;
  }
 return;

} /* end of minterm_encodings */
